//=========================================================================
// Send queries to DB-Library (Microsoft MS SQL) and get answers from it
//=========================================================================
unit VolgaDAO;

interface
uses Windows, SysUtils, VolClass, ComObj,
{$IFDEF VER180}Variants, {$ENDIF}
ActiveX;

const
    //RecordsetOption
  dbDenyWrite = 1;
  dbDenyRead = 2;
  dbReadOnly = 4;
  dbAppendOnly = 8;
  dbInconsistent = 16;
  dbConsistent = 32;
  dbSQLPassThrough = 64;
  dbFailOnError = 128;
  dbForwardOnly = 256;
  dbSeeChanges = 512;
  dbRunAsync = 1024;
  dbExecDirect = 2048;

    //RecordsetType
  dbOpenTable = 1;
  dbOpenDynaset = 2;
  dbOpenSnapshot = 4;
  dbOpenForwardOnly = 8;
  dbOpenDynamic = 16;

    //EditMode
  dbEditNone = 0;
  dbEditInProgress = 1;
  dbEditAdd = 2;
  dbEditChanged = 4;
  dbEditDeleted = 8;
  dbEditNew = 16;

    //DataType
  dbBoolean = 1;
  dbByte = 2;
  dbInteger = 3;
  dbLong = 4;
  dbCurrency = 5;
  dbSingle = 6;
  dbDouble = 7;
  dbDate = 8;
  dbBinary = 9;
  dbText = 10;
  dbLongBinary = 11;
  dbMemo = 12;
  dbGUID = 15;
  dbBigInt = 16;
  dbVarBinary = 17;
  dbChar = 18;
  dbNumeric = 19;
  dbDecimal = 20;
  dbFloat = 21;
  dbTime = 22;
  dbTimeStamp = 23;

  //WorkspaceType
  dbUseODBC = 1;
  dbUseJet = 2;

type
  TFldAnswer = record
    Head: char;                         //type of the field / length in the description of the record
    Size: integer;                      //size of the field / real data size
    Data: PChar;                        //name of the field / data
  end;
  PFldAnswer = ^TFldAnswer;

  TRecAnswer = record                   //one record or description of the structure
    Count: integer;
    ArrFlds: array of PFldAnswer;
  end;
  PRecAnswer = ^TRecAnswer;

  PQueueRec = ^TQueueRec;               //unidirectional list of the record
  TQueueRec = record
    RecData: PRecAnswer;
    Next: PQueueRec;
  end;

  TServAnswer = record
    Buffer: PChar;                      //buffer with answer
    Size: integer;                      //size of buffer
    Rez: integer;                       //result/err cod
  end;
  PServAnswer = ^TServAnswer;

  TVolgaDAOLink = class(TObject)
  private
    FEngine: Variant;
    FDatabase: Variant;
    FWorkSpace: Variant;
    FRecordSet: Variant;
    FField: Variant;
    FAnswDef: PRecAnswer;               //fields of the answers descriptions
    FirstRec: PQueueRec;                //the first record of the list
    LastRec: PQueueRec;                 //the last record of the list
    FAnswDataSize: integer;             //for calculation of the data size
    FAnswRecCount: integer;             //count of the records in answer
    FLoadedDLL: Boolean;
    FConnected: Boolean;
    ARecs: Variant;
    function FillAnswDef: Boolean;
    function FillAnswRec(var pRec: PRecAnswer; ixRow: integer): integer;
    function FillAnswFld(coltype: char; var pFld: PFldAnswer; const RecFld: Variant): integer;
    procedure FreeAnswRec(pRec: PRecAnswer);
    procedure FreeAnswTab;
    procedure PrepareAnswer;            //fill the buffer of the answer 
    function SQLToVolga(coltype, Size: integer): char;
    function GetSQLErrorMsg: string;    //convert SQL type to Volga type
    procedure InitAnswer;
    //function Err_handler: string;
    procedure CopyVarToBuf(Buf: PChar; Value: TVarData; coltype: char);
    procedure AnswerMsg(Msg: string);
    procedure ErrAnswer(Cod: integer; Msg: string);
    procedure DoSelect(InBuf: string);
    procedure DoEdit(InBuf: string);
    procedure DoDDL(InBuf: string);
//    function DoScript(InBuf: PChar; InSize: integer): PServAnswer;
    procedure StartTransaction;
    procedure Commit;
    procedure Rollback;
  public
    FAnswer: PServAnswer;
    procedure DoSQL(InBuf: PChar; InSize: integer);
    constructor Create;
    destructor Destroy; override;
    function Connect(DBName, ConnStr: string): Boolean;
    procedure Disconnect;
    procedure FreeAnswer(QBuf: PChar);
    property SQLErrorMsg: string read GetSQLErrorMsg;
    property Connected: Boolean read FConnected;
    property IsLoaded: Boolean read FLoadedDLL;
  end;

var MSLink: TVolgaDAOLink;

implementation

const SIZEINT = sizeof(integer);
var  ErrAlarm: integer = 0;   // ------------
var  SQLMess: string = '';    //-------------

{ TVolgaDAOLink }

constructor TVolgaDAOLink.Create;
begin
  FLoadedDLL := false;
  FConnected := false;
  FEngine := Unassigned;
  FDatabase := Unassigned;
  FWorkSpace := Unassigned;
  FRecordSet := Unassigned;
  CoInitialize(nil);
  try
    FEngine := CreateOleObject('DAO.PrivateDBEngine.36');
    FEngine.DefaultType := dbUseJet;
    FWorkSpace := FEngine.Workspaces[0]; //first from the list
    FLoadedDLL := true;
  except
  end;
end;

destructor TVolgaDAOLink.Destroy;
begin
  Disconnect;
  FWorkSpace := Unassigned;
  FEngine := Unassigned;
  CoUninitialize;
  inherited;
end;

function TVolgaDAOLink.Connect(DBName, ConnStr: string): Boolean;
var err: Variant;
  iErr: integer;
  errmsg: string;
begin
  Result := false;
  if not IsLoaded then Exit;
  if Connected then Disconnect;
  try
    FDatabase := FWorkSpace.OpenDatabase(DBName, false, false, ConnStr);
  except
    for iErr := 0 to FEngine.Errors.Count - 1 do
    begin
      err := FEngine.Errors[iErr];
      errmsg := Format('%d : %s : %s', [err.Number, err.Description, err.Source]);
    end;
    MessageBox(NULL, PChar(errmsg), 'ForceError', MB_OK);
  end;
  FConnected := not VarIsEmpty(FDatabase);
  Result := FConnected;
end;

procedure TVolgaDAOLink.Disconnect;
begin
  if FConnected then
    FDatabase.Close;
  FDatabase := Unassigned;
  FConnected := false;
end;

procedure TVolgaDAOLink.Commit;
begin
  if Connected then
  begin
    FWorkSpace.CommitTrans;
  end;
end;

procedure TVolgaDAOLink.Rollback;
begin
  if Connected then
  begin
    FWorkSpace.RollBack;
  end;
end;

procedure TVolgaDAOLink.StartTransaction;
begin
  if Connected then
  begin
    FWorkSpace.BeginTrans;
  end;
end;

procedure TVolgaDAOLink.DoSQL(InBuf: PChar; InSize: integer);
var SQLText, keyword: string;
    SQLUText: String;   // Sql upper text
begin
  //define the type SQL-sentence
  SQLText := Trim(StrPas(InBuf));
  keyword := UpperCase(Copy(SQLText,1 ,6));
  If keyword = 'SELECT' then
  begin
    SqlUText := UpperCase(SqlText);
    If Pos('INTO', SqlUText) > 0 Then
      DoEdit(SQLText)
    else
      DoSelect(SQLText);
  end
  else if (keyword = 'UPDATE') or (keyword = 'DELETE')
  or (keyword = 'INSERT') then
    DoEdit(SQLText)
  else if (keyword = 'CREATE') or (keyword = 'ALTER ')
    or (copy(keyword, 1, 5) = 'DROP ') then
    DoDDL(SQLText)
  else  //script?
    ErrAnswer(-1, 'Error in SQL statement');
end;

procedure TVolgaDAOLink.DoEdit(InBuf: string);
var
  CmdBuf: TDoEditBuf;
begin
  ErrAlarm := 0;
  InitAnswer;
  if not Connected then
  begin
    ErrAnswer(-1, 'No connection to DAO');
    Exit;
  end;
  //parse received sentences on separate commands 
  CmdBuf := TDoEditBuf.Create(InBuf);
  //transaction begining
  StartTransaction;
  while CmdBuf.NextCmd = 1 {<> NO_MORE_Cmd} do
  begin
    try
      FDatabase.Execute(string(CmdBuf.cmdbuf), dbFailOnError);
    except
      ErrAnswer(-1, 'One of the string edit statements failed.');
      Break;
    end;
  end;
  //commit or rollback - separately 
  if ErrAlarm < 0 then
  begin
    RollBack;
    ErrAnswer(-1, 'Error when update data.');
  end else begin
    Commit;
    AnswerMsg('All OK');
  end;
  if CmdBuf <> nil then CmdBuf.Free;
end;

procedure TVolgaDAOLink.DoDDL(InBuf: string);
begin
  InitAnswer;
  if not Connected then
  begin
    ErrAnswer(-1, 'No connection to DAO');
    Exit;
  end;
  try
    FDatabase.Execute(InBuf, dbFailOnError);
    AnswerMsg('All OK');
  except
    ErrAnswer(-1, 'DDL statement failed.');
  end;
end;

{function TVolgaDAOLink.DoStored(InBuf: PChar; InSize: integer): PServAnswer;
begin
  InitAnswer;
  if not Connected then
    ErrAnswer(-1, 'Connection error');
  Result := FAnswer;
end;}

procedure TVolgaDAOLink.DoSelect(InBuf: string);
var
  pRec: PRecAnswer;
  QueueRec: PQueueRec;
  lenrec, j: integer;
begin
  ErrAlarm := 0;
  InitAnswer;
  if not Connected then
  begin
    ErrAnswer(-1, 'No connection to DAO');
    Exit;
  end;
  try
    FRecordSet := FDatabase.OpenRecordset(InBuf, dbOpenForwardOnly, dbReadOnly);
  except
    ErrAnswer(-1, 'Error in SELECT statement');
    Exit;
  end;
  //create description of the fields
  if FillAnswDef then
  begin
    FAnswDataSize := 0;
    FAnswRecCount := 0;
    FirstRec := nil;                    //pointer on the first record 
    LastRec := nil;                     //pointer on the last record 
    while not FRecordSet.EOF do
    begin
      ARecs := FRecordSet.GetRows(5);  //read 5 records immediately
      for j := 0 to VarArrayHighBound(ARecs, 2) do
      begin
        lenrec := FillAnswRec(pRec, j);
        if lenrec >= 0 then             //fill record with data
        begin                           //creating of the unidirectional list of the records
          New(QueueRec);
          QueueRec^.RecData := pRec;
          QueueRec^.Next := nil;
          if FirstRec = nil then FirstRec := QueueRec;
          if LastRec <> nil then PQueueRec(LastRec)^.Next := QueueRec;
          LastRec := QueueRec;
          Inc(FAnswRecCount);           //records count
          FAnswDataSize := FAnswDataSize + lenrec + FAnswDef^.Count;  //total size of the buffer for data 
        end;
      end;
    end
  end
  else
  begin
    FRecordSet.Close;
    ErrAnswer(-1, 'Set descriptor fields failed.');
    FreeAnswTab;
  end;
  if FAnswer^.Rez >= 0 then begin
    FRecordSet.Close;
    FRecordSet := Unassigned;
  end;
  PrepareAnswer;                      //fill the buffer of the answer, clear internal fields 
end;

//transformation of the types

function TVolgaDAOLink.SQLToVolga(coltype, Size: integer): char;
begin
  Result := char(0);
  case coltype of
    dbBoolean: Result := 'b';
    dbChar, dbText:
      if Size > 35 then
        Result := 's'
      else
      begin
        if Size = 0 then
          Result := '?'
        else if Size > 9 then
          Result := char(ord('A') + (Size - 10))
        else
          Result := char(ord('1') + (Size - 1));
      end;
    dbInteger: Result := 'l';
    dbLong: Result := 'i';
    dbSingle, dbDouble, dbFloat, dbDecimal, dbNumeric: Result := 'f';
    dbCurrency: Result := 'c';
    dbDate: Result := 'd';
    dbTime: Result := 't';
    dbMemo: Result := 'm';
  end;
end;

procedure TVolgaDAOLink.CopyVarToBuf(Buf: PChar; Value: TVarData; coltype: char);
//write value in the buffer of the field, entered as variant 
var ri: integer;
  rd: double;
  rt: TTimeStamp;
  rc: TDateTime;
begin
  case Value.VType of
    varSmallint:
      Move(Value.VSmallint, Buf^, sizeof(SmallInt));
    varInteger:
      Move(Value.VInteger, Buf^, sizeof(Integer));
    varBoolean:
      Move(Value.VBoolean, Buf^, sizeof(WordBool));
    varSingle:
      begin
        rd := Value.VSingle;
        Move(rd, Buf^, sizeof(double));
      end;
    varDouble:
      Move(Value.VDouble, Buf^, sizeof(double));
    varCurrency:
      begin
        rd := Value.VCurrency;
        Move(rd, Buf^, sizeof(double));
      end;
    varDate:
      begin
        rc := Value.VDate;
        rt := DateTimeToTimeStamp(rc);
        case coltype of
        'd':
          begin
            ri := rt.Date;
            Move(ri, Buf^, SIZEINT);
          end;
        't':
          begin
            ri := rt.Time;
            Move(ri, Buf^, SIZEINT);
          end;
        end;
      end;  //varDate
  end;
end;

//create descriptions of the fields

function TVolgaDAOLink.FillAnswDef: Boolean;
var i, column_count, lennam: integer;
  colnam: string;
  FldName: PChar;
begin
  FAnswDef := nil;
  column_count := FRecordSet.Fields.Count;
  try
    New(FAnswDef);
    FAnswDef^.ArrFlds := nil;
    SetLength(FAnswDef^.ArrFlds, column_count);
    FAnswDef^.Count := column_count;
  except
    FreeAnswRec(FAnswDef);
    Result := false;
    Exit;
  end;
  try
    for i := 0 to column_count - 1 do
    begin
      FField := FRecordSet.Fields[i];
      colnam := FField.Name;
      lennam := Length(colnam);
      New(FAnswDef^.ArrFlds[i]);
      FldName := AllocMem(lennam + 1);
      StrPCopy(FldName, colnam);
      FAnswDef^.ArrFlds[i]^.Data := FldName;
      FAnswDef^.ArrFlds[i]^.Size := FField.Size;
      FAnswDef^.ArrFlds[i]^.Head := SQLToVolga(FField.type, FAnswDef^.ArrFlds[i]^.Size);
      if FAnswDef^.ArrFlds[i]^.Head = '?' then begin
        FAnswDef^.ArrFlds[i]^.Size := 255;  //for string calc fields in select query
        FAnswDef^.ArrFlds[i]^.Head := 's';
      end;
    end;
  except
    FreeAnswRec(FAnswDef);
    Result := false;
    Exit;
  end;
  Result := true;
end;

//fill one record with data

function TVolgaDAOLink.FillAnswRec(var pRec: PRecAnswer; ixRow: integer): integer;
var lenrec, lenfld, i: integer;
begin
  Result := -1;
  pRec := nil;
  try
    New(pRec);
    pRec^.ArrFlds := nil;
    SetLength(pRec^.ArrFlds, FAnswDef^.Count);
    pRec^.Count := FAnswDef^.Count;
  except
    FreeAnswRec(pRec);
    Exit;
  end;
  lenrec := 0;
  for i := 0 to pRec^.Count - 1 do
  begin
    lenfld := FillAnswFld(FAnswDef^.ArrFlds[i]^.Head, pRec^.ArrFlds[i], ARecs[i, ixRow]);
    if lenfld < 0 then
    begin
      FreeAnswRec(pRec);
      Exit;
    end;
    Inc(lenrec, lenfld);
  end;
  Result := lenrec;
end;

//fill one field with data

function TVolgaDAOLink.FillAnswFld(coltype: char; var pFld: PFldAnswer;
  const RecFld: Variant): integer;
var len: integer;
  VarData: TVarData;
  varstr: string;
begin
  Result := -1;
  pFld := nil;
  try
    New(pFld);
    pFld^.Data := nil;
    if (VarType(RecFld) = varEmpty) or (VarType(RecFld) = varNull) then begin
      //pFld^.Head := char(pFld^.Size)
      pFld^.Size := 0;
      pFld^.Head := char(0);
    end
    else
    begin
      varstr := '';
      VarData.VType := varNull;
      if VarType(RecFld) = varOleStr then
        varstr := WideCharToString(TVarData(RecFld).VOleStr)
      else
        VarData := TVarData(RecFld);
      len := 0;
      //special field size
      case coltype of
        'b', 'l': pFld^.Size := sizeof(WordBool);
        'f', 'c': pFld^.Size := sizeof(double);
        'i', 'd', 't': pFld^.Size := SIZEINT;
        else
          begin
            len := Length(varstr);
            if coltype in ['s', 'm', 'g'] then  //long lines and BLOB 
              pFld^.Size := SIZEINT + len
            else  //short lines 
              pFld^.Size := len;
          end;
      end;
      pFld^.Data := AllocMem(pFld^.Size);
      //carrying value in buffer 
      case coltype of
        '1'..'9', 'A'..'Z':
           begin
             pFld^.Head := char(pFld^.Size);
             Move(PChar(varstr)^, (pFld^.Data)^, len);
           end;
        's', 'm', 'g':
          begin
            PInteger(pFld^.Data)^ := len;
            Move(PChar(varstr)^, (pFld^.Data + SIZEINT)^, len);
            pFld^.Head := char(SIZEINT);
          end;
        else
          begin
            CopyVarToBuf(pFld^.Data, VarData, coltype);
            pFld^.Head := char(pFld^.Size);
          end;
      end;
    end;
  except
    Exit;
  end;
  Result := pFld^.Size;
end;

procedure TVolgaDAOLink.PrepareAnswer;  //fill the buffer of the answer
var S, nam: string;
  Flds: PRecAnswer;
  i, offs: integer;
  CurRec: PQueueRec;
  Layout: PChar;

  function RecCpy(Buf: PChar): integer;
  //fill header of the one record with the real length of flilds
  //and fill the field 
  var i: integer;
  begin
    Result := Flds^.Count;              //the size of the record's Header  
    for i := 0 to Flds^.Count - 1 do
    begin
      Buf[i] := Flds^.ArrFlds[i]^.Head; //the size of the field
      if Flds^.ArrFlds[i]^.Size <> 0 then
      begin                             //carrying data in buffer 
        Move((Flds^.ArrFlds[i]^.Data)^, (Buf + Result)^, Flds^.ArrFlds[i]^.Size);
        Result := Result + Flds^.ArrFlds[i]^.Size; //offset
      end;
    end;
  end;

  function FieldDataType2(pFld: PFldAnswer): string;
  begin //type of the field in the table's header  (type and length for lines) 
    Result := '';
    if pFld^.Head = 's' then
      Result := Result + pFld^.Head + IntToStr(pFld^.Size)
    else
      Result := Result + pFld^.Head;
  end;

begin
  S := '';
  for i := 0 to FAnswDef^.Count - 1 do  //fill table's header
  begin
    nam := string(FAnswDef^.ArrFlds[i]^.Data);
    S := S + 'sql.' + nam + ' ' +
      FieldDataType2(FAnswDef^.ArrFlds[i]) + Chr($12);
  end;
  S := S + Chr($14);
  FAnswer^.Size := FAnswDataSize + SIZEINT * 4 + Length(S) + 1;
  FAnswer^.Buffer := AllocMem(FAnswer^.Size);
  offs := 0;
  PInteger(FAnswer^.Buffer)^ := FAnswer^.Size; //size of the answer's buffer
  Inc(offs, SIZEINT);
  FAnswer^.Buffer[offs] := 'S';
  Inc(offs);
  PInteger(FAnswer^.Buffer + offs)^ := FAnswRecCount; //records count
  Inc(offs, SIZEINT);
  PInteger(FAnswer^.Buffer + offs)^ := FAnswDef^.Count; //fields count
  Inc(offs, SIZEINT);
  PInteger(FAnswer^.Buffer + offs)^ := Length(S); //length of the table's header
  Inc(offs, SIZEINT);
  Layout := AllocMem(Length(S)+1);
  StrPCopy(Layout, S);
  Move(Layout^, (FAnswer^.Buffer + offs)^, Length(S));//table's header
  FreeMem(Layout);
  Inc(offs, Length(S));
  if FirstRec <> nil then
  begin
    CurRec := FirstRec;
    repeat
      Flds := CurRec^.RecData;
      if Flds <> nil then
      begin                             //moving records
        offs := offs + RecCpy(FAnswer^.Buffer + offs);
        CurRec := CurRec^.Next;
      end
      else
        Break;
    until CurRec = nil;
  end;
  FreeAnswTab;
end;

//delete whole table from the memory

procedure TVolgaDAOLink.FreeAnswTab;
var CurRec, NxtRec: PQueueRec;
begin
  FreeAnswRec(FAnswDef);
  if FirstRec <> nil then
  begin
    CurRec := FirstRec;
    repeat
      NxtRec := CurRec^.Next;
      if CurRec <> nil then
      begin
        FreeAnswRec(CurRec^.RecData);
        Dispose(CurRec);
      end;
      CurRec := NxtRec;
    until CurRec = nil;
  end;
end;

// Free Memory fields

procedure TVolgaDAOLink.FreeAnswRec(pRec: PRecAnswer);
var i: integer;
begin
  if pRec <> nil then
  begin
    if FAnswDef^.ArrFlds <> nil then
    begin
      for i := 0 to pRec^.Count - 1 do
        if pRec^.ArrFlds[i] <> nil then
        begin
          if pRec^.ArrFlds[i]^.Data <> nil then
            FreeMem(pRec^.ArrFlds[i]^.Data);
          Dispose(pRec^.ArrFlds[i]);
        end;
      pRec^.ArrFlds := nil;             //free dynamic array 
    end;
    Dispose(pRec);
  end;
end;

procedure TVolgaDAOLink.ErrAnswer(Cod: integer; Msg: string);
var len: integer;
begin
  Msg := Msg + SQLMess;
  if FAnswer <> nil then
  begin
    if ErrAlarm <= -1 then
      FAnswer^.Rez := ErrALarm
    else
      FAnswer^.Rez := Cod;
    ErrAlarm := -1;
    if FAnswer^.Buffer <> nil then FreeMem(FAnswer^.Buffer);
    len := Length(Msg) + 1 + 5;  //4 for length + E + $0 
    FAnswer^.Size := len;
    FAnswer^.Buffer := AllocMem(len);
    Move(len, (FAnswer^.Buffer)^, 4);
    StrPCopy(FAnswer^.Buffer + 4, 'E'+Msg);
  end;
  SQLMess := '';
//Disconnect;
end;

procedure TVolgaDAOLink.AnswerMsg(Msg: string);
var len: integer;
begin
  if FAnswer <> nil then
  begin
    if FAnswer^.Buffer <> nil then FreeMem(FAnswer^.Buffer);
    len := Length(Msg) + 1 + 5;  //4 for length + S + $0
    FAnswer^.Size := len;
    FAnswer^.Buffer := AllocMem(len);
    Move(len, (FAnswer^.Buffer)^, 4);
    StrPCopy(FAnswer^.Buffer + 4, 'S'+Msg);
  end;
end;

function TVolgaDAOLink.GetSQLErrorMsg: string;
begin
  Result := SQLMess;
end;

{function TVolgaDAOLink.Err_handler: string;
var FError: Variant;
  iErr: integer;
begin
  Result := '';
  ErrAlarm := -1;
  for iErr := 0 to FEngine.Errors.Count - 1 do
  begin
    FError := FEngine.Errors[iErr];
    Result := Result + Format('%d : %s : %s',
      [FError.Number, FError.Description, FError.Source]);
  end;
end;}

procedure TVolgaDAOLink.InitAnswer;
begin
  if (FAnswer <> nil) and (FAnswer^.Buffer <> nil) then
    FreeAnswer(FAnswer^.Buffer);
  FAnswer^.Buffer := nil;
  FAnswer^.Size := 0;
  FAnswer^.Rez := 0;
end;

procedure TVolgaDAOLink.FreeAnswer(QBuf: PChar);
begin
  if QBuf <> nil then FreeMem(QBuf);
end;

end.

