//******************************************************************************
//  Components for export from TVolgaDatabset to XML, DBF files
//---------------------------------------------------------------------------
//  Copyright  2000, Olga Vlasova, Russia
//  http://www.volgadb.com
//  E-mail: info@volgadb.com
//---------------------------------------------------------------------------
//  Support Kylix/Lilux platform: (c) Vit (Vitaly Nevzorov), 2004, Chicago, USA
//  http://www.delphist.com
//  E-mail: vit@vingrad.ru
//******************************************************************************
unit VolgaExp;

interface

uses
{$IfnDef Linux}Windows, Messages, ComObj, ActiveX, {$EndIf non-Linux}
{$IFDEF VER140}Variants, {$ENDIF}
{$IFDEF VER150}Variants, {$ENDIF}
{$IFDEF VER180}Variants, {$ENDIF}
  SysUtils, Classes, DB;

type

  TExportCallbackEvent = procedure(const Percent: integer) of object;

  //****************************************************************************
  //  TVolgaXMLConverter - exports any TDataset to XML-format
  //****************************************************************************
  TVolgaXMLConverter = class(TComponent)
  private
    FDataset: TDataset;
    FXMLText: TStrings;
    FOrigin: string;
    FEncoding: string;
    FOnCallback: TExportCallbackEvent;
    procedure SetDataset(const Value: TDataset);
    procedure SetOrigin(const Value: string);
    { Private declarations }
  protected
    { Protected declarations }
    procedure PrepareFieldDef(AField: TField);
    procedure PrepareSchema;
    procedure PrepareRow;
    procedure PrepareHeader;
    procedure ConvertToXML;
    function XMLstr(text: string): string;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure XMLToStream(AStream: TStream);
    procedure XMLToFile(AFileName: string);
  published
    { Published declarations }
    property Dataset: TDataset read FDataset write SetDataset;  //dataset, converting to XML
    property Origin: string read FOrigin write SetOrigin; //the table's name in database
    property Encoding: string read FEncoding write FEncoding;  //coding
    property OnCallback: TExportCallbackEvent read FOnCallback write FOnCallback;
  end;

  //****************************************************************************
  //  TVolgaDBFConverter - exports any TDataset to DBF-format
  //****************************************************************************
  TVolgaDBFConverter = class(TComponent)
  private
    { Private declarations }
    FDataset: TDataset;
    FToOEM: Boolean;
    FOnCallback: TExportCallbackEvent;
    procedure SetDataset(const Value: TDataset);
    function CodePageNum: char;
  protected
    { Protected declarations }
    procedure PrepareHeader(AStream: TStream);
    procedure PrepareRow(AStream: TStream);
    procedure ConvertToDBF(AStream: TStream);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DBFToStream(AStream: TStream);
    procedure DBFToFile(AFileName: string);
  published
    { Published declarations }
    property Dataset: TDataset read FDataset write SetDataset;  //dataset, converting to XML
    property ToOEM: Boolean read FToOEM write FToOEM default true;
    property OnCallback: TExportCallbackEvent read FOnCallback write FOnCallback;
  end;

{$IfnDef Linux}
  //****************************************************************************
  //  TVolgaXLSConverter - exports any TDataset to Excel-format
  //****************************************************************************
  TVolgaXLSConverter = class(TComponent)
  private
    { Private declarations }
    FDataset: TDataset;
    FOnCallback: TExportCallbackEvent;
    Fn: string;
    procedure SetDataset(const Value: TDataset);
  protected
    { Protected declarations }
    function PrepareHeader: string;
    function PrepareRow: string;
    procedure ConvertToCsv;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure XLSToFile(AFileName: string);
  published
    { Published declarations }
    property Dataset: TDataset read FDataset write SetDataset;  //dataset, converting to Excel
    property OnCallback: TExportCallbackEvent read FOnCallback write FOnCallback;
  end;
{$EndIf non-Linux}

implementation

{ TVolgaXMLConverter }

constructor TVolgaXMLConverter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FXMLText := TStringList.Create;
  FOrigin := 'Table';
  FEncoding := 'windows-1252';
end;

destructor TVolgaXMLConverter.Destroy;
begin
  FXMLText.Free;
  FDataset := nil;
  inherited Destroy;
end;

procedure TVolgaXMLConverter.ConvertToXML;
var perc, tek, i: integer;
begin
  if not FDataset.Active then FDataset.Open;
  FXMLText.Clear;
  PrepareHeader;
  PrepareSchema;
  FXMLText.Add('<rs:data>');
  FDataset.First;
  perc := 0;
  i := 0;
  while not FDataset.Eof do begin
    Inc(i);
    PrepareRow;
    FDataSet.Next;
    tek := Round(i * 100 / FDataset.RecordCount);
    if (perc <> tek) and Assigned(FOnCallback) then
      FOnCallback(tek); //if the percent realy changed (to prevent often jerking)
    perc := tek;
  end;
  FXMLText.Add('</rs:data>');
  FXMLText.Add('</xml>');
end;

procedure TVolgaXMLConverter.PrepareFieldDef(AField: TField);
begin
  FXMLText.Add('<s:AttributeType name="' + AField.FieldName + '" rs:number="'
    + IntToStr(AField.Index) + '"');
  FXMLText.Add('rs:basetable="' + FOrigin + '" rs:basecolumn="' + AField.FieldName
    + '" rs:maybenull="true">');
//  if AField.IsIndexField then
//    FXMLText.Add('rs:keycolumn="true">');
  case AField.DataType of
    ftString:
      FXMLText.Add('<s:datatype dt:type="string" dt:maxLength="' +
        IntToStr(AField.Size) + '"/>');
    ftSmallint:
      FXMLText.Add('<s:datatype dt:type="i2" dt:maxLength="2" rs:precision="5"'
        + ' rs:fixedlength="true"/>');
    ftInteger, ftAutoInc:
      FXMLText.Add('<s:datatype dt:type="int" dt:maxLength="4" rs:precision="10"'
        + ' rs:fixedlength="true"/>');
    ftWord:
      FXMLText.Add('<s:datatype dt:type="ui2" dt:maxLength="2" rs:precision="5"'
        + ' rs:fixedlength="true"/>');
    ftBoolean:
      FXMLText.Add('<s:datatype dt:type="boolean"/>');
    ftFloat:
      FXMLText.Add('<s:datatype dt:type="float" dt:maxLength="8"'
        + ' rs:precision="19" rs:fixedlength="true"/>');
    ftCurrency:
      FXMLText.Add('<s:datatype dt:type="i8" rs:dbtype="currency" dt:maxLength="8"'
        + ' rs:precision="19" rs:fixedlength="true"/>');
    ftDate:
      FXMLText.Add('<s:datatype dt:type="date"/>');    //dt:type="date.iso8601">');
    ftTime:
      FXMLText.Add('<s:datatype dt:type="time.iso8601"/>');
    ftDateTime:
      FXMLText.Add('<s:datatype dt:type="datetime.iso8601"/>');
    else ;//skiping ftBlob;ftMemo;ftGraphic;ftFmtMemo;ftBlob;
  end;
      FXMLText.Add('</s:AttributeType>');
end;

procedure TVolgaXMLConverter.PrepareHeader;
begin
  FXMLText.Add('<?xml version="1.0" encoding="' + FEncoding + '"?>');
  FXMLText.Add('<xml xmlns:s="uuid:BDC6E3F0-6DA3-11d1-A2A3-00AA00C14882"');
  FXMLText.Add('xmlns:dt="uuid:C2F41010-65B3-11d1-A29F-00AA00C14882"');
  FXMLText.Add('xmlns:rs="urn:schemas-microsoft-com:rowset"');
  FXMLText.Add('xmlns:z="#RowsetSchema">');
end;

procedure TVolgaXMLConverter.PrepareRow;
var s: string; i: integer;
begin
  s := '<z:row';
  for i := 0 to FDataset.FieldCount - 1 do
    if not (FDataset.Fields[i].DataType in [ftBlob,ftMemo,ftGraphic,ftFmtMemo]) then
      s := s + ' '+ FDataset.Fields[i].FieldName + '="' +
      XMLStr(FDataset.Fields[i].AsString) + '"';
  s := s + '/>';
  FXMLText.Add(s);
end;

procedure TVolgaXMLConverter.PrepareSchema;
var i: integer;
begin
  FXMLText.Add('<s:Schema id="RowsetSchema">');
  FXMLText.Add('<s:ElementType name="row" content="eltOnly">');
  for i := 0 to FDataset.FieldCount - 1 do
    PrepareFieldDef(FDataset.Fields[i]);
  FXMLText.Add('</s:ElementType>');
  FXMLText.Add('</s:Schema>');
end;

procedure TVolgaXMLConverter.XMLToFile(AFileName: string);
begin
  if (FDataset = nil) or (AFileName = '') then Exit;
  ConvertToXML;
  FXMLText.SaveToFile(AFileName);
end;

procedure TVolgaXMLConverter.XMLToStream(AStream: TStream);
begin
  if (FDataset = nil) or (AStream = nil) then Exit;
  ConvertToXML;
  FXMLText.SaveToStream(AStream);
end;

procedure TVolgaXMLConverter.SetDataset(const Value: TDataset);
begin
  FDataset := Value;
end;

procedure TVolgaXMLConverter.SetOrigin(const Value: string);
begin
  FOrigin := Value;
end;

procedure TVolgaXMLConverter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FDataset) then
    FDataset := nil;
end;

function TVolgaXMLConverter.XMLstr(text: string): string;
begin
  Result := StringReplace(text, '&', '&amp;', [rfReplaceAll]);
  Result := StringReplace(Result, '"', '&quot;', [rfReplaceAll]);
  Result := StringReplace(Result, '<', '&lt;', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
end;

{ TVolgaDBFConverter }

constructor TVolgaDBFConverter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FToOEM := true;
end;

destructor TVolgaDBFConverter.Destroy;
begin
  FDataset := nil;
  inherited Destroy;
end;

procedure TVolgaDBFConverter.DBFToFile(AFileName: string);
var FS: TMemoryStream;
begin
  if (FDataset = nil) or (AFileName = '') or (UpperCase(ExtractFileExt(AFileName))
   <> '.DBF') then Exit;
  FS := TMemoryStream.Create;
  try
    DBFToStream(FS);
    FS.SaveToFile(AFileName);
  finally
    FS.Free;
  end;
end;

procedure TVolgaDBFConverter.DBFToStream(AStream: TStream);
begin
  if (FDataset = nil) or (AStream = nil) then Exit;
  ConvertToDBF(AStream);
end;

procedure TVolgaDBFConverter.ConvertToDBF(AStream: TStream);
var ch: char;
  perc, tek, i: integer;
begin
  if not FDataset.Active then FDataset.Open;
  PrepareHeader(AStream);
  FDataset.First;
  perc := 0;
  i := 0;
  while not FDataset.Eof do begin
    Inc(i);
    PrepareRow(AStream);
    FDataSet.Next;
    tek := Round(i * 100 / FDataset.RecordCount);
    if (perc <> tek) and Assigned(FOnCallback) then
      FOnCallback(tek); //if the percent realy changed (to prevent often jerking)
    perc := tek;
  end;
  ch := char($1A);
  if FDataset.RecordCount > 0 then
    AStream.Write(ch, 1); //end of the file
end;

procedure TVolgaDBFConverter.PrepareHeader(AStream: TStream);
var yy,mm,dd: word; aa: array[0..2] of char; apos,aint,size,off,frac,i: integer;
  fn: PChar; typ,ch : char; bsize: byte; wpos: word;
begin
  ch := char($03);
  AStream.Write(ch, 1);  //0 type of file ($03 dBASE IV no memo, $8B dBASE IV with memo)
  DecodeDate(Date, yy, mm, dd);
  yy := yy - 1900;
  aa[0] := char(yy);
  aa[1] := char(mm);
  aa[2] := char(dd);
  AStream.Write(aa, 3);  //1-3 Last update (YYMMDD)
  aint := FDataset.RecordCount;
  AStream.Write(aint, 4); //4-7 Number of records in file
  //8-9 Position of first data record
  //10-11 Length of one data record (including delete flag)
  AStream.Position := 12;
  aint := 0;
  AStream.Write(aint, 4); //12-27 reserved
  AStream.Write(aint, 4);
  AStream.Write(aint, 4);
  AStream.Write(aint, 4);
  ch := char(0);
  AStream.Write(ch, 1);  //28 Table Flags  ($01file has a structural .cdx, $02file has a Memo field)
  ch := CodePageNum;
  AStream.Write(ch, 1);  //29 Code page mark
  dd := 0;
  AStream.Write(dd, 2);  //30-31 - reserved
  //Field subrecords
  fn := AllocMem(11);
  off := 1;
  for i := 0 to FDataset.FieldCount - 1 do
  begin
    FillChar(fn^, 11, $00);
    StrLCopy(fn, PChar(FDataset.Fields[i].FieldName), 10);
    AStream.Write(fn^, 11); //0-10 Field name
    frac := 0;
    case FDataset.Fields[i].DataType of
    ftString: begin typ := 'C'; size := FDataset.Fields[i].Size; end;
    ftSmallint: begin typ := 'N'; size := 5; end;
    ftInteger: begin typ := 'N'; size := 10; end;
    ftWord: begin typ := 'N'; size := 5; end;
    ftBoolean: begin typ := 'L'; size := 1; end;
    ftFloat: begin typ := 'N'; size := 16; frac := 4; end;
    ftCurrency: begin typ := 'N'; size := 16; frac := 4; end;
    ftDate: begin typ := 'D'; size := 8; end;
    ftTime: begin typ := 'C'; size := 8; end;
    ftDateTime: begin typ := 'C'; size := 19; end;
    ftAutoInc: begin typ := 'N'; size := 10; end;
    ftMemo:  begin typ := 'M'; size := 10; end;
    ftGraphic, ftBlob: begin typ := 'P'; size := 10; end;
    ftFmtMemo: begin typ := 'M'; size := 10; end;
    else begin typ := 'C'; size := 10; end;
    end;
    AStream.Write(typ, 1);  //11 Field Type
    AStream.Write(off, 4);  //12-15 offset of the field in the record
    Inc(off, size);
    bsize := byte(size);
    AStream.Write(bsize, 1); //16 Length of field (in bytes)
    bsize := byte(frac);
    AStream.Write(bsize, 1);       //17 Number of decimal places
    ch := char(0);
    if (typ = 'M') or (typ = 'P') then ch := char(4);
    AStream.Write(ch, 1); //18 Field Flags ($02Column can store null values, $04Binary column (for CHAR and MEMO only))
    aint := 0;
    AStream.Write(aint, 4); //1931 Reserved
    AStream.Write(aint, 4);
    AStream.Write(aint, 4);
    ch := char($00);
    AStream.Write(ch, 1);
  end;
  ch := char($0D);
  AStream.Write(ch, 1); //Header record terminator (0x0D)
  apos := AStream.Position;
  AStream.Position := 8;
  wpos := word(apos);
  AStream.Write(wpos, 2); //offset of the 1-st record with data
  wpos := word(off);
  AStream.Write(wpos, 2);  //the record's length
  AStream.Position := apos;     //returned to the begining of the records
  FreeMem(fn);
end;

procedure TVolgaDBFConverter.PrepareRow(AStream: TStream);
var s: string; OldSep, ch: char; i: integer;
  dd: word;

  //press text to right edge, adding spaces to length l
  function RightStr(s: string; l: integer): string;
  var i: integer;
  begin
    Result := Trim(s);
    i := Length(Result);
    if i < l then
      Result := StringOfChar(' ', l - i) + Result;
  end;
  //press text to left edge, adding spaces to length l
  function LeftStr(s: string; l: integer): string;
  var i: integer;
  begin
    Result := Trim(s);
    i := Length(Result);
    if i < l then
      Result := Result + StringOfChar(' ', l - i)
    else if i > l then
      Result := Copy(Result, 1, l);
  end;
  //convert to OEM
  function VStrToOem(const AnsiStr: string): string;
  begin
{$IfDef Linux}
    Result:=AnsiStr; {vn: OEM charset is not defined under linux}
{$Else}
    SetLength(Result, Length(AnsiStr));
    if Length(Result) > 0 then
      CharToOem(PChar(AnsiStr), PChar(Result));
{$EndIf Linux}
  end;

begin
  ch := ' ';
  AStream.Write(ch, 1); //space - an undeleted record token
  OldSep := DecimalSeparator;
  DecimalSeparator := '.';
  for i := 0 to FDataset.FieldCount - 1 do
  begin
    if FDataset.Fields[i].DataType in [ftFloat, ftCurrency] then
      s := FormatFloat('0.0000', FDataset.Fields[i].AsFloat)
    else if FDataset.Fields[i].DataType = ftDate then
      s := FormatDateTime('yyyymmdd', FDataset.Fields[i].AsDateTime)
    else if FDataset.Fields[i].DataType = ftTime then
      s := FormatDateTime('hh:nn:ss', FDataset.Fields[i].AsDateTime)
    else if FDataset.Fields[i].DataType = ftDateTime then
      s := FormatDateTime('dd.mm.yyyy hh:nn:ss', FDataset.Fields[i].AsDateTime)
    else begin
      s := FDataset.Fields[i].AsString;      //toOem?
      if FToOEM then s := VStrToOem(s);
    end;
    case FDataset.Fields[i].DataType of
    ftString: begin
      dd := word(FDataset.Fields[i].Size);
      s := LeftStr(s, dd);
      AStream.Write(PChar(s)^, dd); end;
    ftSmallint, ftWord: begin
      s := RightStr(s, 5);
      AStream.Write(PChar(s)^, 5); end;
    ftInteger, ftAutoInc: begin
      s := RightStr(s, 10);
      AStream.Write(PChar(s)^, 10); end;
    ftLargeInt: begin
      s := RightStr(s, 10);
      AStream.Write(PChar(s)^, 10); end;
    ftBoolean: begin
      if FDataset.Fields[i].AsBoolean then ch := 'T' else ch := 'F';
      AStream.Write(ch, 1); end;
    ftFloat, ftCurrency: begin
      s := RightStr(s, 16);
      AStream.Write(PChar(s)^, 16); end;
    ftDate, ftTime: AStream.Write(PChar(s)^, 8);
    ftDateTime: AStream.Write(PChar(s)^, 19);
//    ftMemo:  begin typ := 'M'; size := 10; end;
//    ftGraphic: begin typ := 'P'; size := 4; end;
//    ftFmtMemo: begin typ := 'M'; size := 4; end;
    else ;
    end;
  end;
  DecimalSeparator := OldSep;
end;

procedure TVolgaDBFConverter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FDataset) then
    FDataset := nil;
end;

procedure TVolgaDBFConverter.SetDataset(const Value: TDataset);
begin
  FDataset := Value;
end;

{$IfDef Linux}
function TVolgaDBFConverter.CodePageNum: char;
begin
  Result := char($00);
end;
{$Else}

function TVolgaDBFConverter.CodePageNum: char;
var cp: UINT;
begin
  if FToOEM then
  begin
    cp := GetOEMCP;
    case cp of
      437: Result := char($19); //dBASE ENG cp437
      737: Result := char($86); //dBASE ELL GR437
      850: Result := char($19); //dBASE ENG cp850
      852: Result := char($22); //dBASE HUN cp852
      857: Result := char($88); //dBASE TRK cp857
      860: Result := char($88); //dBASE PTG cp860
      862: Result := char($85); //Hebrew dBASE
      863: Result := char($1C); //dBASE FRC cp863
      865: Result := char($17); //dBASE NOR cp865
      866: Result := char($26); //dBASE RUS cp866
      867: Result := char($20); //dBASE CSY cp867
      868: Result := char($8E); //dBASE BUL 868
      874: Result := char($50); //dBASE THA cp874
      932: Result := char($13); //dBASE JPN cp932
      936: Result := char($4D); //dBASE CHS cp936
      949: Result := char($4E); //dBASE KOR cp949
      950: Result := char($4F); //dBASE CHT cp950
      else Result := char($00);  //no codepage
    end;
  end
  else
  begin
    cp := GetACP;
    case cp of
      437: Result := char($1B); //Paradox 'ascii'
      592: Result := char($65); //pdx ISO L_2 Czech
      737: Result := char($69); //Paradox Greek GR437
      850: Result := char($39); //Paradox 'intl' 850
      852: Result := char($68); //Paradox Hun 852 DC
      857: Result := char($6E); //Paradox 'turk'857
      861: Result := char($3B); //Paradox ISL 861
      862: Result := char($6B); //Paradox 'hebrew'
      865: Result := char($04); //Paradox 'nordan'
      866: Result := char($67); //Paradox Cyrr 866
      867: Result := char($7B); //Paradox Czech 867
      868: Result := char($71); //Paradox BUL 868
      874: Result := char($48); //Paradox Thai 874
      932: Result := char($03); //Paradox 'japan'
      936: Result := char($45); //Paradox China 936
      949: Result := char($46); //Paradox Korea 949
      950: Result := char($47); //Paradox Taiwan 950
      1250: Result := char($7F); //Pdox ANSI Hun. DC
      1251: Result := char($7E); //Pdox ANSI Cyrillic
      1252: Result := char($57); //'ascii' ANSI
      1253: Result := char($80); //Pdox ANSI Greek
      1254: Result := char($83); //Pdox ANSI Turkish
      1255: Result := char($75); //Paradox ANSI HEBREW
      1256: Result := char($74); //Borland ANSI Arabic
      else Result := char($00);  //no codepage
    end;
  end;
end;
{$EndIf Linux}

(* Codepages and their codes
  8 Description: SQL Link ROMAN8      $C9
437 Description: Paradox 'ascii'      $1B
437 Description: Paradox 'intl'       $02
437 Description: Paradox 'swedfin'    $06
437 Description: Paradox ESP 437      $3A
437 Description: dBASE NLD cp437      $09
437 Description: dBASE FIN cp437      $0B
437 Description: dBASE FRA cp437      $0D
437 Description: dBASE DEU cp437      $0F
437 Description: dBASE ITA cp437      $11
437 Description: dBASE SVE cp437      $15
437 Description: dBASE ESP cp437      $18
437 Description: dBASE ENG cp437      $19
437 Description: dBASE ENU cp437      $1B
437 Description: FoxPro German 437    $5E
437 Description: FoxPro Nordic 437    $5F
437 Description: Sybase SQL Dic437    $CC
592 Description: pdx ISO L_2 Czech    $65
737 Description: Paradox Greek GR437  $69
737 Description: dBASE ELL GR437      $86
850 Description: dBASE NLD cp850      $09
850 Description: dBASE FRA cp850      $0E
850 Description: dBASE DEU cp850      $0F
850 Description: dBASE ITA cp850      $11
850 Description: dBASE ESP cp850      $18
850 Description: dBASE SVE cp850      $15
850 Description: dBASE ENG cp850      $19
850 Description: dBASE FRC cp850      $1D
850 Description: dBASE PTB cp850      $25
850 Description: dBASE ENU cp850      $37
850 Description: Paradox 'intl' 850   $39
850 Description: FoxPro Nordic 850    $60
850 Description: Oracle SQL WE850     $CA
850 Description: Sybase SQL Dic850    $CB
852 Description: dBASE CSY cp852      $1F
852 Description: dBASE HUN cp852      $22
852 Description: dBASE PLK cp852      $23
852 Description: dBASE SLO cp852      $87
852 Description: Paradox Czech 852    $64
852 Description: Paradox Polish 852   $66
852 Description: Paradox Hun 852 DC   $68
852 Description: Paradox Slovene 852  $6D
852 Description: pdx Czech 852 'CH'   $79
857 Description: Paradox 'turk'       $6E
857 Description: dBASE TRK cp857      $88
860 Description: dBASE PTG cp860      $24
861 Description: Paradox ISL 861      $3B
862 Description: Paradox 'hebrew'     $6B
862 Description: Hebrew dBASE         $85
863 Description: dBASE FRC cp863      $1C
865 Description: Paradox 'nordan'     $04
865 Description: Paradox 'nordan40'   $05
865 Description: dBASE DAN cp865      $08
865 Description: dBASE NOR cp865      $17
866 Description: dBASE RUS cp866      $26
866 Description: Paradox Cyrr 866     $67
867 Description: dBASE CSY cp867      $20
867 Description: pdx Czech 867 'CH'   $7A
867 Description: Paradox Czech 867    $7B
868 Description: dBASE BUL 868        $8E
868 Description: Paradox BUL 868      $71
874 Description: Paradox Thai 874     $48
874 Description: dBASE THA cp874      $50
895 Description: FoxPro Czech DOS895  $9C
932 Description: Paradox 'japan'      $03
932 Description: dBASE JPN cp932      $13
932 Description: dBASE JPN Dic932     $56
932 Description: Access Japanese      $5D
932 Description: Pdox 'ascii' Japan   $1B
936 Description: Paradox China 936    $45
936 Description: dBASE CHS cp936      $4D
949 Description: Paradox Korea 949    $46
949 Description: dBASE KOR cp949      $4E
950 Description: dBASE CHT cp950      $4F
950 Description: Paradox Taiwan 950   $47
1250 Description: Pdox ANSI Czech     $78
1250 Description: pdx ANSI Czech 'CH' $7C
1250 Description: Pdox ANSI Polish    $7D
1250 Description: Pdox ANSI Hun. DC   $7F
1250 Description: Pdox ANSI Slovene   $82
1250 Description: pdx ANSI ISO L_2 CZ $92
1250 Description: FoxPro Czech 1250   $9B
1251 Description: Pdox ANSI Cyrillic  $7E
1251 Description: Pdox ANSI Bulgaria  $94
1252 Description: Borland DAN Latin-1 $27
1252 Description: Borland NLD Latin-1 $28
1252 Description: Borland FIN Latin-1 $29
1252 Description: Borland FRA Latin-1 $2A
1252 Description: Borland FRC Latin-1 $2B
1252 Description: Borland DEU Latin-1 $2C
1252 Description: Borland ISL Latin-1 $2D
1252 Description: Borland ITA Latin-1 $2E
1252 Description: Borland NOR Latin-1 $30
1252 Description: Borland ESP Latin-1 $31
1252 Description: Borland SVE Latin-1 $33
1252 Description: Borland ENG Latin-1 $34
1252 Description: Borland ENU Latin-1 $35
1252 Description: Borland PTG Latin-1 $36
1252 Description: Pdox ANSI Intl      $3C
1252 Description: Pdox ANSI Intl850   $3D
1252 Description: Pdox ANSI Spanish   $3E
1252 Description: Pdox ANSI Swedfin   $3F
1252 Description: Pdox ANSI Nordan4   $40
1252 Description: 'ascii' ANSI        $57
1252 Description: 'WEurope' ANSI      $58
1252 Description: 'Spanish' ANSI      $59
1252 Description: Access General      $5A
1252 Description: Access Nord/Danish  $5B
1252 Description: Access Swed/Finnish $5C
1252 Description: FoxPro German 1252  $61
1252 Description: FoxPro Nordic 1252  $62
1252 Description: DB2 SQL ANSI DEU    $CD
1253 Description: Pdox ANSI Greek     $80
1253 Description: Access Greece       $91
1253 Description: MSSQL ANSI Greek    $D1
1254 Description: Pdox ANSI Turkish   $83
1255 Description: Paradox ANSI HEBREW $75
1256 Description: Borland ANSI Arabic $74
*)

{$IfnDef Linux}
{ TVolgaXLSConverter }

constructor TVolgaXLSConverter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  CoInitialize(nil);
end;

destructor TVolgaXLSConverter.Destroy;
begin
  FDataset := nil;
  CoUninitialize;
  inherited Destroy;
end;

procedure TVolgaXLSConverter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FDataset) then
    FDataset := nil;
end;

procedure TVolgaXLSConverter.ConvertToCsv;
var A: TStrings;
begin
  if not FDataset.Active then FDataset.Open;
  A := TStringList.Create;
  A.Add(PrepareHeader);
  FDataset.First;
  while not FDataset.Eof do begin
    A.Add(PrepareRow);
    FDataSet.Next;
  end;
  A.SaveToFile(Fn);
  A.Free;
end;

function TVolgaXLSConverter.PrepareHeader: string;
var j: integer;
begin
  Result := '';
  for j := 0 to FDataset.FieldCount - 1 do begin
    if FDataset.Fields[j].Visible and not FDataset.Fields[j].IsBlob then
      Result := Result + AnsiQuotedStr(FDataset.Fields[j].DisplayLabel, '"') + #9;
  end;
  SetLength(Result, Length(Result) - 1);
end;

function TVolgaXLSConverter.PrepareRow: string;
var j: integer;
begin
  Result := '';
  for j := 0 to FDataset.FieldCount - 1 do begin
    if FDataset.Fields[j].Visible and not FDataset.Fields[j].IsBlob then
      Result := Result + AnsiQuotedStr(FDataset.Fields[j].AsString, '"') + #9;
  end;
  SetLength(Result, Length(Result) - 1);
end;

procedure TVolgaXLSConverter.SetDataset(const Value: TDataset);
begin
  FDataset := Value;
end;

procedure TVolgaXLSConverter.XLSToFile(AFileName: string);
const
  xlDelimited = $00000001;
  xlWindows = $00000002;
  xlDoubleQuote = $00000001;
  xlNormal = $FFFFEFD1;
  xlExcel9795 = 43;
var FExcel: Variant; FSheet: Variant;
begin
  try
    FExcel := CreateOleObject('Excel.Application');
  except
    DatabaseError('Cannot load Microsoft Excel', FDataset);
  end;
  if Assigned(FOnCallback) then FOnCallback(10);
  Fn := AFileName + '.txt';
  try
    ConvertToCsv;
    if Assigned(FOnCallback) then FOnCallback(55);
    try
      FExcel.Workbooks.OpenText(Fn, xlWindows, 1, xlDelimited,
        xlDoubleQuote, False, True, False, False ,False, False);
      if Assigned(FOnCallback) then FOnCallback(70);
      FSheet := FExcel.ActiveWorkbook.ActiveSheet;
      FSheet.Range['A1','A1'].Select;
      FExcel.ActiveCell.EntireRow.Font.Bold := true;
      FSheet.Cells.ShrinkToFit := true;
      FSheet.Columns.AutoFit;
      if Assigned(FOnCallback) then FOnCallback(85);
      FExcel.ActiveWorkbook.SaveAs(AFileName, xlExcel9795);
      if Assigned(FOnCallback) then FOnCallback(95);
      FExcel.ActiveWorkbook.Close(false);
    except
      DatabaseError('Error when exporting data to Microsoft Excel', FDataset);
    end;
  finally
    if FileExists(Fn) then DeleteFile(Fn);
    FSheet := Unassigned;
    FExcel.Quit;
    FExcel := Unassigned;
  end;
  if Assigned(FOnCallback) then FOnCallback(100);
end;
{$EndIf non-Linux}

end.
