//---------------------------------------------------------------------------
//  TVolgaDataset - inherited from TDataSet, used for flat files and in-memory tables
//  It inherits and publishes all properties, methods and events from TDataset,
//  that are inherited and published in standard TTable
//---------------------------------------------------------------------------
//  Copyright  2000-2004, Olga Vlasova, Russia
//  http://www.volgadb.com
//  E-mail: info@volgadb.com
//---------------------------------------------------------------------------
//  Support Kylix/Linux platform: (c) Vit (Vitaly Nevzorov), 2004, Chicago, USA
//  http://www.delphist.com
//  E-mail: vit@vingrad.ru
//---------------------------------------------------------------------------
unit VolgaTbl;
{ $DEFINE VOLGAT}
{$DEFINE PRO}

interface

uses Classes, DB, SysUtils,             {vn: Db -> DB Linux names are case sensitive}
{$IFDEF Linux}
  QForms, QControls, QDialogs,
{$ELSE}
  Windows, Forms, Controls, Dialogs,
{$ENDIF Linux}
{$IFDEF VER140}Variants, {$ENDIF}
{$IFDEF VER150}Variants, {$ENDIF}
{$IFDEF VER180}Variants, {$ENDIF}
{$IFDEF PRO}
  VolgaAlg, VolgaLex, VolgaExp,
{$IFDEF Linux}
  QStdCtrls, QConsts,
{$ELSE}
  StdCtrls, ComObj, ActiveX, Consts,
{$ENDIF Linux}
{$ENDIF}
{$IFDEF INTLIB}VolgaMs, VolMsSQLTest, {$ENDIF}
  VolgaConst;

const MAXSLEN: integer = 35;            //max length of vftSmallStr field (9 digits + 26 letters)

type
  PWordBool = ^WordBool;

  TVolgaUpdStatus = (vusUnmodified, vusModified, vusInserted, vusDeleted, vusAbsent);
  TVolgaLoadMode = (lmClear, lmFile, lmClone, lmStream);
  TVolgaFieldType = (vftUnknown, vftSmallInt, vftInteger, vftLargeInt, vftBoolean, vftDouble,
    vftCurrency, vftDate, vftTime, vftDateTime, vftAutoInc, vftSmallStr, vftString,
    vftBlob, vftMemo, vftFmtMemo, vftGraphic);
  TVolgaBatchMode = (vbmAppend, vbmAppendUpdate, vbmCopy, vbmCopyDef);
{$IFDEF PRO}
  TVolgaGroupKind = (vgkSum, vgkMin, vgkMax, vgkAvg, vgkCount);
{$ENDIF}
  TVolgaSortOptions = set of (vsoCaseInsensitive, vsoDescendSort, vsoInsertInOrder);
  TVolgaSaveOptions = set of (savAutoBackup, savSlowMode, savSortOnIndex);

const StrFieldType: array[TVolgaFieldType] of string = ('Unknown', 'SmallInt',
    'Integer', 'Int64', 'Boolean', 'Double', 'Currency', 'Date', 'Time', 'DateTime',
    'AutoInc', 'SmallStr', 'String', 'BLOB', 'Memo', 'FmtMemo', 'Graphic');

type
  TVolgaRecord = record                 //datastore record
    Data: PChar;                        //current data
    OldData: PChar;                     //old data before changings (for OldValue)
    RecordData: Pointer;                //pointer for arbitrary data for this record
    Status: TVolgaUpdStatus;            //status: unmodified/modified/inserted/deleted
    Size: integer;                      //size of whole record when it will be stored in file
  end;
  PVolgaRecord = ^TVolgaRecord;

  TVolgaField = record                  //one field description
    Name: string;                       //field name
    Origin: string;                     //source table name
    DataSize: Word;                     //length of field data or pointer length
    Size: Word;                         //declared field size
    Offset: integer;                    //offset from start of record
    KeyValue: string;                   //for searching needs
    DataType: TVolgaFieldType;          //Volga type of field
    Align: TAlignment;                  //alignment from TField
    DefValue: string;                   //default value
    Description: string;                //description of field
  end;
  PVolgaField = ^TVolgaField;

  //------------------ structure for keeping key data for search methods -------
  TKeyIndex = (kiIndex, kiSearch, kiRangeStart, kiRangeEnd);

  PKeyBuffer = ^TKeyBuffer;
  TKeyBuffer = packed record
    Data: PChar;
    Kind: TKeyIndex;
    KeyFI: TList;                       //list of VolgaCursor fields
    Fields: TList;                      //list of Dataset fields
    Exclusive: Boolean;
    FieldCount: Integer;
  end;
  //------------------- end of key structure -----------------------------------

  TVolgaDatabase = class;
  TVolgaDataset = class;
  TVolgaCursor = class;

  TCursorFilterEvent = procedure(Sender: TVolgaCursor; pRec: PVolgaRecord; var Accept:
    Boolean) of object;
{$IFDEF PRO}
  TPasswordEvent = function(var NewPassw: string): Boolean of object;
  TInvalidPasswordEvent = procedure(WrongPassw: string; var TryAgain: Boolean) of object;
  TDestroyRecDataEvent = procedure(Sender: TVolgaDataset; var RecordData: Pointer) of
    object;
  TDestroyCursorRecEvent = procedure(var RecordData: Pointer) of object;
{$ENDIF}
  TCallbackEvent = procedure(Sender: TVolgaDataset; const Percent: integer) of object;
  TCursorCallbackEvent = procedure(const Percent: integer) of object;

//********* TBlobStream - for BLOB fields implementation (memo,graphic...) *****

  TVolgaBlobStream = class(TMemoryStream)
  private
    FField: TBlobField;
    FDataSet: TVolgaDataset;
    FFieldNo: integer;
    FModified: Boolean;
    procedure ReadBlobData;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    procedure Truncate;
  end;

{ TVolgaMasterLink }

  TVolgaMasterLink = class(TDetailDataLink)
  private
    FDataSet: TVolgaDataSet;
    FFieldNames: string;
    FFields: TList;
    FOnMasterChange: TNotifyEvent;
    FOnMasterDisable: TNotifyEvent;
    procedure SetFieldNames(const Value: string);
  protected
    procedure ActiveChanged; override;
    procedure CheckBrowseMode; override;
    function GetDetailDataSet: TDataSet; override;
    procedure LayoutChanged; override;
    procedure RecordChanged(Field: TField); override;
  public
    constructor Create(DataSet: TVolgaDataSet);
    destructor Destroy; override;
    property FieldNames: string read FFieldNames write SetFieldNames;
    property Fields: TList read FFields;
    property OnMasterChange: TNotifyEvent read FOnMasterChange write FOnMasterChange;
    property OnMasterDisable: TNotifyEvent read FOnMasterDisable write FOnMasterDisable;
  end;

//********* TVolgaCursor - data storage ****************************************

  TVolgaCursor = class(TObject)
  private
    FData: TList;                       //list of all data records
    FFldInfo: TList;                    //list of fieldinfo records
    FFieldLayout: PChar;                //buffer for storing fieldinfo
    FFldLen: integer;                   //length of field layout info buffer
    FFldCount: integer;                 //field count in record
    FRecSize: integer;                  //record size
    FPartialOpen: Boolean;              //flag - open cursor regardless of any error occured
    FFilterClause: TCursorFilterEvent;
{$IFDEF PRO}
    FIsFiltered: Boolean;
    FPassword: string;
    FPasswBuf: PChar;
    FCiphered: Boolean;
    FLastAutoInc: integer;
    FAutoIncName: string;
    FAutoIncField: PVolgaField;
    FOnDestroyRec: TDestroyCursorRecEvent;
{$ENDIF}
    FNotes: string;                     //for Table notes
    FOnCallback: TCursorCallbackEvent;
    function AddNewRecord(Stream: TStream; var Msg: string): Boolean;
    function InitFieldDefs(var Msg: string): Boolean;
    function GetRecordCount: integer;
    function GetFieldCount: integer;
    function GetFieldInfo(Index: integer): PVolgaField;
    function GetData(Index: integer): PVolgaRecord;
    function LoadStruFromStream(Stream: TStream; var Msg: string;
      var rec_count: integer): Boolean;
    procedure CalcRecSize(pRec: PVolgaRecord);
    function CalcStreamSize: integer;
    procedure FillFieldLayout;
  public
    constructor Create;
    destructor Destroy; override;
    procedure CancelUpdates;
    procedure Clear(WithFields: Boolean);
    procedure ClearBuf(RecBuf: PChar);
    procedure ClearData(var RecBuf: PChar);
    procedure CommitUpdates;
    procedure CopyRecordBuf(SourceBuf, DestBuf: PChar);
    function CreateUndoCopy(pRec: PVolgaRecord): Boolean;
    function DeleteRecord(CurRec: PVolgaRecord; var Msg: string): Boolean;
    function FindFieldInfo(const Name: string): PVolgaField;
    function GetFieldText(FI: PVolgaField; P: PVolgaRecord; IsOld: Boolean; YearFirst:
      Boolean): string;
    function LoadFieldInfo(Stream: TStream; var Msg: string): Boolean;
    function LoadFromStream(Stream: TStream; var Msg: string): Boolean;
    function SaveRecord(Buf: PChar; CurRec: PVolgaRecord; var Msg: string): Boolean;
    function SaveToStream(AStream: TStream): integer;
    procedure SaveToFile(const Fname: string; Options: TVolgaSaveOptions;
      const FieldNames: string; const IgnoreCase: Boolean);
    procedure Sort(const FieldNames: string; const IgnoreCase: Boolean);
    procedure UndoRecord(pRec: PVolgaRecord);
{$IFDEF PRO}
    function CheckDuplicate(Buf: PChar; NonCheckedRec: PVolgaRecord; const Fn: string):
    boolean;
    procedure SetLastAutoInc(NewValue: integer);
{$ENDIF}
    property FieldCount: integer read GetFieldCount;
    property FieldInfo[Index: integer]: PVolgaField read GetFieldInfo;
    property FilterClause: TCursorFilterEvent read FFilterClause write FFilterClause;
    property PartialOpen: Boolean read FPartialOpen write FPartialOpen;
{$IFDEF PRO}
    property Password: string read FPassword write FPassword;
    property OnDestroyRec: TDestroyCursorRecEvent read FOnDestroyRec write FOnDestroyRec;
    property Notes: string read FNotes write FNotes;
{$ENDIF}
    property RecData[Index: integer]: PVolgaRecord read GetData;
    property RecordCount: integer read GetRecordCount;
    property RecSize: integer read FRecSize;
    property OnCallback: TCursorCallbackEvent read FOnCallback write FOnCallback;
  end;

//********* TVolgaDataSet - abstract Volga dataset *****************************

  TVolgaDataset = class(TDataSet)
  private
    FRecordSize: Word;                  //dataset record size without service info
    FRecBufSize: Word;                  //whole record size with service info
    FCurRec: integer;                   //number of current logical data record
    FSortList: TList;                   //list of records in logical order (sorted or ranged or filtered)
    FOrigin: string;                    //source table name without path and extension
    FNotes: string;                     //arbitrary notes stored within Volga file
    FReadOnly: Boolean;                 //flag of readonly dataset
    FRecReadOnly: Boolean;              //flag of readonly record
    FIndexFieldNames: string;           //list of index fields
    FRangeFieldNames: string;           //list of range fields
    FUniqueFieldNames: string;          //list of unique key fields
    FLoadedBuf: Boolean;                //flag of completely data loading
    FIsRanged: Boolean;                 //flag of ranged dataset
    FPartialOpen: Boolean;              //flag - open dataset regardless of any error occured
    FFiltRec: PVolgaRecord;             //current record for applying filter conditions
    FCloneList: TList;                  //list of cloned datasets
{$IFDEF PRO}
    FPassword: string;
    FCheckViol: Boolean;
    FOnDestroy: TDestroyRecDataEvent;
    FOnFindRecord: TFilterRecordEvent;
    FShowDeleted: Boolean;
{$ENDIF}
    FMasterLink: TVolgaMasterLink;
    FOnNeed: TNotifyEvent;
    FLoadTime: Cardinal;
    FLoadSize: integer;
    FOnCallback: TCallbackEvent;
    FSortOptions: TVolgaSortOptions;
    FSaveOptions: TVolgaSaveOptions;
    FDatabase: TVolgaDatabase;
    FIsSorted: Boolean;                 //flag of sorted dataset
    FExternalStream: TStream;           //external stream for loading
    FKeyBuffers: array[TKeyIndex] of PKeyBuffer; //structure for keep key info
    FKeyBuffer: PKeyBuffer;             //current search buffer (for dsSetKey state)

    procedure AllocKeyBuffers;
    procedure FreeKeyBuffers;
    function InitKeyBuffer(KeyIndex: TKeyIndex; FieldNames: string = ''): PKeyBuffer;
    procedure SetKeyFields(KeyIndex: TKeyIndex; const Values: array of const); overload;
    procedure SetKeyFields(KeyIndex: TKeyIndex; const Values: string); overload;
    function GetKeyFieldCount: Integer;
    procedure SetKeyFieldCount(Value: Integer);
    procedure CheckSetKeyMode;
    procedure SetKeyBuffer(KeyIndex: TKeyIndex; Clear: Boolean);
    function SetCursorRange: Boolean;
    function ResetCursorRange: Boolean;
    function GetKeyExclusive: Boolean;
    procedure SetKeyExclusive(Value: Boolean);
    function GetIndexFieldCount: Integer;
    function GetRangeFieldCount: Integer;
    function GetIndexField(Index: Integer): TField;
    procedure SetIndexField(Index: Integer; Value: TField);
    procedure SetReadOnly(value: Boolean);
    function GetFieldOffset(Num: integer): integer;
    procedure ModifyRec(Buf: PChar);
    procedure InsertRec(Buf: PChar);
    procedure SetIndexFieldNames(const value: string);
    function AcceptRange(pRec: PVolgaRecord): Boolean;
    function InternalLocate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions; const FromStart: Boolean = true): Boolean;
    function GetActiveRecBuf(var RecBuf: PChar): Boolean;
    function GetNeedToSave: Boolean;
    function InternalFind(const istart: integer; Sorted: Boolean): integer;
    procedure SetDataSource(Value: TDataSource);
    function GetMasterFields: string;
    procedure SetMasterFields(const Value: string);
    procedure SetRangeFieldNames(const Value: string);
    procedure DoCallback(const Percent: integer);
    procedure SetSortOptions(const Value: TVolgaSortOptions);
    function FastSeek(const Fn: string): Boolean; //check if we can use fast searching
{$IFDEF PRO}
    procedure SetShowDeleted(const Value: Boolean);
    function InternalFindNearest(const istart: integer): integer;
    procedure DoDestroyRec(var RecordData: Pointer);
    procedure CheckFilterFields;
    function GetPassword: string;
    procedure SetPassword(const Value: string);
    function GetRecordData: Pointer;
    procedure SetRecordData(const Value: Pointer);
{$ENDIF}
    function GetCloneCount: integer;
    procedure CheckCursorOpen;
    function GetVersion: string;
    procedure SetVersion(const Value: string);
    function GetStreamSize: integer;
    procedure MoveInsertedToOrder;

  protected
    FNeedToSave: integer;               //flag - if any changed records are not saved yet
    FCursor: TVolgaCursor;              //internal data storage cursor
{$IFDEF PRO}
    FFilterFI: TList;                   //list of filter pairs
    AcceptFilterList: TList;
{$ENDIF}
    procedure SetOrigin(const Value: string); virtual;
    procedure SetDatabase(const Value: TVolgaDatabase); virtual;
    procedure AddClone(DS: TVolgaDataset);
    procedure RemoveClone(DS: TVolgaDataset);
    procedure InternalLoadData; virtual; abstract;
    procedure InternalApplyUpdates; virtual; abstract;
    procedure DestroyCursor; virtual; abstract;
    procedure InternalLoad(Stream: TStream; Keep: Boolean = false);
{$IFDEF PRO}
    procedure AcceptFilter(Sender: TVolgaCursor; pRec: PVolgaRecord;
      var Accept: Boolean);             //check filter conditions for filter expression or WHERE in SELECT
    procedure ClearFilterExpr;
    function FindRecord(Restart, GoForward: Boolean): Boolean; override;
{$ENDIF}
    //virtual protected abstract methods from standard TDataSet
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure ClearCalcFields(Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
      override;
    function GetRecordSize: Word; override;
    function GetDataSource: TDataSource; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalRefresh; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    //virtual protected non-abstract methods from standard TDataSet
    function GetCanModify: Boolean; override;
    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    procedure SetFiltered(Value: Boolean); override;
    procedure DoOnNewRecord; override;
    procedure DoBeforeDelete; override;
    procedure DoBeforeInsert; override;
    procedure DoBeforeEdit; override;
    procedure InternalCancel; override;
{$IFDEF PRO}
    procedure SetFilterText(const Value: string); override;
    property Password: string read GetPassword write SetPassword;
    property CheckKeyViolation: Boolean read FCheckViol write FCheckViol default false;
{$ENDIF}
    procedure CheckMasterRange;
    procedure MasterChanged(Sender: TObject);
    procedure MasterDisabled(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;

    //virtual public non-abstract methods from standard TDataSet
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; overload; override;
    function GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean; overload;
      override;
    function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant; override;
    function UpdateStatus: TUpdateStatus; override;

    //additional public methods
    procedure ApplyUpdates; virtual;
    procedure CancelUpdates; virtual;
    procedure CommitUpdates; virtual;
    procedure CopyRecord;               //copy current record to new one using OnNewRecords rules
    procedure EmptyTable;               //clear internal cursor without closing dataset
    function ExistField(const Fn: string): Boolean; //is this field exists?
    function GetDescription(const FieldName: string): string;  //description of a field
    procedure SetDescription(const FieldName: string; const descr: string);  //set description of a field
    function GetDefaultValue(const FieldName: string): string;  //default value of a field
    procedure SetDefaultValue(const FieldName: string; const value: string);  //set default value of a field
{$IFDEF PRO}
    function CheckDuplicate(const FieldName: string = ''): Boolean; overload; //if there exist record with same field values
    procedure ExportToTxt(const FileName: string; const Delimiter: char = #9); //export to text file with delimiters
    procedure ExportToXML(const FileName, Encoding: string); //export to XML
    procedure ExportToDBF(const FileName: string; ToOEM: Boolean = false); //export to DBF
    procedure ExportToExcel(const FileName: string); //export to Excel
    function SaveToStream(AStream: TStream): integer;
    procedure SetOriginToFields; virtual;
    procedure RevertRecord;             //rollback deleting or changing a record
    function IsChanged(const FieldName: string): Boolean; //is this field changed? (only before CommitUpdates)
    procedure FindNearest(const KeyValues: array of const);
    procedure GotoNearest;
    function IsPasswordProtected: Boolean;
    function GetListRecValues(const FieldName: string; const ch: char): string; //list of field values, satisfying filter condition
    property RecordData: Pointer read GetRecordData write SetRecordData; //pointer to arbitrary data
    property ShowDeleted: Boolean read FShowDeleted write SetShowDeleted; //flag - show deleted records
{$ENDIF}
    function GetSum(const FieldName: string): Extended; //get a sum of field values from all visible records
    function Find(const KeyFields: string; const KeyValues: array of const;
      const FromFirstRec: Boolean = true): Boolean; {overload;} //find using array
    function FindTab(const KeyFields, KeyValues: string;
      const FromFirstRec: Boolean = true): Boolean; //overload; //find using string with tabs
    function GetTabValues(const KeyFields: string): string; //get a string of field values separationg with tabs
    procedure GotoCurrent(DS: TVolgaDataset); //synchronize two datasets
    function Like(const KeyFields: string; const KeyValues: string; AtStartPos: Boolean = false;
      FromFirstRec: Boolean = true): Boolean; //context searching
    procedure SaveToFile(const FileName: string);
    procedure SetRange(const StartValues, EndValues: array of const); overload; //using array
    procedure SetRange(const StartValues, EndValues: string); overload; //using string with tabs
    procedure SortRecords(ToFirst: Boolean = false); //force sorting on IndexFieldNames
    procedure SetListOfFields(const KeyFields: string; const KeyValues: array of const);
    function FindKey(const KeyValues: array of const): Boolean;
    procedure EditKey;
    function GotoKey: Boolean;
    procedure SetKey;
    procedure ApplyRange;
    procedure CancelRange;
    procedure EditRangeEnd;
    procedure EditRangeStart;
    procedure SetRangeEnd;
    procedure SetRangeStart;

    property KeyFieldCount: integer read GetKeyFieldCount write SetKeyFieldCount;
    property KeyExclusive: Boolean read GetKeyExclusive write SetKeyExclusive;
    property IndexFieldCount: Integer read GetIndexFieldCount;
    property RangeFieldCount: Integer read GetRangeFieldCount;
    property IndexFields[Index: Integer]: TField read GetIndexField write SetIndexField;
    property CloneCount: integer read GetCloneCount; //number of cloned datasets
    property IsRanged: Boolean read FIsRanged; //is Range active?
    property LoadSize: integer read FLoadSize; //loading buffer size
    property LoadTime: Cardinal read FLoadTime; //loading buffer time
    property NeedToSave: Boolean read GetNeedToSave; //is where any changes (before ApplyUpdates)
    property Origin: string read FOrigin write SetOrigin;
    property RecReadOnly: Boolean read FRecReadOnly write FRecReadOnly; //readonly record flag
    property StreamSize: integer read GetStreamSize;

    //Public properties derived from TDataset
    property Bof;
    property Bookmark;
    property DefaultFields;
    property Designer;
    property Eof;
    property FieldCount;
    property Fields;
    property FieldValues;
    property Found;
    property Modified;
    property RecordCount;
    property State;

  published
    property Database: TVolgaDatabase read FDatabase write SetDatabase;
    property IndexFieldNames: string read FIndexFieldNames write SetIndexFieldNames;
    property MasterFields: string read GetMasterFields write SetMasterFields;
    property MasterSource: TDataSource read GetDataSource write SetDataSource;
    property Notes: string read FNotes write FNotes;
    property PartialOpen: Boolean read FPartialOpen write FPartialOpen default false;
    property RangeFieldNames: string read FRangeFieldNames write SetRangeFieldNames;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly default false;
    property SaveOptions: TVolgaSaveOptions read FSaveOptions write FSaveOptions default
    [];
    property SortOptions: TVolgaSortOptions read FSortOptions write SetSortOptions default
    [];
    property UniqueFieldNames: string read FUniqueFieldNames write FUniqueFieldNames;
    property Version: string read GetVersion write SetVersion;

    //Published out of TDataset
    property Active;
    property AutoCalcFields;
{$IFDEF PRO}
    property Filter;
{$ENDIF}
    property Filtered;
    property AfterCancel;
    property AfterClose;
    property AfterDelete;
    property AfterEdit;
    property AfterInsert;
    property AfterOpen;
    property AfterPost;
    property AfterScroll;
    property BeforeCancel;
    property BeforeClose;
    property BeforeDelete;
    property BeforeEdit;
    property BeforeInsert;
    property BeforeOpen;
    property BeforePost;
    property BeforeScroll;
    property OnCalcFields;
    property OnFilterRecord;
    property OnNewRecord;
    property OnNeedToSaveChange: TNotifyEvent read FOnNeed write FOnNeed;
    property OnCallback: TCallbackEvent read FOnCallback write FOnCallback;
{$IFDEF PRO}
    property OnFindRecord: TFilterRecordEvent read FOnFindRecord write FOnFindRecord;
    property OnDestroyRecordData: TDestroyRecDataEvent read FOnDestroy write FOnDestroy;
{$ENDIF}
  end;

//********* TVolgaTable - table for flat Volga files ***************************

  TVolgaTable = class(TVolgaDataSet)
  private
    FTableName: string;
    FLoadMode: TVolgaLoadMode;
    FStoreDefs: Boolean;
    FCloneDS: TVolgaDataset;
    procedure SetTableName(const value: string);
    procedure SetLoadMode(value: TVolgaLoadMode);
    procedure SetCloneDS(value: TVolgaDataset);
    function FieldDefsStored: Boolean;
    procedure InternalLoadByClearMode;  //load from FieldDefs
    procedure InternalLoadFromFile;     //load from file
    procedure InternalLoadFromCursor;   //load from another dataset (clone)
    function PrepareLayoutStr(FD: TFieldDefs): string;
    function GetFullTableName: string;
{$IFDEF PRO}
    function GetTableNotes: string;
    procedure SetTableNotes(const Value: string);
{$ENDIF}
  protected
    procedure SetDatabase(const Value: TVolgaDatabase); override;
    procedure DefChanged(Sender: TObject); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure InternalLoadData; override;
    procedure InternalApplyUpdates; override;
    procedure InternalClose; override;
    procedure DestroyCursor; override;
  public
    constructor Create(AOwner: TComponent); override;
    function BatchMove(ASource: TDataSet; AMode: TVolgaBatchMode = vbmAppend): integer;
    procedure CancelUpdates; override;
    procedure CommitUpdates; override;
    procedure CreateTable;
    procedure DeleteTable;
    procedure LoadFromFile(const Fn: string);
    procedure LoadFromCursor(CloneDB: TVolgaDataset);
    procedure RenameTable(const NewTableName: string);
{$IFDEF PRO}
    procedure SetLastAutoInc(NewValue: integer);
    procedure LoadFromStream(AStream: TStream); //load from any stream
    procedure ImportFromTxt(const FileName: string; const Delimiter: char = #9); //import into opened table
    function GroupBy(ASource: TVolgaDataset; const GroupFlds, SumFlds: string;
      const Kind: TVolgaGroupKind = vgkSum): Boolean;
    property TableNotes: string read GetTableNotes write SetTableNotes;
{$ENDIF}
  published
    property CloneDS: TVolgaDataset read FCloneDS write SetCloneDS;
    property TableName: string read FTableName write SetTableName;
    property LoadMode: TVolgaLoadMode read FLoadMode write SetLoadMode default lmClear;
    property FieldDefs stored FieldDefsStored;
    property StoreDefs: Boolean read FStoreDefs write FStoreDefs default false;
{$IFDEF PRO}
    property CheckKeyViolation;
    property Password;
{$ENDIF}
  end;

{$IFDEF PRO}
//********* TVolgaCustomQuery **************************************************

  TVolgaCustomQuery = class(TVolgaDataSet)
  private
    FSQL: TStrings;
    FText: string;
    FParams: TParams;
    function GetParamsCount: Word;
    function GetSQL: TStrings;
    procedure SetParamsList(const Value: TParams);
    procedure SetSQL(const Value: TStrings);
    procedure QueryChanged(Sender: TObject);
  protected
    FPrepared: Boolean;
    procedure DestroyCursor; override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure InternalClose; override;
    procedure InternalRefresh; override;
    procedure ReadParamData(Reader: TReader);
    procedure WriteParamData(Writer: TWriter);
    function GetPreparedSQL(DecSep, DatSep: char; DatFmt: string): string;
    procedure Prepare; virtual; abstract;
    procedure UnPrepare; virtual; abstract;
  public
    constructor Create(AOwner: TComponent); override;
    function ParamByName(const Value: string): TParam;
    procedure SetOriginToFields; override;
    property Prepared: Boolean read FPrepared;
    property ParamCount: Word read GetParamsCount;
  published
    property CheckKeyViolation;
    property SQL: TStrings read GetSQL write SetSQL;
    property Params: TParams read FParams write SetParamsList stored False;
    property Origin;
  end;
{$ENDIF}

//********* TVolgaDatabase *****************************************************

  TVolgaDatabase = class(TComponent)
  private
    FDSList: TList;
    FUpdated: TList;
    FDatabasePath: string;
    FDateQuot: char;
    FDateFormat: string;
    function GetDataSet(Index: integer): TVolgaDataset;
    function GetDataSetCount: integer;
    function GetUpdatedCount: integer;
    function GetUpdatedDataSet(Index: integer): TVolgaDataset;
    procedure SetDatabasePath(const Value: string);
    procedure Close;
  protected
    procedure Add(DS: TVolgaDataset);
    procedure Remove(DS: TVolgaDataset);
    procedure AddUpdated(DS: TVolgaDataset);
    procedure DeleteUpdated(DS: TVolgaDataset);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ApplyUpdates(const ADataSets: array of TVolgaDataset);
    procedure CancelUpdates(const ADataSets: array of TVolgaDataset);
    property DataSetCount: integer read GetDataSetCount;
    property DataSets[Index: integer]: TVolgaDataset read GetDataSet;
    property UpdatedCount: integer read GetUpdatedCount;
    property UpdatedDataSets[Index: integer]: TVolgaDataset read GetUpdatedDataSet;
{$IFDEF PRO}
    property DateQuot: char read FDateQuot; //for DAO - #, else "
    procedure Restructure(const Fn: string; AList: TStrings);
{$ENDIF}
  published
{$IFDEF PRO}
    property DateFormat: string read FDateFormat write FDateFormat;
{$ENDIF}
    property DatabasePath: string read FDatabasePath write SetDatabasePath;
  end;

//********* TVolgaSession ******************************************************
  TVolgaSession = class(TObject)
  private
    FDatabases: TList;
{$IFDEF PRO}
    FPasswords: TStringList;            //list of all entered passwords
    FOnPassword: TPasswordEvent;
    FOnInvalidPassword: TInvalidPasswordEvent;
    FTryCount: integer;
    function GetPassword(var NewPassw: string): Boolean; //for Volga tables
{$ENDIF}
    function GetDatabase(Index: Integer): TVolgaDatabase;
    function GetDatabaseCount: Integer;
  protected
{$IFDEF PRO}
    Sapphire: TSapphire;                //Sapphire engine for ciphering
    function CheckPassw(Original: PChar; var newpasw: string): Boolean;
{$ENDIF}
    procedure Add(Database: TVolgaDatabase);
    procedure Delete(Database: TVolgaDatabase);
  public
    constructor Create;
    destructor Destroy; override;
{$IFDEF PRO}
    procedure AddPassword(const passw: string); //add password for session
    procedure DeletePassword(const passw: string); //remove password from session
    procedure ClearPasswords;                      //remove all passwords from session
    function CheckStreamHeader(AStream: TStream): Boolean; //check if stream has a correct Volga header
    property OnPassword: TPasswordEvent read FOnPassword write FOnPassword;
    property OnInvalidPassword: TInvalidPasswordEvent read FOnInvalidPassword write FOnInvalidPassword;
    property TryCount: integer read FTryCount write FTryCount;
{$ENDIF}
    property DatabaseCount: Integer read GetDatabaseCount;
    property Databases[Index: Integer]: TVolgaDatabase read GetDatabase;
    procedure GetTableNames(const DatabasePath, Pattern: string; Extensions: Boolean;
      List: TStrings); overload;
    procedure GetTableNames(const Database: TVolgaDatabase; Pattern: string;
      Extensions: Boolean; List: TStrings); overload;
  end;

const SIZEINT: integer = sizeof(integer);
const SIZEINT64: integer = sizeof(int64);
const SIZEDOUB: integer = sizeof(double);
const SIZECHAR: integer = sizeof(PChar);
const SIZEBOOL: integer = sizeof(WordBool);
const FieldSeparator: char = Chr($12);  //field separator in stream
const RecordSeparator: char = Chr($14); //record separator in stream
const Tabulator: char = Chr($09);
const Quote2: char = Chr($22);          //quotation mark
const Quote1: char = Chr($27);          //apostrophe
const RETURN: string = #13#10;

//****************************************************************************
var VolgaDB: TVolgaSession;             //global variable of current session
//****************************************************************************

function GetUnicod: string;     //unique codegenerator 8 bytes code
function GetUnicod10: string;   //unique codegenerator 10 bytes code
//overloaded field type functions
function GetFldType(ft: TFieldType; size: integer): TVolgaFieldType; overload;
function GetFldType(ft: TVolgaFieldType): string; overload;
function GetFldType(ft: string): TVolgaFieldType; overload;
function GetFldType2(ft: string): TFieldType;
function GetCharType(ft: TVolgaFieldType; ASize: integer): char;
//functions like C++   expr ? a1 : a2;
function _if(b: Boolean; a1, a2: integer): integer; overload;
function _if(b: Boolean; a1, a2: string): string; overload;
//divide a string into string array using delimiter
procedure Split(const s: string; const delim: string; SplitArray: TStrings);
//convert all spaces to underline symbol in string
function vSpaceToChar(const s: string; p: char = '_'): string;
//functions for working with buffer
function IsBufEmpty(Buffer: PChar; size: integer): Boolean;
procedure KillBuf(Buffer: Pointer);
procedure CopyLongString(BufSource, BufDest: PChar);
procedure CopyBlobField(BufSource, BufDest: PChar);
function CompareBuf(const key1, key2: PChar; FI: PVolgaField; IgnoreCase: Boolean):
integer;
function CompareRec(Buf1, Buf2: PChar; ListFI: TList; Count: integer = -1): integer;
procedure AssignBufFromStr(Buf: PChar; const Value: string; FI: PVolgaField);
procedure AssignBufFromVar(Buf: PChar; const Value: TVarRec; FI: PVolgaField);
//get TVolgaField list using string of field names
procedure GetFldInfoList(Sender: TVolgaCursor; List: TList; const FieldNames: string);
function GetCursor(Dataset: TVolgaDataset): TVolgaCursor;

implementation

{$IFDEF PRO}
const LenPassw: integer = 40;           //sizeof(TCLSID) + 2;
{$ENDIF}

{$IFDEF VOLGAT}
const FVolgaCounter: integer = 0;
const texta: string =
  'You are using unregistered version of VolgaDB Engine.'#13#13'Please register at http://www.volgadb.com/ !';
const textb: string =
  'This feature is unavailable in unregistered version of VolgaDB Engine.'#13#13'Please register at http://www.volgadb.com/ !';
const textc: string = 'VolgaDB Engine trial';
{$ENDIF}
const _version: string = '5.10';
const UsingFieldType: array[TVolgaFieldType] of TFieldType = (ftUnknown, ftSmallInt,
    ftInteger, ftLargeInt, ftBoolean, ftFloat, ftCurrency, ftDate, ftTime, ftDateTime,
    ftAutoInc, ftString, ftString, ftBlob, ftMemo, ftFmtMemo, ftGraphic);

type
  TVolgaRecInfo = record                //this structure is added to buffer of each record when TDataSet is asking for it
    Bookmark: Pointer;                  //current record pointer - a bookmark
    BookmarkFlag: TBookmarkFlag;        //enum TBookmarkFlag {bfCurrent,bfBOF,bfEOF,bfInserted}
    RecordNum: integer;                 //record number in FSortList + 1
  end;
  PVolgaRecInfo = ^TVolgaRecInfo;

  TVolgaBookmark = record               //used for access to bookmark
    pToRec: Pointer;                    //current record pointer - a bookmark
  end;
  PVolgaBookmark = ^TVolgaBookmark;

var FSortFI: TList;                     //fields list for sorting
  IgnoreC: Boolean = false;
  DescSort: Boolean = false;            //decrease sorting
{$IFDEF PRO}
  crcl, crch: byte;
{$ENDIF}

//========= sort function for TList (list of records) ==========================

function VolgaSort(Item1, Item2: Pointer): integer;
//variable FSortFI is defined before sorting is started
var FI: PVolgaField;
  j: integer;
begin
  Result := 0;
  j := 0;
  while (Result = 0) and (j < FSortFI.Count) do
  begin
    FI := PVolgaField(FSortFI.Items[j]); //field for comparing values
    if (PVolgaRecord(Item1)^.Status <> vusDeleted) and
      (PVolgaRecord(Item1)^.Status <> vusAbsent) then
      Result := CompareBuf(PVolgaRecord(Item1)^.Data + FI^.Offset,
      PVolgaRecord(Item2)^.Data + FI^.Offset, FI, IgnoreC)      //casesensitive or not
    else
      Result := 0;
    if DescSort then Result := -Result; //descending sort
    Inc(j);
  end;
end;

function _if(b: Boolean; a1, a2: integer): integer; overload;
begin
  if b then
    Result := a1
  else
    Result := a2;
end;

function _if(b: Boolean; a1, a2: string): string; overload;
begin
  if b then
    Result := a1
  else
    Result := a2;
end;

function vStrLen(Buffer: PChar): integer;
begin
  try
    Result := StrLen(Buffer);
  except
    Result := 0;
  end;
end;

//function that extracts a part of string, divided by tabulation

function ExtractValue(const Values: string; var Pos: integer): string;
var I: integer;
begin
  I := Pos;
  while (I <= Length(Values)) and (Values[I] <> Tabulator) do
    Inc(I);
  Result := Copy(Values, Pos, I - Pos);
  if (I <= Length(Values)) and (Values[I] = Tabulator) then Inc(I);
  Pos := I;
end;

//transformation all spaces into underlinings in the string

function vSpaceToChar(const s: string; p: char): string;
var i: integer;
begin
  Result := s;
  for i := 1 to Length(Result) do
    if Result[i] = ' ' then Result[i] := p; //specified symbol instead of space
end;

procedure Split(const s: string; const delim: string; SplitArray: TStrings);
var a: string;
  i, l: integer;
begin
  SplitArray.Clear;
  a := s;
  l := Length(delim);
  while Length(a) > 0 do
  begin
    i := Pos(delim, a);
    if i = 0 then
    begin
      SplitArray.Add(a);
      Exit;
    end;
    SplitArray.Add(Copy(a, 1, i - 1));
    a := Copy(a, i + l, MaxInt);
  end;
end;

{$J+}
function GetUnicod: string;     //unique 8 bytes code
const counter: integer = 0;
  starttime: integer = 0;
  base: byte = 62;
const tran: string[62] =
  '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
var hh, mi, ss, cc: word;
  date2000: TDateTime;
  aaa, bbb, ddd: string[3];
  base2, base3: longint;

  function Conv10toN(number: longint; numbyte: byte): string;
  {return the number with the BASE modulus, transformed to NUMBYTE byte}
  var a, b, c, d: integer;
  begin
    if numbyte = 2 then
      if number >= base2 then
        number := number - base2
      else
    else if number >= base3 then
      number := number - base3;
    a := number div base;               {integer part}
    b := number mod base;               {remainder}
    if numbyte = 3 then
    begin
      c := a div base;                  {integer part}
      d := a mod base;                  {remainder}
      Result := tran[c + 1] + tran[d + 1] + tran[b + 1];
    end
    else
      Result := tran[a + 1] + tran[b + 1];
  end;

begin
  base2 := base * base;
  base3 := base2 * base;
  DecodeTime(Time, hh, mi, ss, cc);
  if starttime = 0 then
  begin
    {we start counting  form the moment of the program launch}
    Counter := hh * 60 + mi;
    starttime := 1;
  end
  else
    Inc(Counter);
  if Counter >= base2 then Counter := 0;
  date2000 := EncodeDate(2000, 11, 1);
  {the first 3 bytes - a year, month, day, hour}
  aaa := Conv10toN(Trunc(Date - date2000) * 24 + hh, 3);
  {the second 2 bytes - minutes,seconds}
  bbb := Conv10toN(mi * 60 + ss, 2);
  {the last 3 bytes - milliseconds + the reference counter till it's value is equal to base*base}
  ddd := Conv10toN(((cc div 10) * base2 div 2) + Counter, 3);
  Result := aaa + bbb + ddd;
end;

function GetUnicod10: string;   //unique 10 bytes code
{  10 }
const counter: Cardinal = 0;
  starttime: integer = 0;
  base: byte = 36;
  date2001: TDateTime = 0;
  base2: Cardinal = 1296;
  base3: Cardinal = 46656;
  tic: Cardinal = 0;
const tran: string[36] =
  '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';
var hh, mi, ss, cc: word;
  aaa, ddd: string[3];
  bbb: string[4];

  function Conv10toN(number: Cardinal; numbyte: byte): string;
  {return the number with the BASE modulus, transformed to NUMBYTE byte}
  var a, b, c, d: byte;
  begin
    a := number div base3;
    number := number - base3 * a;
    b := number div base2;
    number := number - base2 * b;
    c := number div base;
    d := number mod base;
    Result := tran[a + 1] + tran[b + 1] + tran[c + 1] + tran[d + 1];
    Result := Copy(Result, 5 - numbyte, numbyte);
  end;

begin
  DecodeTime(SysUtils.Time, hh, mi, ss, cc);
  if starttime = 0 then
  begin
    {we start counting  form the moment of the program launch}
    date2001 := EncodeDate(2001, 1, 1);
    tic := GetTickCount;
    Counter := hh * 60 + mi; 
    starttime := 1;
  end
  else
    Inc(Counter);
  if Counter >= base3 then Counter := 0;
  aaa := Conv10toN(Trunc(Date - date2001), 3);
  bbb := Conv10toN(hh * 36000 + mi * 600 + ss * 10 + (cc div 100), 4);
  ddd := Conv10toN((GetTickCount - tic + Counter) mod base3, 3);
  Result := aaa + bbb + ddd;
end;
{$J-}

{$IFDEF PRO}
//CRC calculation

procedure Crc(data: byte);
var a, c: byte;
  b: boolean;
begin
  crcl := crcl xor data;
  for a := 1 to 8 do
  begin
    b := false;
    c := 0;
    if ((crcl and 1) = 1) then b := true;
    if ((crch and 1) = 1) then c := $80;
    crcl := (crcl shr (1)) or c;
    crch := crch shr (1);
    if b then
    begin
      crch := crch xor $A0;
      crcl := crcl xor $01;
    end;
  end;
end;
{$ENDIF}

//three overloaded functions, returning supported type of the field

function GetFldType(ft: string): TVolgaFieldType; overload;
var i: TVolgaFieldType;
begin
  for i := Low(TVolgaFieldType) to High(TVolgaFieldType) do
    if ft = StrFieldType[i] then
    begin
      Result := i;
      Exit;
    end;
  Result := vftUnknown;
end;

function GetFldType(ft: TVolgaFieldType): string; overload;
begin
  Result := StrFieldType[ft];
end;

function GetFldType(ft: TFieldType; size: integer): TVolgaFieldType; overload;
begin
  case ft of
    ftString:
      if size > MAXSLEN then
        Result := vftString
      else
        Result := vftSmallStr;
    ftInteger: Result := vftInteger;
    ftLargeInt: Result := vftLargeInt;
    ftAutoInc: Result := vftAutoInc;
    ftSmallint, ftWord: Result := vftSmallInt;
    ftBoolean: Result := vftBoolean;
    ftFloat: Result := vftDouble;
    ftCurrency: Result := vftCurrency;
    ftDate: Result := vftDate;
    ftTime: Result := vftTime;
    ftDateTime: Result := vftDateTime;
    ftBlob: Result := vftBlob;
    ftMemo: Result := vftMemo;
    ftFmtMemo: Result := vftFmtMemo;
    ftGraphic: Result := vftGraphic;
  else
    Result := vftUnknown;
  end;
end;

function GetFldType2(ft: string): TFieldType;
begin
  Result := UsingFieldType[GetFldType(ft)];
end;

function GetCharType(ft: TVolgaFieldType; ASize: integer): char;
//type of the field as 1 byte
begin
  case ft of
    vftString:
      Result := 's';
    vftSmallStr:
      if ASize <= 9 then                //forming '1...'9''A...'Z' symbol
        Result := Chr(ASize + Ord('0'))
      else
        Result := Chr(ASize + Ord('A') - 10);
    vftBlob:
      Result := 'o';
    vftGraphic:
      Result := 'g';
    vftMemo:
      Result := 'm';
    vftFmtMemo:
      Result := 'r';                    //rtf-file
    vftSmallInt:
      Result := 'l';
    vftInteger:
      Result := 'i';
    vftLargeInt:
      Result := 'e';
    vftAutoInc:
      Result := 'a';
    vftCurrency:
      Result := 'c';
    vftDouble:
      Result := 'f';
    vftBoolean:
      Result := 'b';
    vftDate:
      Result := 'd';
    vftTime:
      Result := 'v';
    vftDateTime:
      Result := 't';
  else                                  //to prevent compiler errors
    Result := 's';
  end;
end;

function GetCursor(Dataset: TVolgaDataset): TVolgaCursor;
begin
  Dataset.CheckCursorOpen;
  Result := Dataset.FCursor;
end;

//******************************************************************************
//  functions for working with buffers
//******************************************************************************

function IsBufEmpty(Buffer: PChar; size: integer): Boolean;
//is buffer filled with zeroes?
var i: integer;
begin
  Result := true;
  for i := 0 to size - 1 do
  begin
    Result := Buffer[i] = #0;
    if not Result then Break;
  end;
end;

procedure KillBuf(Buffer: Pointer);
//clearing the chosened string using adress in buffer
var AdrStr: Pointer;
  Str: PChar;
begin
  AdrStr := Pointer(Pointer(Buffer)^);
  //Move(Buffer^, AdrStr, SIZECHAR);      //copy the address from the buffer
  if AdrStr <> nil then
  begin
    Str := PChar(AdrStr);
    if Str <> nil then FreeMem(Str);    //pointer on the current memo or string
    //AdrStr := nil;
  end;
  FillChar(Buffer^, SIZECHAR, $00);     //clearing the buffer
end;

procedure CopyLongString(BufSource, BufDest: PChar);
//copying of the string if it's adress is specified in buffer (the length of each buffer is SIZECHAR)
var Str1, Str2: PChar;
  AdrStr1, AdrStr2: Pointer;
  slen: integer;
begin
  KillBuf(BufDest);                     //clear old value
  AdrStr1 := Pointer(Pointer(BufSource)^);
  //Move(BufSource^, AdrStr1, SIZECHAR);  //read the address of the source string
  if AdrStr1 <> nil then
  begin
    Str1 := PChar(AdrStr1);
    slen := vStrLen(Str1);              //define it's real length
    if slen > 0 then
    begin
      Str2 := AllocMem(slen + 1);       //buffer for new string
      Move(Str1^, Str2^, slen);         //  -
      AdrStr2 := Pointer(Str2);         //  
      Pointer(Pointer(BufDest)^) := AdrStr2;
      //Move(AdrStr2, BufDest^, SIZECHAR); //copy the pointer on string
    end;
  end;
end;

procedure CopyBlobField(BufSource, BufDest: PChar);
//copying BLOB if it's address is specified in buffer (the length of each buffer SIZECHAR+SIZECHAR)
var buf1, buf2: PChar;
  AdrBlob1, AdrBlob2: Pointer;
  slen: integer;
begin
  KillBuf(BufDest);                     //clear old value
  AdrBlob1 := Pointer(Pointer(BufSource)^);
  //Move(BufSource^, AdrBlob1, SIZECHAR); //read the address source BLOB
  if AdrBlob1 <> nil then
  begin
    buf1 := PChar(AdrBlob1);            //source BLOB address
    Move((BufSource + SIZECHAR)^, slen, SIZEINT); //read the BLOB length
    if slen > 0 then
    begin
      buf2 := AllocMem(slen);           //buffer for new BLOB
      Move(buf1^, buf2^, slen);         //copy blob-field
      AdrBlob2 := Pointer(buf2);        //new BLOB address
      Pointer(Pointer(BufDest)^) := AdrBlob2;
      //Move(AdrBlob2, BufDest^, SIZECHAR); //copy the pointer on blob
      PInteger(BufDest + SIZECHAR)^ := slen;
      //Move(slen, (BufDest + SIZECHAR)^, SIZEINT); //copy the BLOB length
    end;
  end;
end;

function CompareBuf(const key1, key2: PChar; FI: PVolgaField; IgnoreCase: Boolean):
integer;
//comparison of the specified field in different buffer (key1,key2 -  buffers of the field, offset is already aplyed)

  function CompStr(const key1, key2: PChar): longint;
  var len, len2, minlen: longint;
  begin
    len := vstrlen(key1);
    len2 := vstrlen(key2);
    if (len = 0) or (len2 = 0) then
    begin
      Result := len - len2;
      Exit;
    end;
    minlen := len;
    if minlen > len2 then minlen := len2;
    if IgnoreCase then
      Result := AnsiStrLIComp(key1, key2, minlen)
    else
      Result := StrLComp(key1, key2, minlen);
    if (Result = 0) and (len <> len2) then
      Result := len - len2;             //if strings beginnings are equal, then we select the longest
  end;

var ri1, ri2: integer;
  rd1, rd2: double;
  rl1, rl2: smallint;
  rb1, rb2: WordBool;
  r64, r641, r642: int64;
  buf1, buf2: PChar;
  AdrStr1, AdrStr2: Pointer;
  slen1, slen2: integer;
begin
  Result := 0;
  case FI^.DataType of                  //field type dependent comparison
    vftSmallInt:
      begin
        rl1 := PSmallInt(key1)^;
        rl2 := PSmallInt(key2)^;
        Result := rl1 - rl2;
      end;
    vftInteger, vftAutoInc, vftDate, vftTime:
      begin
        ri1 := PInteger(key1)^;
        ri2 := PInteger(key2)^;
        Result := ri1 - ri2;
      end;
    vftLargeInt:
      begin
        r641 := PInt64(key1)^;
        r642 := PInt64(key2)^;
        //Range overflow protection
        r64 := r641 - r642;
        if r64 < 0 then
          Result := -1
        else if r64 > 0 then
          Result := 1
        else
          Result := 0;
      end;
    vftBoolean:
      begin
        rb1 := PWordBool(key1)^;
        rb2 := PWordBool(key2)^;
        Result := _if(rb1 = rb2, 0, _if(Boolean(rb1), -1, 1));
      end;
    vftDouble, vftCurrency, vftDateTime:
      begin
        rd1 := PDouble(key1)^;
        rd2 := PDouble(key2)^;
        Result := _if(rd1 > rd2, 1, _if(rd1 < rd2, -1, 0));
      end;
    vftSmallStr:
      begin
        if FI^.Align = taRightJustify then
        begin                           //sorting the short strings by right edge
          Result := vstrlen(key1) - vstrlen(key2);
          if Result <> 0 then Exit;     //otherwise compare as strings with the equal length
        end;
        Result := CompStr(key1, key2);
      end;
    vftString:
      begin
        AdrStr1 := Pointer(Pointer(key1)^);
        AdrStr2 := Pointer(Pointer(key2)^);
        if not LongBool(AdrStr1) or not LongBool(AdrStr2) then
        begin                           //one of the addresses is empty (the binary zeroes)
          Result := _if(LongBool(AdrStr1), 1, _if(LongBool(AdrStr2), -1, 0));
          Exit;
        end;
        Result := CompStr(PChar(AdrStr1), PChar(AdrStr2));
      end;
    vftMemo, vftFmtMemo:                //Blobs have no binary 0 at the end of buffer
      begin
        AdrStr1 := Pointer(Pointer(key1)^);
        AdrStr2 := Pointer(Pointer(key2)^);
        slen1 := PInteger(key1 + SIZECHAR)^;
        slen2 := PInteger(key2 + SIZECHAR)^;
        if not LongBool(AdrStr1) or not LongBool(AdrStr2) then
        begin                           //one of the addresses is empty (the binary zeroes)
          Result := _if(LongBool(AdrStr1), 1, _if(LongBool(AdrStr2), -1, 0));
          Exit;
        end;
        if (slen1 = 0) or (slen2 = 0) then
          Result := slen1 - slen2
        else
        begin
          buf1 := AllocMem(slen1 + 1);
          Move(AdrStr1^, buf1^, slen1);
          buf2 := AllocMem(slen2 + 1);
          Move(AdrStr2^, buf2^, slen2);
          Result := CompStr(buf1, buf2);
          Freemem(buf1);
          Freemem(buf2);
        end;
      end;
  end;
end;

function CompareRec(Buf1, Buf2: PChar; ListFI: TList; Count: integer = -1): integer;
var j, cnt: integer;
  FI: PVolgaField;
begin                                   //compare specified fields of two buffers with records
  Result := 0;
  j := 0;
  cnt := ListFI.Count;
  if Count >= 0 then cnt := Count;
  while (Result = 0) and (j < cnt) do
  begin
    FI := PVolgaField(ListFI.Items[j]); //the field, in that we are searching
    Result := CompareBuf(Buf1 + FI^.Offset, Buf2 + FI^.Offset, FI, false);
    Inc(j);
  end;
end;

procedure AssignBufFromStr(Buf: PChar; const Value: string; FI: PVolgaField);
//write value, recorded as string in field buffer and add spaces till the length of the field
var ri, slen, E: integer;
  rl: smallint;
  rd: double;
  r64: int64;
  rb: WordBool;
  rdt: TDateTime;
  rt: TTimeStamp;
  Str: string;
  Str2: PChar;
  AdrStr2: Pointer;
begin
  slen := Length(Value);
  if FI^.DataType in [vftString, vftMemo, vftFmtMemo] then
    KillBuf(Buf);                       //clear old value
  FillChar(Buf^, FI^.DataSize, $00);
  if slen = 0 then Exit;

  case FI^.DataType of
    vftSmallStr:
      begin
        if slen > FI^.Size then slen := FI^.Size;
        StrLCopy(Buf, PChar(Value), slen);
      end;
    vftString:
      begin
        if slen > FI^.Size then slen := FI^.Size;
        Str2 := AllocMem(FI^.Size + 1); //allocate the memory
        StrLCopy(Str2, PChar(Value), slen); //and copy the string
        AdrStr2 := Pointer(Str2);
        Pointer(Pointer(Buf)^) := AdrStr2;
      end;
    vftMemo, vftFmtMemo:
      begin
        if slen > FI^.Size then slen := FI^.Size;
        Str2 := AllocMem(slen + 1);     //allocate the memory
        StrLCopy(Str2, PChar(Value), slen); //and copy the string
        AdrStr2 := Pointer(Str2);
        Pointer(Pointer(Buf)^) := AdrStr2;
        PInteger(Buf + SIZECHAR)^ := slen;
      end;
    vftSmallInt:
      begin
        Val(Value, rl, E);
        if E = 0 then                   //some convertation errors?
          PSmallInt(Buf)^ := rl;
      end;
    vftInteger, vftAutoInc:
      begin
        Val(Value, ri, E);
        if E = 0 then                   //some convertation errors?
          PInteger(Buf)^ := ri;
      end;
    vftLargeInt:
      begin
        Val(Value, r64, E);
        if E = 0 then                   //some convertation errors?
          PInt64(Buf)^ := r64;
      end;
    vftDouble, vftCurrency:
      try
        rd := StrToFloat(PChar(Value));
        PDouble(Buf)^ := rd;
      except;
      end;
    vftDate:
      try
        rdt := StrToDate(Value);
        rt := DateTimeToTimeStamp(rdt);
        ri := rt.Date;
        PInteger(Buf)^ := ri;
      except;
      end;
    vftTime:
      try
        rdt := StrToTime(Value);
        rt := DateTimeToTimeStamp(rdt);
        ri := rt.Time;
        PInteger(Buf)^ := ri;
      except;
      end;
    vftDateTime:
      try
        rdt := StrToDateTime(Value);
        rt := DateTimeToTimeStamp(rdt);
        rdt := TimeStampToMSecs(rt);
        PDouble(Buf)^ := rdt;
      except;
      end;
    vftBoolean:
      begin
        Str := AnsiUpperCase(Value);
        if (Str = 'TRUE') or (Str = '') or (Str = 'YES') or (Str = 'T') or (Str = '1')
          then
          rb := true
        else
          rb := false;
        PWordBool(Buf)^ := rb;
      end;
  end;
end;

procedure AssignBufFromVar(Buf: PChar; const Value: TVarRec; FI: PVolgaField);
//write value in field buffer, entered as variant
var ri: integer;
  rd: double;
  //rb: WordBool;
  r64: int64;
  rt: TTimeStamp;
  rc: TDateTime;
  Str: string;
begin
  //for BLOB?????????????
  if FI^.DataType in [vftString, vftMemo, vftFmtMemo] then
    KillBuf(Buf);                       //clear old value
  FillChar(Buf^, FI^.DataSize, $00);
  with Value do
    if VType in [vtChar, vtString, vtPChar, vtAnsiString] then
    begin
      case VType of
        vtChar:
          Str := VChar;
        vtString:
          Str := VString^;
        vtPChar:
          Str := VPChar;
        vtAnsiString:
          Str := string(VAnsiString);
      end;
      AssignBufFromStr(Buf, Str, FI);
    end
    else
      case VType of
        vtInteger:
          PInteger(Buf)^ := VInteger;
          //Move(VInteger, Buf^, SIZEINT);
        vtInt64:
          begin
            r64 := VInt64^;
            Move(r64, Buf^, SIZEINT64);
          end;
        vtBoolean:
          begin
            PWordBool(Buf)^ := VBoolean;
            //rb := VBoolean;
            //Move(rb, Buf^, SIZEBOOL);
          end;
        vtExtended:
          begin
            rd := VExtended^;
            case FI^.DataType of
              vftDate:
                begin
                  rt := DateTimeToTimeStamp(rd);
                  ri := rt.Date;
                  PInteger(Buf)^ := ri;
                  //Move(ri, Buf^, SIZEINT);
                end;
              vftTime:
                begin
                  rt := DateTimeToTimeStamp(rd);
                  ri := rt.Time;
                  PInteger(Buf)^ := ri;
                  //Move(ri, Buf^, SIZEINT);
                end;
              vftDateTime:
                begin
                  rt := DateTimeToTimeStamp(rd);
                  rc := TimeStampToMSecs(rt);
                  PDouble(Buf)^ := rc;
                  //Move(rc, Buf^, SIZEDOUB);
                end;
            else                        //vftDouble, vftCurrency
              PDouble(Buf)^ := rd;
              //Move(rd, Buf^, SIZEDOUB);
            end;
          end;
        vtCurrency:
          Move(VCurrency^, Buf^, SIZEDOUB);
        vtVariant:
          if not VarIsEmpty(VVariant^) then
          begin
            Str := string(VVariant^);
            AssignBufFromStr(Buf, Str, FI);
          end;
      //else leave the buffer as it is
      end;
end;

procedure GetFldInfoList(Sender: TVolgaCursor; List: TList;
  const FieldNames: string);
var FI: PVolgaField;
  iPos: integer;
begin
  iPos := 1;
  while iPos <= Length(FieldNames) do
  begin
    FI := Sender.FindFieldInfo(Db.ExtractFieldName(FieldNames, iPos)); //described in db.pas
    if (FI <> nil) and (FI^.DataType <> vftMemo) and (FI^.DataType <> vftFmtMemo)
      and (FI^.DataType <> vftGraphic) and (FI^.DataType <> vftBlob) then
      List.Add(FI);                     //don't works with blob-fields
  end;
end;

//==============================================================================
// TVolgaMasterLink
//==============================================================================

constructor TVolgaMasterLink.Create(DataSet: TVolgaDataSet);
begin
  inherited Create;
  FDataSet := DataSet;                  //owner - detail dataset
  FFields := TList.Create;
end;

destructor TVolgaMasterLink.Destroy;
begin
  FFields.Free;
  inherited Destroy;
end;

procedure TVolgaMasterLink.ActiveChanged;
begin                                   //Changes in the Active property
  FFields.Clear;
  if Active then
  try                                   //fields list of the link from the master-table
    DataSet.GetFieldList(FFields, FFieldNames);
  except
    FFields.Clear;
    raise;
  end;
  if FDataSet.Active and not (csDestroying in FDataSet.ComponentState) then
    if Active and (FFields.Count > 0) then
    begin
      if Assigned(FOnMasterChange) then FOnMasterChange(Self);
    end
    else if Assigned(FOnMasterDisable) then
      FOnMasterDisable(Self);
end;

procedure TVolgaMasterLink.CheckBrowseMode;
begin
  if FDataSet.Active then FDataSet.CheckBrowseMode;
end;

function TVolgaMasterLink.GetDetailDataSet: TDataSet;
begin
  Result := FDataSet;
end;

procedure TVolgaMasterLink.LayoutChanged;
begin                                   //the data source has changed the number of records buffered.
  ActiveChanged;
end;

procedure TVolgaMasterLink.RecordChanged(Field: TField);
begin                                   //the contents of one of the fields in the current record changes.
  if (DataSource.State <> dsSetKey) and FDataSet.Active and
    (FFields.Count > 0) and ((Field = nil) or (FFields.IndexOf(Field) >= 0)) and
    Assigned(FOnMasterChange) and (DataSource.DataSet is TVolgaDataSet) then
    FOnMasterChange(Self);
end;

procedure TVolgaMasterLink.SetFieldNames(const Value: string);
begin
  if FFieldNames <> Value then
  begin
    FFieldNames := Value;
    ActiveChanged;                      //recreate the fields list
  end;
end;

//******************************************************************************
//      TVolgaBlobStream - an object for support of the BLOB-fields
//******************************************************************************

constructor TVolgaBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
begin
  FField := Field;
  FFieldNo := FField.FieldNo;
  FDataSet := TVolgaDataset(FField.DataSet);
  FDataSet.CheckActive;                 //if not opened then exception
  if FDataSet.IsEmpty then Exit;
  if Mode <> bmRead then                //if we are going to to write, then Mode must not be bmRead
  begin
    if FField.ReadOnly then
      DatabaseErrorFmt(V_READONLYFLD, [FField.DisplayName], FDataSet);
    if (FDataSet.State <> dsEdit) and (FDataSet.State <> dsInsert) then
      DatabaseError(V_NOTEDITMODE, FDataSet);
  end;
  if Mode = bmWrite then
    Truncate                            //clear the stream
  else
    ReadBlobData;
end;

destructor TVolgaBlobStream.Destroy;
var Offset, len: integer;
  RecBuf, BlobBuf: PChar;
  AdrStr: Pointer;
begin
  if FModified then
  try
    Offset := FDataSet.GetFieldOffset(FFieldNo); //offset of the field in record
    RecBuf := FDataSet.ActiveBuffer;    //buffer of the current active record
    KillBuf(RecBuf + Offset);           //clear old value
    GetMem(BlobBuf, Size);              //allocate the memory under blob
    Move(Memory^, BlobBuf^, Size);
    AdrStr := Pointer(BlobBuf);
    Pointer(Pointer(RecBuf + Offset)^) := AdrStr;
    //Move(AdrStr, (RecBuf + Offset)^, SIZECHAR); //address blob
    len := Size;
    PInteger(RecBuf + Offset + SIZECHAR)^ := len;
    //Move(len, (RecBuf + Offset + SIZECHAR)^, SIZEINT); //length blob
    FField.Modified := true;
    FDataSet.DataEvent(deFieldChange, longint(FField)); //the field changed!!!!!!!
  except
    Application.HandleException(Self);
  end;
  inherited;
end;

procedure TVolgaBlobStream.ReadBlobData;
var Offset, len: integer;
  AdrStr: Pointer;
  RecBuf, BlobBuf: PChar;
begin
  Offset := FDataSet.GetFieldOffset(FFieldNo); //offset of the field pre-recorded
  if not FDataSet.GetActiveRecBuf(RecBuf) then Exit; //buffer of the current active record
  AdrStr := Pointer(Pointer(RecBuf + Offset)^);
  len := PInteger(RecBuf + Offset + SIZECHAR)^;
  //Move((RecBuf + Offset)^, AdrStr, SIZECHAR); //copy the address
  //Move((RecBuf + Offset + SIZECHAR)^, len, SIZEINT); //copy the length
  if (AdrStr <> nil) and (len > 0) then
  begin
    BlobBuf := PChar(AdrStr);
    Position := 0;
    SetSize(len);
    Move(BlobBuf^, Memory^, len);
  end;
end;

procedure TVolgaBlobStream.Truncate;
//clear BLOB
begin
  Clear;
  FModified := true;
end;

function TVolgaBlobStream.Write(const Buffer; Count: Integer): Longint;
begin
  Result := inherited Write(Buffer, Count);
  FModified := true;                    //we realy fill in buffer when the copy BLOB-stream is destroyed!!!
end;

//******************************************************************************
//          TVolgaCursor - an internal cursor with data
//******************************************************************************

constructor TVolgaCursor.Create;
begin
  FData := TList.Create;
  FFldInfo := TList.Create;
  FFieldLayout := nil;
  FRecSize := 0;
  FPartialOpen := false;
{$IFDEF PRO}
  FCiphered := false;
  FAutoIncName := '';
  FLastAutoInc := 0;
  FAutoIncField := nil;
{$ENDIF}
  FNotes := '';
end;

destructor TVolgaCursor.Destroy;
begin
  Clear(true);
  FData.Free;
  FData := nil;
  FFldInfo.Free;
  FFldInfo := nil;
end;

function TVolgaCursor.AddNewRecord(Stream: TStream; var Msg: string): Boolean;
var
  pRec: PVolgaRecord;
  slen, flen, reclen: integer;
  MemoFld, Mesto: PChar;
  AdrMemo: Pointer;
  Field: PVolgaField;
  FL: PChar;
  i: integer;
  wronglen, Accept: Boolean;
label LabelClear;
begin
  Msg := V_ENDBUF;
  Result := false;
  FL := AllocMem(FFldCount + 1);
  try
    New(pRec);                          //locating the new structure in memory
    pRec^.Status := vusUnmodified;
    pRec^.OldData := nil;
    pRec^.RecordData := nil;
    pRec^.Data := AllocMem(FRecSize);   //allocate the memory for data  writing
    //lengthes of the fields in this record
    Stream.ReadBuffer(FL^, FFldCount);  //we are located after lengths of the fields descriptions  in record
    reclen := FFldCount;
  except                                //EReadError
    FreeMem(FL);
    Exit;
  end;
  try
    //produce the transformation of the sent data in correspondence with the fields type
    for i := 0 to FFldInfo.Count - 1 do
    begin                               //for each field
      Field := FieldInfo[i];
      Mesto := pRec^.Data + Field^.Offset;
      flen := integer(FL[i]);           //real length of the field
      case Field^.DataType of
        vftString, vftMemo, vftFmtMemo, vftGraphic, vftBlob:
          wronglen := (flen <> SIZEINT) and (flen <> 0);
      else
        wronglen := flen > Field^.DataSize;
      end;
      if wronglen then
      begin
        Msg := Format(V_FLDLEN, [Field^.Name, flen, FData.Count + 1]);
        EReadError.Create('');
        Result := false;
        Exit;
      end;
      if flen > 0 then                  //if this field is nonblank
      begin
        Inc(reclen, flen);
        case Field^.DataType of
          vftSmallInt, vftInteger, vftLargeInt, vftAutoInc, vftBoolean, vftDouble,
            vftCurrency, vftDate, vftTime, vftDateTime, vftSmallStr: //short strings always have 0x00 in the end
            Stream.ReadBuffer(Mesto^, flen);
          vftString:
            begin
              Stream.ReadBuffer(slen, flen);
              if (slen >= 0) and (slen < 8192) then //till 8192 bytes - TStringField limit
              begin
                Inc(reclen, slen);
                MemoFld := AllocMem(slen + 1);
                Stream.ReadBuffer(MemoFld^, slen); //string-field itself
                AdrMemo := Pointer(MemoFld);
                Pointer(Pointer(Mesto)^) := AdrMemo;
                //Move(AdrMemo, Mesto^, SIZECHAR); //pointer on string
              end
              else
              begin
                Msg := Format(V_STRLEN, [Field^.Name, slen, FData.Count + 1]);
                EReadError.Create('');
              end;
            end;
          vftMemo, vftFmtMemo, vftGraphic, vftBlob:
            begin
              Stream.ReadBuffer(slen, flen);
              if (slen >= 0) and (slen < 1024 * 1024 * 1024) then //till 1 Gb - our limit
              begin
                Inc(reclen, slen);
                MemoFld := AllocMem(slen);
                Stream.ReadBuffer(MemoFld^, slen); //Blob-field itself
                AdrMemo := Pointer(MemoFld);
                Pointer(Pointer(Mesto)^) := AdrMemo;
                //Move(AdrMemo, Mesto^, SIZECHAR); //pointer on Blob
                PInteger(Mesto + SIZECHAR)^ := slen;
                //Move(slen, (pRec^.Data + Field^.Offset + SIZECHAR)^, SIZEINT); //Blob length
              end
              else
              begin
                Msg := Format(V_BLOBLEN, [Field^.Name, slen, FData.Count + 1]);
                EReadError.Create('');
              end;
            end;
        end;
      end;                              //else (the field is empty := nil)
      Result := true;                   //for correct ending of the cycle
    end;
  except;
  end;
  FreeMem(FL);
  if Result then
  begin
    Msg := '';                          //if record does not suit to filter, it is not error
    pRec^.Size := reclen;
{$IFDEF PRO}
    Accept := true;
    if FIsFiltered then FFilterClause(Self, pRec, Accept);
    if Accept then                      //if filter is nominated for the record loading, then applying it
{$ENDIF}
      FData.Add(pRec)
{$IFDEF PRO}
    else
      goto LabelClear;
{$ENDIF}
  end
  else                                  //record was not successfully read or does not suit to filter
  begin
    LabelClear:
    ClearData(pRec^.Data);              //clearing unused buffer
    Dispose(pRec);
  end;
end;

procedure TVolgaCursor.CancelUpdates;
var pRec: PVolgaRecord;
  i: integer;
begin
  for i := 0 to FData.Count - 1 do
  begin
    pRec := PVolgaRecord(FData.Items[i]);
    case pRec^.Status of
      vusUnmodified: ;                  //nothing to do
      vusModified:                      //restore default values
        UndoRecord(pRec);
      vusInserted:                      //delete newly entered values
        begin
          pRec^.Status := vusAbsent;
          ClearData(pRec^.OldData);
          ClearData(pRec^.Data);
{$IFDEF PRO}
          if Assigned(FOnDestroyRec) then FOnDestroyRec(pRec^.RecordData);
{$ENDIF}
          pRec^.RecordData := nil;
        end;
      vusDeleted:
        UndoRecord(pRec);               //restore default values if they were
      vusAbsent: ;                      //nothing to do
    end;
  end;
end;

procedure TVolgaCursor.CalcRecSize(pRec: PVolgaRecord);
//calc size of whole record when it will be written to stream (file)
var
  FI: PVolgaField;
  fldbuf: PChar;
  reclen, len: integer;
  j: integer;
  //bl: WordBool;
  //bd: double;
  AdrStr: Pointer;
begin
  if (pRec^.Status = vusAbsent) then Exit;
  reclen := FFldCount;                  //the fields lengthes
  for j := 0 to FFldCount - 1 do
  begin
    FI := FieldInfo[j];
    case FI^.DataType of
      vftSmallInt:
          if PSmallInt(pRec^.Data + FI^.Offset)^ <> 0 then
            Inc(reclen, 2);
      vftInteger, vftAutoInc, vftDate, vftTime:
          if PInteger(pRec^.Data + FI^.Offset)^ <> 0 then
            Inc(reclen, SIZEINT);
      vftLargeInt:
          if PInt64(pRec^.Data + FI^.Offset)^ <> 0 then
            Inc(reclen, SIZEINT64);
      vftBoolean:
          if PWordBool(pRec^.Data + FI^.Offset)^ then   //false values are not written
            Inc(reclen, SIZEBOOL);
      vftDouble, vftCurrency, vftDateTime:
          if PDouble(pRec^.Data + FI^.Offset)^ <> 0 then  //zero values are not written
            Inc(reclen, SIZEDOUB);
      vftSmallStr:
        begin
          //fldbuf := AllocMem(FI^.DataSize); //buffer for the short strings copying
          //Move((pRec^.Data + FI^.Offset)^, fldbuf^, FI^.DataSize); //including the closing 000
          //len := vstrlen(fldbuf);
          len := vstrlen(pRec^.Data + FI^.Offset);
          if len > 0 then               //empty strings are not written
            Inc(reclen, len);
          //FreeMem(fldbuf);
        end;
      vftString, vftMemo, vftFmtMemo, vftGraphic, vftBlob:
        begin
          AdrStr := Pointer(Pointer(pRec^.Data + FI.Offset)^);
          //Move((pRec^.Data + FI.Offset)^, AdrStr, SIZECHAR); //copy in address buffer
          if AdrStr <> nil then
          begin
            fldbuf := PChar(AdrStr);    //strings and blob are taken from existing buffer immediately
            if FI^.DataType = vftString then
              len := vstrlen(fldbuf)
            else                        //blob length is specified obviously
              len := PInteger(pRec^.Data + FI^.Offset + SIZECHAR)^;
              //Move((pRec^.Data + FI^.Offset + SIZECHAR)^, len, SIZEINT); //copy the length
            if len > 0 then             //empty strings are not written
            begin
              Inc(reclen, SIZEINT);     //the length of the length
              Inc(reclen, len);         //the length of the string or blob
            end;
          end;
        end;
    end;
  end;
  pRec^.Size := reclen;
end;

function TVolgaCursor.CalcStreamSize: integer;
//calc size of stream (file) for whole data
var siz, i: integer;
  pRec: PVolgaRecord;
begin
  siz := 0;
  Inc(siz, SIZEINT);                    //the size of the whole buffer
  Inc(siz, 1);                          //special symbol
{$IFDEF PRO}
  if FCiphered then
    Inc(siz, LenPassw);
{$ENDIF}
  Inc(siz, SIZEINT);                    //the record's counter
  Inc(siz, SIZEINT);                    //fields counter
  //information about the structure
  Inc(siz, SIZEINT);                    //the length of the information about structure
  Inc(siz, FFldLen);                    //information (the last separator is included)
  //note in table
  if FNotes > '' then
  begin
    Inc(siz, 1);                        //notes flag
    Inc(siz, 2);                        //lenght of notes
    Inc(siz, Length(FNotes));
  end;
  for i := 0 to FData.Count - 1 do
  begin
    pRec := PVolgaRecord(FData.Items[i]);
    siz := siz + pRec^.Size;
  end;
  Result := siz;
end;

{$IFDEF PRO}

function TVolgaCursor.CheckDuplicate(Buf: PChar; NonCheckedRec: PVolgaRecord; const Fn:
  string): boolean;
var start, non: integer;
  FSearchFI: TList;
  pRec: PVolgaRecord;
begin
  Result := false;
  if Fn <> '' then
  begin
    FSearchFI := TList.Create;
    GetFldInfoList(Self, FSearchFI, Fn); //fields list for searching in it
  end
  else
    FSearchFI := FFldInfo;
  start := 0;
  non := FData.IndexOf(NonCheckedRec);
  while not Result and (start < FData.Count) do //linear search
  begin
    pRec := PVolgaRecord(FData.Items[start]);
    Result := (start <> non) and (pRec^.Status <> vusDeleted) and (pRec^.Status <>
      vusAbsent) and (CompareRec(Buf, pRec^.Data, FSearchFI, FSearchFI.Count) = 0);
    if not Result then
      Inc(start);
  end;
  if Fn <> '' then
    FSearchFI.Free;
end;
{$ENDIF}

procedure TVolgaCursor.Clear(WithFields: Boolean);
var pRec: PVolgaRecord;
  F: PVolgaField;
  i: integer;
begin
  for i := FData.Count - 1 downto 0 do
  begin
    pRec := PVolgaRecord(FData.Items[i]);
    ClearData(pRec^.OldData);
    ClearData(pRec^.Data);
{$IFDEF PRO}
    if Assigned(FOnDestroyRec) then FOnDestroyRec(pRec^.RecordData);
{$ENDIF}
    Dispose(pRec);
  end;
{$IFDEF PRO}
  FLastAutoInc := 0;
{$ENDIF}
  FData.Clear;
  if WithFields then                    //delete with field info and field layout
  begin
{$IFDEF PRO}
    FCiphered := false;
    FAutoIncName := '';
    FAutoIncField := nil;
{$ENDIF}
    FNotes := '';
    for i := FFldInfo.Count - 1 downto 0 do
    begin
      F := (PVolgaField(FFldInfo.Items[i]));
      Dispose(F);
    end;
    FFldInfo.Clear;
    if FFieldLayout <> nil then         //fields descriptions buffer
    begin
      FreeMem(FFieldLayout);
      FFieldLayout := nil;
    end;
{$IFDEF PRO}
    if FPasswBuf <> nil then
    begin
      FreeMem(FPasswBuf);
      FPasswBuf := nil;
      FPassword := '';
    end;
{$ENDIF}
  end;
end;

procedure TVolgaCursor.ClearBuf(RecBuf: PChar);
var j: integer;
begin
  if RecBuf = nil then Exit;
  for j := 0 to FFldInfo.Count - 1 do
  begin                                 //free the memory from string-fields
    case FieldInfo[j]^.DataType of
      vftString, vftMemo, vftFmtMemo, vftGraphic, vftBlob:
        KillBuf(RecBuf + FieldInfo[j]^.Offset);
    end;
  end;
end;

procedure TVolgaCursor.ClearData(var RecBuf: PChar);
begin
  if RecBuf <> nil then                 //can be Absent!!
  begin
    ClearBuf(RecBuf);
    FreeMem(RecBuf);
    RecBuf := nil;
  end;
end;

procedure TVolgaCursor.CommitUpdates;
var pRec: PVolgaRecord;
  i: integer;
begin
  for i := 0 to FData.Count - 1 do
  begin
    pRec := PVolgaRecord(FData.Items[i]);
    if pRec^.Status = vusDeleted then
    begin
      pRec^.Status := vusAbsent;
      ClearData(pRec^.OldData);
      ClearData(pRec^.Data);
{$IFDEF PRO}
      if Assigned(FOnDestroyRec) then FOnDestroyRec(pRec^.RecordData);
{$ENDIF}
      pRec^.RecordData := nil;
    end
    else if pRec^.Status <> vusAbsent then
    begin
      pRec^.Status := vusUnmodified;
      ClearData(pRec^.OldData);
    end;
  end;
end;

procedure TVolgaCursor.CopyRecordBuf(SourceBuf, DestBuf: PChar);
var FI: PVolgaField;
  j: integer;
begin                                   //copy the whole record and buffers of the long string-fields and BLOB
  for j := 0 to FFldInfo.Count - 1 do
  begin
    FI := FieldInfo[j];
    case FI^.DataType of
      vftString:
        CopyLongString(SourceBuf + FI^.Offset, DestBuf + FI^.Offset);
      vftMemo, vftFmtMemo, vftGraphic, vftBlob:
        CopyBlobField(SourceBuf + FI^.Offset, DestBuf + FI^.Offset);
    else                                //the way of the copying of every field except string and BLOB
      Move((SourceBuf + FI^.Offset)^, (DestBuf + FI^.Offset)^, FI^.DataSize);
    end;
  end;
end;

function TVolgaCursor.CreateUndoCopy(pRec: PVolgaRecord): Boolean;
begin
  Result := false;
  if (pRec = nil) or (pRec^.Status = vusDeleted) or (pRec^.Status = vusAbsent) then
    Exit;
  if pRec^.OldData = nil then
  begin                                 //record is not modified yet - copy it
    pRec^.OldData := AllocMem(FRecSize); //allocate the memory for data that is necessary for record's UNDO-copy
    CopyRecordBuf(pRec^.Data, pRec^.OldData);
  end;
  Result := true;
end;

function TVolgaCursor.DeleteRecord(CurRec: PVolgaRecord;
  var Msg: string): Boolean;
var Index: integer;
  pRec: PVolgaRecord;
begin
  Msg := '';
  Result := false;
  //removing of the existing record
  Index := FData.IndexOf(CurRec);
  if Index <= -1 then                   //there is no record with such pointer
  begin
    Msg := V_RECNOTFOUND;
    Exit;
  end;
  pRec := PVolgaRecord(FData.Items[Index]);
  if (pRec^.Status = vusDeleted) or (pRec^.Status = vusAbsent) then
  begin                                 //record was removed earlier
    Msg := V_RECDELETED;
    Exit;
  end;
  if (pRec^.Status = vusInserted) then
  begin
    pRec^.Status := vusAbsent;
    ClearData(pRec^.OldData);
    ClearData(pRec^.Data);
{$IFDEF PRO}
    if Assigned(FOnDestroyRec) then FOnDestroyRec(pRec^.RecordData);
{$ENDIF}
    pRec^.RecordData := nil;
  end
  else
    pRec^.Status := vusDeleted;
  Result := true;
end;

procedure TVolgaCursor.FillFieldLayout;
var temp, dlina, typ, nn: string; i: integer;
  FI: PVolgaField;
begin
  //fill field layout buffer with field info
  if FFieldLayout <> nil then         //fields descriptions buffer
  begin
    FreeMem(FFieldLayout);
    FFieldLayout := nil;
  end;
  temp := '';
  for i := 0 to FFldInfo.Count - 1 do
  begin
    FI := FieldInfo[i];
    dlina := '';
    typ := GetCharType(FI^.DataType, FI^.Size);
    if typ = 's' then dlina := IntToStr(FI^.Size);
    nn := vSpaceToChar(FI^.Name);        //enderscores instead of spaces
    temp := temp + FI^.Origin + '.' + nn + ' ' + typ + dlina;
    if (FI^.DefValue > '') or (FI^.Description > '') then
    begin
      temp := temp + '~';
      if FI^.DefValue > '' then
        temp := temp + FI^.DefValue;
      temp := temp + '~';
      if FI^.Description > '' then
        temp := temp + FI^.Description;
    end;
    temp := temp + FieldSeparator;
  end;
  temp := temp + RecordSeparator;
  FFldLen := Length(temp);
  FFieldLayout := AllocMem(FFldLen + 1);  //header field info with last delimiter
  StrLCopy(FFieldLayout, PChar(temp), FFldLen);
end;

function TVolgaCursor.FindFieldInfo(const Name: string): PVolgaField;
var i: integer;
begin
  for i := 0 to FFldInfo.Count - 1 do
  begin
    Result := PVolgaField(FFldInfo.Items[i]);
    if AnsiCompareText(Result^.Name, Name) = 0 then //caseinsencitive
      Exit;
  end;
  Result := nil;
end;

function TVolgaCursor.GetData(Index: integer): PVolgaRecord;
begin
  Result := PVolgaRecord(FData.Items[Index]);
end;

function TVolgaCursor.GetFieldCount: integer;
begin
  Result := FFldInfo.Count;
end;

function TVolgaCursor.GetFieldInfo(Index: integer): PVolgaField;
begin
  if Index < 0 then
    Result := nil
  else if Index >= FFldInfo.Count then
    Result := nil
  else
    Result := PVolgaField(FFldInfo.Items[Index]);
end;

function TVolgaCursor.GetFieldText(FI: PVolgaField; P: PVolgaRecord;
  IsOld: Boolean; YearFirst: Boolean): string;
var
  ri, ofs, len: integer;
  rl: smallint;
  r64: int64;
  rd: double;
  //rc: Comp;
  rb: WordBool;
  rt: TTimeStamp;
  dtr: TDateTimeRec;
  AdrStr: Pointer;
  typ: TVolgaFieldType;
  Buf, Str: PChar;
const fmtD: string = 'yyyy/mm/dd';
  fmtDT: string = 'yyyy/mm/dd hh:nn:ss';
begin
  Result := V_NOTFOUND;
  ofs := FI^.Offset;
  len := FI^.DataSize;
  typ := FI^.DataType;
  //Buf := AllocMem(len);                 //buffer for the field's value
  //read the bytes from the record with the offset of this field
  if IsOld and (P^.OldData <> nil) then //value before all changes
    Buf := P^.OldData + ofs
    //Move((P^.OldData + ofs)^, Buf^, len)
  else
    Buf := P^.Data + ofs;
    //Move((P^.Data + ofs)^, Buf^, len);  //the last changed value
  case typ of
    vftSmallInt:
      begin
        rl := PSmallInt(Buf)^;
        //Move(Buf^, rl, len);
        Result := IntToStr(rl);
      end;
    vftInteger, vftAutoInc:
      begin
        ri := PInteger(Buf)^;
        //Move(Buf^, ri, len);
        Result := IntToStr(ri);
      end;
    vftLargeInt:
      begin
        r64 := PInt64(Buf)^;
        //Move(Buf^, r64, len);
        Result := inttostr(r64);
      end;
    vftBoolean:
      begin
        rb := PWordBool(Buf)^;
        //Move(Buf^, rb, len);
        Result := _if(rb, V_YES, V_NO);
      end;
    vftDouble, vftCurrency:
      begin
        rd := PDouble(Buf)^;
        //Move(Buf^, rd, len);            //only 6 signs after full stop is important, the rest is rubbish -> del it
        Result := FormatFloat('0.######', rd);
      end;
    vftDate:
      begin
        ri := PInteger(Buf)^;
        //Move(Buf^, ri, len);
        if ri > 0 then
        begin
          rt.Date := ri;
          rt.Time := 0;
          if YearFirst then
            Result := FormatDateTime(fmtD, TimeStampToDateTime(rt))
          else
            Result := DateToStr(TimeStampToDateTime(rt))
        end
        else
          Result := '';
      end;
    vftTime:
      begin
        ri := PInteger(Buf)^;
        //Move(Buf^, ri, len);
        if ri > 0 then
        begin
          rt.Time := ri;
          rt.Date := 0;
          Result := TimeToStr(TimeStampToDateTime(rt));
        end
        else
          Result := '';
      end;
    vftDateTime:
      begin
        Move(Buf^, dtr, len);
        try
          rt := MSecsToTimeStamp(dtr.DateTime);
        except
          rt.Time := 0;
          rt.Date := 0;
        end;
        if YearFirst then
          Result := FormatDateTime(fmtDT, TimeStampToDateTime(rt))
        else
          Result := DateTimeToStr(TimeStampToDateTime(rt));
      end;
    vftSmallStr:
      Result := string(Buf);
    vftString, vftMemo, vftFmtMemo:
      begin
        AdrStr := Pointer(Pointer(Buf)^);
        //Move(Buf^, AdrStr, SIZECHAR);   //copy the address from buffer
        if AdrStr <> nil then
        begin
          Str := PChar(AdrStr);         //strings and blob are taken from the existing buffer immediately
          if FI^.DataType = vftString then
            len := strlen(Str)
          else                          //blob length is specified obviously
            len := PInteger(Buf + SIZECHAR)^;
            //Move((Buf + SIZECHAR)^, len, SIZEINT); //copy the length
          if len > 0 then
          begin                         //zero values are not written
            Result := string(Str);
            SetLength(Result, len);     //important for MEMO!!!!
          end;
        end
        else
          Result := '';
      end;
  end;
  //FreeMem(Buf);
end;

function TVolgaCursor.GetRecordCount: integer;
begin
  Result := FData.Count;
end;

function TVolgaCursor.InitFieldDefs(var Msg: string): Boolean;
var j, p, p2, ipos, i, k: integer;
  fldnam, flddesc: string;
  FI: PVolgaField;
  fldtyp: char;
begin
  ipos := 0;
  FRecSize := 0;
  Result := true;
  Msg := '';
  FFldInfo.Capacity := FFldCount;
  //reading all information about the structure of the sent FFieldLayout buffer
  for i := 0 to FFldCount - 1 do
  begin                                 //for each field
    if ipos >= FFldLen then break;      //is position less then whole header length?
    j := 0;
    for k := ipos to FFldLen - 1 do     //from current position till end of header
      if FFieldLayout[k] = FieldSeparator then
      begin
        j := k - ipos;                  //define a length of one field definition
        break;
      end;
    if j > 0 then                       //separator is found
    begin
      //description of the field in the following format: "name T" or "name Tlll"
      //or "name Tlll~default~description"
      New(FI);
      flddesc := Copy(string(FFieldLayout + ipos), 1, j); //full description of the field
      p := Pos(' ', flddesc);           //a name is from start position to first space
      fldnam := Copy(flddesc, 1, p - 1); //the full field's name
      fldtyp := flddesc[p + 1];         //field type
      p2 := Pos('~', flddesc);          //if where a description part?
      if fldtyp = 's' then              //for long strings
        if p + 1 = j then               //next symbol after field type is field separator
          FI^.Size := 100
        else
        begin                           //stringfield length if it is given
          if p2 > 0 then
            FI^.Size := StrToInt(Copy(flddesc, p + 2, p2 - p - 2))
          else
            FI^.Size := StrToInt(Copy(flddesc, p + 2, j - p - 1));
        end
      else
        FI^.Size := 0;
      p := Pos('.', fldnam);            //searching for the table name /field name separator
      FI^.Name := Copy(fldnam, p + 1, MaxInt); // Length(fldnam) - p);
      FI^.Origin := _if(p > 0, Copy(fldnam, 1, p - 1), 'XXX');
      FI^.Align := taLeftJustify;
      FI^.DefValue := '';
      FI^.Description := '';
      if p2 > 0 then                    //last part with default value and description
      begin
        flddesc := Copy(flddesc, p2 + 1, MaxInt);
        p2 := Pos('~', flddesc);
        if p2 > 0 then
        begin
          FI^.DefValue := Copy(flddesc, 1, p2 - 1);
          FI^.Description := Copy(flddesc, p2 + 1, MaxInt);
        end
        else
          FI^.Description := flddesc;
      end;
      case fldtyp of
        'b':
          begin
            FI^.DataType := vftBoolean;
            FI^.DataSize := SIZEBOOL;
          end;
        'c':
          begin
            FI^.DataType := vftCurrency;
            FI^.Align := taRightJustify;
            FI^.DataSize := SIZEDOUB;
          end;
        'd':
          begin
            FI^.DataType := vftDate;
            FI^.DataSize := SIZEINT;
          end;
        'e':
          begin
            FI^.DataType := vftLargeInt;
            FI^.Align := taRightJustify;
            FI^.DataSize := SIZEINT64;
          end;
        'f':
          begin
            FI^.DataType := vftDouble;
            FI^.Align := taRightJustify;
            FI^.DataSize := SIZEDOUB;
          end;
        'g':
          begin
            FI^.DataType := vftGraphic;
            FI^.DataSize := SIZECHAR + SIZEINT;
          end;                          //pointer+length
        'i':
          begin
            FI^.DataType := vftInteger;
            FI^.Align := taRightJustify;
            FI^.DataSize := SIZEINT;
          end;
        'a':
          begin
            FI^.DataType := vftAutoInc;
            FI^.Align := taRightJustify;
            FI^.DataSize := SIZEINT;
{$IFDEF PRO}
            if FAutoIncName = '' then
              FAutoIncName := FI^.Name; //name of Autoinc field
{$ENDIF}
          end;
        'l':
          begin
            FI^.DataType := vftSmallInt;
            FI^.Align := taRightJustify;
            FI^.DataSize := 2;
          end;
        'm':
          begin
            FI^.DataType := vftMemo;
            FI^.DataSize := SIZECHAR + SIZEINT;
          end;                          //pointer+length
        'o':
          begin
            FI^.DataType := vftBlob;
            FI^.DataSize := SIZECHAR + SIZEINT;
          end;                          //pointer+length
        'r':
          begin
            FI^.DataType := vftFmtMemo; //rtf-file
            FI^.DataSize := SIZECHAR + SIZEINT;
          end;                          //pointer+length
        's':
          begin
            FI^.DataType := vftString;  //string-fields length is 100 by default
            FI^.DataSize := SIZECHAR;
          end;
        't':
          begin
            FI^.DataType := vftDateTime;
            FI^.DataSize := SIZEDOUB;
          end;
        'v':
          begin
            FI^.DataType := vftTime;
            FI^.DataSize := SIZEINT;
          end;
      else
        begin                           //numerals: 1-9 and letters: A-Z
          FI^.DataType := vftSmallStr;
          if fldtyp <= '9' then
            FI^.Size := Ord(fldtyp) - Ord('0')
          else
            FI^.Size := Ord(fldtyp) - Ord('A') + 10; //from A to Z
          FI^.DataSize := FI^.Size + 1;
        end;                            //short strings have 0x00 in the end
      end;
      FI^.KeyValue := '';
      FI^.Offset := FRecSize;           //offset from the beginning of the record's buffer
      FFldInfo.Add(FI);
      Inc(FRecSize, FI^.DataSize);
      Inc(ipos, j + 1);                 //go to the following field
    end
    else
    begin
      Msg := V_FLDDEF;
      Result := false;
    end;
  end;
end;

function TVolgaCursor.LoadFieldInfo(Stream: TStream; var Msg: string): Boolean;
var rec_count: integer;
begin
  Result := LoadStruFromStream(Stream, Msg, rec_count);
  if not Result then Clear(true);       //clearing memory at the opening error
  FreeMem(FFieldLayout);                //because it is reading only for fields information gathering!!!
  FFieldLayout := nil;
end;

function TVolgaCursor.LoadStruFromStream(Stream: TStream;
  var Msg: string; var rec_count: integer): Boolean;
//load only the structure from the buffer
var byte_ans: integer;
  sim: char;
  buf: PChar;
begin
  Result := false;
  Msg := '';
{$IFDEF PRO}
  FCiphered := false;
{$ENDIF}
  try                                   //ReadBuffer returnes EReadError when reading error
    Stream.ReadBuffer(byte_ans, SIZEINT); //declared length of the sent buffer
    Stream.ReadBuffer(sim, 1);
    case sim of
      'E':                              //error
        begin
          buf := AllocMem(Stream.Size - SIZEINT); //for zero in the end
          Stream.ReadBuffer(buf^, Stream.Size - SIZEINT - 1);
          Msg := string(buf);           //sent report about the error
          FreeMem(buf);
          Exit;
        end;
{$IFDEF PRO}
      'P': FCiphered := true;
{$ENDIF}
      'S': ;                            //simple
      else
        begin
          Msg := V_BUFSTRU;       
          Exit;
        end;
    end;
{$IFDEF PRO}
    if FCiphered then
    begin                               //the password is entered by user or taken from the list of the passwords
      if FPasswBuf = nil then
        FPasswBuf := AllocMem(LenPassw);
      Stream.ReadBuffer(FPasswBuf^, LenPassw); //read the key from this file
      //decipher the GUID key by user's password (send ciphered, return deciphered)
      if not VolgaDB.CheckPassw(FPasswBuf, FPassword) then
      begin
        Msg := V_INVALIDPASSW;
        Exit;
      end;                              //else
    end;
    //continue if password is correct or there is no password
{$ENDIF}
    Stream.ReadBuffer(rec_count, SIZEINT); //declared record count
    Stream.ReadBuffer(FFldCount, SIZEINT); //declared field count
    Stream.ReadBuffer(FFldLen, SIZEINT);   //declared header length
    if FFldLen < 100000 then
    begin
      FFieldLayout := AllocMem(FFldLen + 1); //header field info with last delimiter
      Stream.ReadBuffer(FFieldLayout^, FFldLen);
      if InitFieldDefs(Msg) then          //create FieldDefs and calculate record length
        Result := true;
    end
    else
      Msg := V_BUFSTRU; 
  except
    Msg := V_BUFSTRU;                   //'reading the buffer error, read '+Stream.Position
    if FPartialOpen then Result := true; //continue
  end;
end;

function TVolgaCursor.LoadFromStream(Stream: TStream; var Msg: string): Boolean;
var rec_count, perc, tek, offs: integer;
  a: string;
  sim: char;
  noteslen: smallint;
  buf: PChar;
begin
  Result := false;
  perc := 0;
  if LoadStruFromStream(Stream, Msg, rec_count) then
  begin
{$IFDEF PRO}
    if FCiphered then
    begin
      //decryption of the stream before it's parsing
      offs := 4 * SIZEINT + 1 + LenPassw + FFldLen;
      a := string(FPasswBuf);
      SetLength(a, LenPassw - 2);
      VolgaDB.Sapphire.Key := a;
      VolgaDB.Sapphire.DecryptBuf(PChar(TMemoryStream(Stream).Memory) + offs,
        Stream.Size - offs);
    end;
    FIsFiltered := Assigned(FFilterClause);
{$ENDIF}
    //f there is TableNotes then extract them
    tek := FFldCount;
    if rec_count = 0 then tek := 3;     //sim for notes+noteslen
    if Stream.Position <= Stream.Size - tek then
    begin
      offs := Stream.Position;          //we are located after the table's header
      Stream.ReadBuffer(sim, 1);
      if sim = '*' then
      begin
        Stream.ReadBuffer(noteslen, 2); //the length of the notes
        buf := AllocMem(noteslen + 1);
        Stream.ReadBuffer(buf^, noteslen); //the text of the notes
        FNotes := StrPas(buf);
        FreeMem(buf);
      end
      else
        Stream.Position := offs;        //there is no notes in table
    end;
    FData.Capacity := rec_count;
    while (Stream.Position <= Stream.Size - FFldCount) and (FData.Count < rec_count) do
      if not AddNewRecord(Stream, Msg) then //add next record, shift the pointer
        break
      else
      begin
        tek := Round(FData.Count * 100 / rec_count);
        if (perc <> tek) and Assigned(FOnCallback) then
          FOnCallback(tek);             //if the percent realy changed to prevent often jerking
        perc := tek;
      end;
    if Msg = '' then
      if (FData.Count <> rec_count)
{$IFDEF PRO} and not Assigned(FFilterClause){$ENDIF} then
        Msg := Format(V_NUMREC, [FData.Count, rec_count])
      else
        Result := true;
    FData.Capacity := FData.Count;      //lear everything that did not pass through the filter
{$IFDEF PRO}
    if Result and (FAutoIncName > '') then SetLastAutoInc(0);
{$ENDIF}
    if FPartialOpen then Result := true; //even with errors
  end;
  if not Result then Clear(true);       //clearing memory at opening error
end;

{$IFDEF PRO}

procedure TVolgaCursor.SetLastAutoInc(NewValue: integer);
var pRec: PVolgaRecord;
  last: integer;
begin
  if FAutoIncName = '' then Exit;
  FAutoIncField := FindFieldInfo(FAutoIncName);
  last := 0;
  if FData.Count > 0 then
  try
    Sort(FAutoIncName, false);
    pRec := PVolgaRecord(FData.Items[FData.Count - 1]);
    last := PInteger(pRec^.Data + FAutoIncField^.Offset)^;
  except
  end;
  if NewValue > last then last := NewValue;
  FLastAutoInc := last;
end;
{$ENDIF}

function TVolgaCursor.SaveRecord(Buf: PChar; CurRec: PVolgaRecord;
  var Msg: string): Boolean;
var pRec: PVolgaRecord;
  Index: integer;
begin
  Msg := '';
  Result := false;
  if CurRec <> nil then
  begin                                 //modification of the existing record
    Index := FData.IndexOf(CurRec);
    if Index <= -1 then                 //there is no record with such pointer
    begin
      Msg := V_RECNOTFOUND;
      Exit;
    end;
    pRec := PVolgaRecord(FData.Items[Index]);
    if (pRec.Status = vusDeleted) or (pRec.Status = vusAbsent) then
    begin                               //record was removed earlier
      Msg := V_RECDELETED;
      Exit;
    end;
    if pRec^.Status = vusUnmodified then pRec^.Status := vusModified;
  end
  else                                  //for the new record CurRec := nil
  begin
    New(pRec);                          //allocate the new structure in memory
    pRec^.Status := vusInserted;
    pRec^.OldData := nil;
    pRec^.RecordData := nil;
    FData.Add(pRec);
    pRec^.Data := AllocMem(FRecSize);   //allocate the memory for data that we want to write
  end;
  CopyRecordBuf(Buf, pRec^.Data);       //allocate the sent data
{$IFDEF PRO}
  if (CurRec = nil) and (FAutoIncName > '') then
  begin
    Inc(FLastAutoInc);
    PInteger(pRec^.Data + FAutoIncField^.Offset)^ := FLastAutoInc;
  end;
{$ENDIF}
  CalcRecSize(pRec);
  Result := true;
end;

function TVolgaCursor.SaveToStream(AStream: TStream): integer;
var
  FI: PVolgaField;
  pRec: PVolgaRecord;
  fldbuf, LF, Mesto: PChar;
{$IFDEF PRO}
  bufcrypt: PChar;
{$ENDIF}
  posbefore, len, posafter, posstart: integer;
  bi, rc, i, j, perc, tek, offs: integer;
  rl: smallint;
  b64: int64;
  bl: WordBool;
  bd: double;                           //variable for determination = 0
  AdrStr: Pointer;
  sim: char;
{$IFDEF PRO}
  ClassID: {$IFDEF Linux}TGUID{$ELSE}TCLSID{$ENDIF Linux};
  strClassID: string;
{$ENDIF}
begin
{$IFDEF PRO}
  FCiphered := FPassword > '';
  if FCiphered then
    sim := 'P'
  else
{$ENDIF}
    sim := 'S';
  //writing the file's header
  perc := 0;
  posafter := 0;
  posstart := AStream.Position;
  //the record starts with the current position of the stream!!!!
  AStream.Write(posafter, SIZEINT);     //here the size of the whole buffer will be written
  AStream.Write(sim, 1);                //official symbol
{$IFDEF PRO}
  if (sim = 'P') then
  begin
    //generate the key. With that key we will cipher the file
{$IFDEF Linux}
    CreateGUID(ClassID);
{$ELSE}
    CoCreateGuid(ClassID);
{$ENDIF Linux}
    strClassID := GUIDToString(ClassID);
    //copy in buffer
    if FPasswBuf = nil then
      FPasswBuf := AllocMem(LenPassw);
    StrPCopy(FPasswBuf, strClassID);    //always a new GUID for every saving data
    //add CRC
    crcl := $FF;
    crch := $FF;
    for i := 0 to LenPassw - 3 do
      Crc(ord(FPasswBuf[i]));
    FPasswBuf[LenPassw - 2] := chr(crcl);
    FPasswBuf[LenPassw - 1] := chr(crch);
    bufcrypt := AllocMem(LenPassw);
    Move(FPasswBuf^, bufcrypt^, LenPassw);
    //encode it with the user's password
    VolgaDB.Sapphire.Key := FPassword;
    VolgaDB.Sapphire.EncryptBuf(bufcrypt, LenPassw - 2); //without CRC
    AStream.Write(bufcrypt^, LenPassw); //write the key
    FreeMem(bufcrypt);
  end;
{$ENDIF}
  AStream.Write(posafter, SIZEINT);     //here the record's count will be written
  AStream.Write(FFldCount, SIZEINT);    //fields count
  FillFieldLayout;                      //fill new field layout buffer and it's length
  //write information about the structure
  AStream.Write(FFldLen, SIZEINT);      //the length of the information about structure
  AStream.Write(FFieldLayout^, FFldLen); //information (the last separator is included)
  //note in table
  if FNotes > '' then
  begin
    sim := '*';
    AStream.Write(sim, 1);
    rl := Length(FNotes);
    fldbuf := AllocMem(rl + 1);
    StrPCopy(fldbuf, FNotes);
    AStream.Write(rl, 2);
    AStream.Write(fldbuf^, rl);
    FreeMem(fldbuf);
  end;
  //writing separate records
  rc := 0;
  LF := AllocMem(FFldCount);            //here will be the lengthes of the fields
  posafter := AStream.Position;         //we are located after the end of the structure description
  for i := 0 to FData.Count - 1 do
  begin
    pRec := PVolgaRecord(FData.Items[i]);
    if (pRec^.Status = vusAbsent) or (pRec^.Status = vusDeleted) then continue;
    Inc(rc);                            //written records calculation
    //writing the fields lengthes
    FillChar(LF^, FFldCount, $00);      //zeroize
    posbefore := AStream.Position;      //current position before the entering the next record
    AStream.Position := AStream.Position + FFldCount; //after the infomation about the length
    for j := 0 to FFldCount - 1 do
    begin
      FI := FieldInfo[j];
      Mesto := pRec^.Data + FI^.Offset;
      case FI^.DataType of
        vftSmallInt:
          begin
            rl := PSmallInt(Mesto)^;
            //Move(Mesto^, rl, 2);
            if rl <> 0 then
            begin                       //zero values are not written
              LF[j] := Chr(2);
              AStream.Write(rl, 2);
            end;
          end;
        vftInteger, vftAutoInc, vftDate, vftTime:
          begin
            bi := PInteger(Mesto)^;
            //Move(Mesto^, bi, SIZEINT);
            if bi <> 0 then
            begin                       //zero values are not written
              LF[j] := Chr(SIZEINT);
              AStream.Write(bi, SIZEINT);
            end;
          end;
        vftLargeInt:
          begin
            b64 := PInt64(Mesto)^;
            //Move(Mesto^, b64, SIZEINT64);
            if b64 <> 0 then
            begin                       //zero values are not written
              LF[j] := Chr(SIZEINT64);
              AStream.Write(b64, SIZEINT);
            end;
          end;
        vftBoolean:
          begin
            bl := PWordBool(Mesto)^;
            //Move(Mesto^, bl, SIZEBOOL);
            if bl then
            begin
              LF[j] := Chr(SIZEBOOL);
              AStream.Write(Mesto^, SIZEBOOL);
            end;
          end;
        vftDouble, vftCurrency, vftDateTime:
          begin
            bd := PDouble(Mesto)^;
            //Move(Mesto^, bd, SIZEDOUB);
            if bd <> 0 then
            begin                       //zero values are not written
              LF[j] := Chr(SIZEDOUB);
              AStream.Write(bd, SIZEDOUB);
            end;
          end;
        vftSmallStr:
          begin
            //fldbuf := AllocMem(FI^.DataSize); //buffer for the short strings copying
            //Move(Mesto^, fldbuf^, FI^.DataSize); //ncluding the closing 000
            //len := vstrlen(fldbuf);
            len := vstrlen(Mesto);
            if len > 0 then             //zero values are not written
            begin
              LF[j] := Chr(len);        //what is the real length?
              //AStream.Write(fldbuf^, len); //writing without the closing 000
              AStream.Write(Mesto^, len); //writing without the closing 000
            end;
            //FreeMem(fldbuf);
          end;
        vftString, vftMemo, vftFmtMemo, vftGraphic, vftBlob:
          begin
            AdrStr := Pointer(Pointer(Mesto)^);
            //Move(Mesto^, AdrStr, SIZECHAR); //copy in address buffer
            if AdrStr <> nil then
            begin
              fldbuf := PChar(AdrStr);  //strings and blob are taken from existing buffer immediately
              if FI^.DataType = vftString then
                len := vstrlen(fldbuf)
              else                      //blob length is specified obviously
                len := PInteger(Mesto + SIZECHAR)^;
                //Move((Mesto + SIZECHAR)^, len, SIZEINT); //copy the length
              if len > 0 then           //zero values are not written
              begin
                LF[j] := Chr(SIZEINT);  //the length of the length
                AStream.Write(len, SIZEINT); //the length of the string or blob
                AStream.Write(fldbuf^, len); //string or blob
              end;
            end;
          end;
      end;
      posafter := AStream.Position;
      AStream.Position := posbefore;
      AStream.Write(LF^, FFldCount);    //fields lengthes before the beginning of writing
      AStream.Position := posafter;
    end;
    tek := Round((i + 1) * 100 / FData.Count);
    if (perc <> tek) and Assigned(FOnCallback) then
      FOnCallback(tek);                 //if the percent realy changed to prevent often jerking
    perc := tek;
  end;
  FreeMem(LF);
  //the whole file length
  posafter := AStream.Position;         //we are located in the end of the file
{$IFDEF PRO}
  if not FCiphered then
    AStream.Position := posstart + SIZEINT + 1 //skip the official symbol
  else
    AStream.Position := posstart + SIZEINT + 1 + LenPassw; //skip the official symbol and password
{$ELSE}
  AStream.Position := posstart + SIZEINT + 1; //skip the official symbol
{$ENDIF}
  AStream.Write(rc, SIZEINT);           //real record's count
  AStream.Position := posstart;         //go to beginning
  posbefore := posafter - posstart;
  AStream.Write(posbefore, SIZEINT);    //buffer's length
{$IFDEF PRO}
  //the stream encryption after the fields description
  //!!!!!!  -        !!!!
  if FCiphered and (AStream is TMemoryStream) then
  begin
    offs := 4 * SIZEINT + 1 + LenPassw + FFldLen;
    VolgaDB.Sapphire.Key := strClassID;
    VolgaDB.Sapphire.EncryptBuf(PChar(TMemoryStream(AStream).Memory) + offs, posafter -
      offs);
  end;
{$ENDIF}
  AStream.Position := posafter;         //go to the end of the stream
{$IFDEF VOLGAT}
  ShowMessage(texta);
{$ENDIF}
  Result := posbefore;
end;

{$IFDEF Linux}

procedure CopyFile(Source, Target: string; FailIfExists: boolean); {(c) Vit, 2004, www.delphist.com}
var SourceStream, TargetStream: TFileStream;
begin
  if FailIfExists and FileExists(Target) then exit;
  SourceStream := TFileStream.create(Source, fmOpenRead);
  try
    TargetStream := TFileStream.create(Target, fmCreate or fmOpenWrite);
    try
      SourceStream.Position := 0;
      TargetStream.CopyFrom(SourceStream, SourceStream.Size);
    finally
      TargetStream.Free;
    end;
  finally
    SourceStream.Free;
  end;
end;
{$ENDIF Linux}

procedure TVolgaCursor.SaveToFile(const Fname: string; Options: TVolgaSaveOptions;
  const FieldNames: string; const IgnoreCase: Boolean);
var
  mems: TMemoryStream;
  fils: TFileStream;
  sizefil, streamsize: integer;

  function CheckBlobs: Boolean;
  var i: integer;
  begin
    Result := false;
    for i := 0 to FFldCount - 1 do
    begin
      case FieldInfo[i]^.DataType of
        vftMemo, vftFmtMemo, vftGraphic, vftBlob:
          begin
            Result := true;
            Exit;
          end;
      end;
    end;
  end;

begin
  //making backup-copy
  if savAutoBackup in Options then
    CopyFile(PChar(fname), PChar(ChangeFileExt(fname, '.~d')), false);
  //sorting data before saving
  if savSortOnIndex in Options then
    Sort(FieldNames, IgnoreCase);
  streamsize := CalcStreamSize;
  //saving when the stream is in memory
  mems := TMemoryStream.Create;
  try
    mems.SetSize(streamsize);
    sizefil := SaveToStream(mems);
    mems.Position := 0;
    fils := TFileStream.Create(fname, fmCreate);
    try
      if sizefil <> 0 then fils.WriteBuffer((mems.Memory)^, sizefil);
    finally
      fils.Free;
    end;
  finally
    mems.Free;
  end;
end;

procedure TVolgaCursor.Sort(const FieldNames: string; const IgnoreCase: Boolean);
begin
  FSortFI := TList.Create;
  IgnoreC := IgnoreCase;                //the letter's register ignoring token
  GetFldInfoList(Self, FSortFI, FieldNames); //fields list in source table
  if FSortFI.Count > 0 then
    FData.Sort(VolgaSort);
  FSortFI.Free;
  FSortFI := nil;
end;

procedure TVolgaCursor.UndoRecord(pRec: PVolgaRecord);
begin
  if pRec^.OldData <> nil then
  begin
    ClearData(pRec^.Data);
    pRec^.Data := AllocMem(FRecSize);   //allocate the memory for record's UNDO-copy's data
    CopyRecordBuf(pRec^.OldData, pRec^.Data);
    ClearData(pRec^.OldData);           //delete the record's UNDO-copy
  end;
  pRec^.Status := vusUnmodified;
  CalcRecSize(pRec);
end;

//******************************************************************************
//          TVolgaDataset - our data source
//******************************************************************************

constructor TVolgaDataset.Create(AOwner: TComponent);
var KeyIndex: TKeyIndex;
begin
  inherited;
  FSortList := TList.Create;
  FLoadedBuf := false;
  FCurRec := -1;
  FNeedToSave := 0;
  FIsRanged := false;
  FReadOnly := false;
  FRecReadOnly := false;
  FPartialOpen := false;
  FLoadSize := 0;
  FLoadTime := 0;
  FCloneList := TList.Create;
  FIsSorted := false;
{$IFDEF PRO}
  FPassword := '';
  FShowDeleted := false;
{$ENDIF}
  FDatabase := nil;
  FExternalStream := nil;
  for KeyIndex := Low(TKeyIndex) to High(TKeyIndex) do
    FKeyBuffers[KeyIndex] := nil;
  FMasterLink := TVolgaMasterLink.Create(Self);
  FMasterLink.OnMasterChange := MasterChanged;
  FMasterLink.OnMasterDisable := MasterDisabled;
{$IFDEF VOLGAT}
  Inc(FVolgaCounter);
  if (FVolgaCounter = 1) and not (csDesigning in ComponentState) then
    ShowMessage(texta);
  if (FVolgaCounter >= 21) then
    DatabaseError('Trial version restriction!'#13#13 + texta, Self);
{$ENDIF}
end;

destructor TVolgaDataset.Destroy;
begin
{$IFDEF VOLGAT}
  Dec(FVolgaCounter);
{$ENDIF}
  if Active then Close;
  FCloneList.Free;
  FCloneList := nil;
  FSortList.Free;                       //buffer with the logical records
  FSortList := nil;
{$IFDEF PRO}
  ClearFilterExpr;
{$ENDIF}
  FMasterLink.Free;
  DestroyCursor;                        //abstract
  inherited;
end;

{$IFDEF PRO}

procedure TVolgaDataset.AcceptFilter(Sender: TVolgaCursor; pRec: PVolgaRecord;
  var Accept: Boolean);
//load this record in cursor?
var j, compar: integer;
  FC: PVolgaCField;
  FI, FI2: PVolgaField;
  buf: PChar;
  ignorecase: Boolean;

  function CompareDataTypes(t1, t2: TVolgaFieldType): Boolean;
  begin
    Result := false;
    if t1 = t2 then
      Result := true
    else
      case t1 of
        vftInteger, vftAutoInc, vftDate, vftTime:
          if (t2 = vftInteger) or (t2 = vftAutoInc) or (t2 = vftDate) or (t2 = vftTime) then
            Result := true;
        vftDouble, vftCurrency, vftDateTime:
          if (t2 = vftDouble) or (t2 = vftCurrency) or (t2 = vftDateTime) then
            Result := true;
      end;
  end;

begin
  //AcceptFilterList is equal to either FFilterFI or FWhereList
  Accept := true;
  ignorecase := foCaseInsensitive in FilterOptions;
  for j := 0 to AcceptFilterList.Count - 1 do //let's think that all expressions are bound with AND
  begin
    FC := AcceptFilterList.Items[j];    //for the filter we think that the table's names of every fields are equal
//  if (FC^.Name2 > '') and (FC^.Alias <> FC.Alias2) then Continue; //it is JOIN!!!!!!!
    FI := Sender.FindFieldInfo(FC^.Name); //first field
    if (FC^.CompText = '') and (FC^.Name2 > '') then
      FI2 := Sender.FindFieldInfo(FC^.Name2) //fields of the one table are compared with each other
    else
      FI2 := nil;
    if FI <> nil then
    begin
      if FC^.Clause = vcLike then
        Accept := FC^.Maska.Matches(Sender.GetFieldText(FI, pRec, false, false))
      else
      begin
        buf := nil;
        if FI2 <> nil then
          if CompareDataTypes(FI2^.DataType, FI^.DataType) then
            buf := pRec^.Data + FI2^.Offset //the second field's buffer
          else  //todo: convert value to needed datatype using strings
            DatabaseError(V_TYPESNOTEQUAL, Self)
        else
        begin
          buf := AllocMem(FI^.DataSize);
          AssignBufFromStr(buf, FC^.CompText, FI); //moving the string in buffer for the comparison
        end;
        compar := CompareBuf(pRec^.Data + FI^.Offset, buf, FI, ignorecase); //compare the buffers
        case FC^.Clause of
          vcEqual: Accept := compar = 0;
          vcGreat: Accept := compar > 0;
          vcLess: Accept := compar < 0;
          vcGreatEqual: Accept := compar >= 0;
          vcLessEqual: Accept := compar <= 0;
          vcNotEqual: Accept := compar <> 0;
        end;
        if FI2 = nil then
        begin
          if FI^.DataType in [vftString, vftMemo, vftFmtMemo] then
          KillBuf(buf);
          FreeMem(buf);
        end;
      end;
      if not Accept then break;
    end
    else                                //the field is not found
    begin
      Accept := false;
      Break;
    end;
  end;
end;
{$ENDIF}

function TVolgaDataset.AcceptRange(pRec: PVolgaRecord): Boolean;
var range1, range2: integer;
begin
  range1 := CompareRec(pRec^.Data, FKeyBuffers[kiRangeStart]^.Data,
    FKeyBuffers[kiRangeStart]^.KeyFI, FKeyBuffers[kiRangeStart]^.FieldCount);
  range2 := CompareRec(pRec^.Data, FKeyBuffers[kiRangeEnd]^.Data,
    FKeyBuffers[kiRangeEnd]^.KeyFI, FKeyBuffers[kiRangeEnd]^.FieldCount);
  Result := (range1 > 0) or (not FKeyBuffers[kiRangeStart]^.Exclusive and (range1 = 0));
  if Result then
    Result := (range2 < 0) or (not FKeyBuffers[kiRangeEnd]^.Exclusive and (range2 = 0));
end;

procedure TVolgaDataset.AddClone(DS: TVolgaDataset);
begin
  if FCloneList.IndexOf(DS) < 0 then
    FCloneList.Add(DS);
  DS.FCursor := Self.FCursor;
{$IFDEF VOLGAT}
  ShowMessage(texta);
{$ENDIF}
end;

procedure TVolgaDataset.AllocKeyBuffers;
var KeyIndex: TKeyIndex;
begin
  try
    for KeyIndex := Low(TKeyIndex) to High(TKeyIndex) do
    begin
      New(FKeyBuffers[KeyIndex]);
      FKeyBuffers[KeyIndex]^.Kind := KeyIndex;
      FKeyBuffers[KeyIndex]^.KeyFI := TList.Create;
      FKeyBuffers[KeyIndex]^.Fields := TList.Create;
      FKeyBuffers[KeyIndex]^.FieldCount := 0;
      FKeyBuffers[KeyIndex]^.Data := AllocMem(FRecordSize);
      InitKeyBuffer(KeyIndex);
    end;
  except
    FreeKeyBuffers;
    raise;
  end;
  FIsRanged := false;                   //when index changed range is cancelled
end;

function TVolgaDataset.AllocRecordBuffer: PChar;
begin
  Result := AllocMem(FRecBufSize);
end;

procedure TVolgaDataset.ApplyRange;
begin
  CheckCursorOpen;
  if Active then SetState(dsBrowse);
  if SetCursorRange and Active then First;
end;

procedure TVolgaDataset.ApplyUpdates;
begin
  if State in dsEditModes then Post;
  InternalApplyUpdates;                 //it is necessary to call the CancelUpdates in Exception's handler
end;

function TVolgaDataset.BookmarkValid(Bookmark: TBookmark): Boolean;
begin
  //we send the bookmark's address
  Result := FSortList.IndexOf(PVolgaBookmark(Bookmark)^.pToRec) >= 0;
end;

procedure TVolgaDataset.CancelRange;
begin
  CheckCursorOpen;
  if Active then SetState(dsBrowse);
  UpdateCursorPos;
  if ResetCursorRange and Active then Resync([]);
end;

procedure TVolgaDataset.CancelUpdates;
begin
  if Active and NeedToSave then
  begin
    if State in dsEditModes then Cancel;
    FCursor.CancelUpdates;              //physically cancel all changes in buffer
{$IFDEF PRO}
    FShowDeleted := false;
{$ENDIF}
    SortRecords;
    FNeedToSave := 0;
    if FDatabase <> nil then
      FDatabase.DeleteUpdated(Self);
    if Assigned(FOnNeed) then FOnNeed(Self);
  end;
end;

procedure TVolgaDataset.CommitUpdates;
begin
  if Active and NeedToSave then
  begin
    FCursor.CommitUpdates;              //physically commit all changes to buffer
{$IFDEF PRO}
    FShowDeleted := false;
{$ENDIF}
    SortRecords;                        //building the logical sequence of the records again
    FNeedToSave := 0;
    if FDatabase <> nil then
      FDatabase.DeleteUpdated(Self);
    if Assigned(FOnNeed) then FOnNeed(Self);
  end;
end;

procedure TVolgaDataset.ClearCalcFields(Buffer: PChar);
begin
  FillChar((Buffer + FRecordSize)^, CalcFieldsSize, $00);
end;

procedure TVolgaDataset.CheckCursorOpen;
begin
  if not IsCursorOpen then DatabaseError(V_DATASETCLOSED, Self);
end;

{$IFDEF PRO}

function TVolgaDataset.CheckDuplicate(const FieldName: string = ''): Boolean;
var RI: PVolgaRecInfo;
begin
  CheckActive;
  RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
  Result := FCursor.CheckDuplicate(ActiveBuffer, PVolgaRecord(RI^.Bookmark), FieldName);
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaDataset.CheckFilterFields;
//make sure that there are fields from the request in the real table
var i: integer;
  F: PVolgaCField;
begin
  for i := 0 to FFilterFI.Count - 1 do
  begin
    F := PVolgaCField(FFilterFI.Items[i]); //we take the field
    case F^.TokenType of
      vttIdent, vttQuoted:
        if FCursor.FindFieldInfo(F^.Name) = nil then
          DatabaseErrorFmt('Field %s is not found', [F^.Name], Self);
    end;
    case F^.TokenType2 of
      vttIdent, vttQuoted:
        if (FCursor.FindFieldInfo(F^.Name2) = nil) then
          if F^.TokenType2 = vttIdent then
            DatabaseErrorFmt('Field %s is not found', [F^.Name2], Self)
          else
          begin
            F^.CompText := F^.Name2;
            //create the mask for the condition of the Like comparison
            if F^.Clause = vcLike then
              F^.Maska := TVolgaMask.Create(F^.CompText);
          end;
      vttNumber:
        F^.CompText := F^.Name2;
    end;
  end;
end;
{$ENDIF}

procedure TVolgaDataset.CheckMasterRange;
var s1: string;
begin
  if FMasterLink.Active and (FMasterLink.Fields.Count > 0) then
  begin
    //update Range values
    s1 :=
      TVolgaDataset(FMasterLink.DataSource.DataSet).GetTabValues(FMasterLink.FieldNames);
    SetRange(s1, s1);                   //apply the new Range
  end;
end;

procedure TVolgaDataset.CheckSetKeyMode;
begin
  if State <> dsSetKey then
    DatabaseError(V_NOTSETKEYSTATE, Self);
end;

{$IFDEF PRO}

procedure TVolgaDataset.ClearFilterExpr;
var i: integer;
  F: PVolgaCField;
begin
  if FFilterFI <> nil then
  begin
    for i := 0 to FFilterFI.Count - 1 do
    begin
      F := PVolgaCField(FFilterFI.Items[i]);
      Dispose(F);
    end;
    FFilterFI.Free;
  end;
  FFilterFI := nil;
end;

{$ENDIF}

function TVolgaDataset.CompareBookmarks(Bookmark1,
  Bookmark2: TBookmark): Integer;
begin
  if (Bookmark1 = nil) and (Bookmark2 = nil) then
    Result := 0                         //both bookmarks are empty
  else if Bookmark1 = nil then
    Result := -1                        //only first is empty
  else if Bookmark2 = nil then
    Result := 1                         //only second is empty
  else if longint(PVolgaBookmark(Bookmark1)^.pToRec) < //both are not empty
    longint(PVolgaBookmark(Bookmark2)^.pToRec) then
    Result := -1
  else if longint(PVolgaBookmark(Bookmark1)^.pToRec) >
    longint(PVolgaBookmark(Bookmark2)^.pToRec) then
    Result := 1
  else
    Result := 0;
end;

procedure TVolgaDataset.CopyRecord;
//copy currenr record in new with the OnNewRecord settings
var buf, actbuf: PChar;
  i: integer;
  FI: PVolgaField;
  F: TField;
begin
  CheckCursorOpen;                      //if not opened, then exception
  CheckBrowseMode;
  buf := PVolgaRecord(FSortList.Items[RecNo - 1])^.Data;
  Append;                               //OnNewRecord is called
  for i := 0 to FCursor.FieldCount - 1 do
  begin
    FI := FCursor.FieldInfo[i];
    actbuf := ActiveBuffer + FI^.Offset;
    //if the field is not the unique key
    if Pos(FI^.Name, UniqueFieldNames) = 0 then
    begin
      F := FindField(FI^.Name);
      if F <> nil then
      begin
        if FI^.DataType = vftString then
          CopyLongString(buf + FI^.Offset, actbuf)
        else if FI^.DataType in [vftMemo, vftFmtMemo, vftGraphic, vftBlob] then
          CopyBlobField(buf + FI^.Offset, actbuf)
        else
          Move((buf + FI^.Offset)^, actbuf^, FI^.DataSize);
        DataEvent(deFieldChange, longint(F));
      end;
    end;
  end;
end;

function TVolgaDataset.CreateBlobStream(Field: TField;
  Mode: TBlobStreamMode): TStream;
begin
  Result := TVolgaBlobStream.Create(TBlobField(Field), Mode);
end;

procedure TVolgaDataset.DoBeforeDelete;
begin
  if RecReadOnly or ReadOnly then
    DatabaseError(V_RECREADONLYDEL, Self);
  inherited;
end;

procedure TVolgaDataset.DoBeforeEdit;
var RI: PVolgaRecInfo;
begin
  if RecReadOnly or ReadOnly then
    DatabaseError(V_RECREADONLYCHAN, Self);
  inherited;
  RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
  if not FCursor.CreateUndoCopy(PVolgaRecord(RI^.Bookmark)) then Abort; //UNDO-copy creation was not successful
  Inc(FNeedToSave);
  if (FNeedToSave = 1) and Assigned(FOnNeed) then FOnNeed(Self);
  if FDatabase <> nil then
    FDatabase.AddUpdated(Self);
end;

procedure TVolgaDataset.DoBeforeInsert;
begin
  if ReadOnly then
    DatabaseError(V_ADDREADONLY, Self);
  inherited;
  Inc(FNeedToSave);
  if (FNeedToSave = 1) and Assigned(FOnNeed) then FOnNeed(Self);
  if FDatabase <> nil then
    FDatabase.AddUpdated(Self);
end;

procedure TVolgaDataset.DoCallback(const Percent: integer);
begin
  if Assigned(FOnCallback) then FOnCallback(Self, Percent);
end;

{$IFDEF PRO}

procedure TVolgaDataset.DoDestroyRec(var RecordData: Pointer);
begin
  if Assigned(FOnDestroy) then FOnDestroy(Self, RecordData);
end;
{$ENDIF}

procedure TVolgaDataset.DoOnNewRecord;
var de: string;
  F: TField;
  FI: PVolgaField;
  i: integer;
  DetailList: TList;
begin
  for i := 0 to FieldCount - 1 do
  begin
    F := Fields[i];
    de := F.DefaultExpression;
    FI := FCursor.FindFieldInfo(F.FieldName);
    //default value from volga table
    if (FI <> nil) and (de = '') and (FI^.DefValue > '') then
      de := FI^.DefValue;
    if (F.FieldNo > 0) and (de > '') then //the field is not Calculated and it has the default value
    begin
      if (de = 'TODAY') and (F is TDateField) then
        F.AsDateTime := Date            //today!
      else if (de = 'NOW') and (F is TDateTimeField) then
        F.AsDateTime := Now             //now!
      else if (de = 'NOW') and (F is TTimeField) then
        F.AsDateTime := Time            //now!
      else if (de = 'UNICOD') and (F is TStringField) then
        F.AsString := GetUnicod         //unique 8 bytes code
      else
        F.AsString := de;
    end;
  end;
  if FMasterLink.Active and (FMasterLink.Fields.Count > 0) then
  begin
    DetailList := TList.Create;
    GetFieldList(DetailList, RangeFieldNames);
    for I := 0 to FMasterLink.Fields.Count - 1 do
      TField(DetailList[i]).Value := TField(FMasterLink.Fields[I]).Value;
    DetailList.Free;
  end;
  inherited DoOnNewRecord;
end;

procedure TVolgaDataset.EditKey;
begin
  SetKeyBuffer(kiIndex, False);
end;

procedure TVolgaDataset.EditRangeEnd;
begin
  SetKeyBuffer(kiRangeEnd, False);
end;

procedure TVolgaDataset.EditRangeStart;
begin
  SetKeyBuffer(kiRangeStart, False);
end;

procedure TVolgaDataset.EmptyTable;
begin
  CheckActive;                          //if not opened, then exception
  DisableControls;
  FCursor.Clear(false);                 //the info about the fields is not deleted
  FCurRec := -1;
  FSortList.Clear;
  ClearBuffers;
  First;
  EnableControls;
//  Inc(FNeedToSave); //it is necesary to save the changes
//  if (FNeedToSave = 1) and Assigned(FOnNeed) then FOnNeed(Self);
//  if FDatabase <> nil then
//    FDatabase.AddUpdated(Self);
end;

function TVolgaDataset.ExistField(const Fn: string): Boolean;
begin
  Result := FindField(Fn) <> nil;
end;

{$IFDEF PRO}

procedure TVolgaDataset.ExportToDBF(const FileName: string; ToOEM: Boolean);
//Export in DBF format
var DBFC: TVolgaDBFConverter;
  num: integer;
begin
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  CheckActive;                          //if not opened, then exception
  DisableControls;
  num := RecNo;
  DBFC := TVolgaDBFConverter.Create(Self);
  DBFC.ToOEM := ToOEM;
  DBFC.Dataset := Self;
  if Assigned(FOnCallback) then
    DBFC.OnCallback := DoCallback;
  DBFC.DBFToFile(FileName);
  DBFC.Free;
  RecNo := num;
  EnableControls;
{$ENDIF}
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaDataset.ExportToTxt(const FileName: string; const Delimiter: char = #9);
var i, j, num: integer;
  s: string;
  A: TStrings;
begin                                   //export in text file
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  CheckActive;                          //if not opened, then exception
  num := RecNo;
  A := TStringList.Create;
  DisableControls;
  s := '';
  for j := 0 to FieldCount - 1 do
  begin
    if not Fields[j].IsBlob then
      if Delimiter = #9 then
        s := s + Fields[j].DisplayLabel + Delimiter
      else
        s := s + AnsiQuotedStr(Fields[j].DisplayLabel, '"') + Delimiter;
  end;
  SetLength(s, Length(s) - 1);
  A.Add(s);
  First;
  for i := 0 to RecordCount - 1 do
  begin
    s := '';
    for j := 0 to FieldCount - 1 do
    begin
      if not Fields[j].IsBlob then
        if Delimiter = #9 then
          s := s + Fields[j].AsString + Delimiter
        else
          s := s + AnsiQuotedStr(Fields[j].AsString, '"') + Delimiter;
    end;
    SetLength(s, Length(s) - 1);
    A.Add(s);
    Next;
  end;
  A.SaveToFile(FileName);
  A.Free;
  RecNo := num;
  EnableControls;
{$ENDIF}
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaDataset.ExportToXML(const FileName, Encoding: string);
//Export in XML format
var XMLC: TVolgaXMLConverter;
  num: integer;
begin
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  CheckActive;                          //if not opened, then exception
  DisableControls;
  num := RecNo;
  XMLC := TVolgaXMLConverter.Create(Self);
  XMLC.Encoding := Encoding;
  XMLC.Origin := Origin;
  XMLC.Dataset := Self;
  if Assigned(FOnCallback) then
    XMLC.OnCallback := DoCallback;
  XMLC.XMLToFile(FileName);
  XMLC.Free;
  RecNo := num;
  EnableControls;
{$ENDIF}
end;
{$ENDIF}

{$IFDEF PRO}
{$IFDEF Linux}

procedure TVolgaDataset.ExportToExcel(const Fn: string);
begin
  raise Exception.Create('Capability is not supported!');
end;
{$ELSE}

procedure TVolgaDataset.ExportToExcel(const FileName: string);
//Export in XLS (Excel) format
var ExcelC: TVolgaXLSConverter;
  num: integer;
begin
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  CheckActive;                          //if not opened, then exception
  DisableControls;
  num := RecNo;
  ExcelC := TVolgaXLSConverter.Create(Self);
  ExcelC.Dataset := Self;
  if Assigned(FOnCallback) then
    ExcelC.OnCallback := DoCallback;
  ExcelC.XLSToFile(FileName);
  ExcelC.Free;
  RecNo := num;
  EnableControls;
{$ENDIF}
end;
{$ENDIF Linux}
{$ENDIF}

{$IFDEF PRO}
//--------- used in FindFirst, FindNext, FindPrior, FindLast where filter is active

function TVolgaDataset.FindRecord(Restart, GoForward: Boolean): Boolean;
var IsFound: Boolean; index: integer;
  OldState: TDataSetState;

  function SearchRec(start: integer): integer;
  var pRec: PVolgaRecord;
  begin
    Result := start;
    IsFound := false;
    if (Result < 0) or (Result >= RecordCount) then Exit;
    while true do
    begin
      IsFound := true;
      pRec := PVolgaRecord(FSortList.Items[Result]);
      if Filter > '' then               //filter as expression
        AcceptFilter(FCursor, pRec, IsFound);
      if IsFound and Assigned(OnFilterRecord) then //filter as procedure
      begin
        FFiltRec := pRec;               //record, that is tested in filter
        OnFilterRecord(Self, IsFound);
      end;
      if IsFound then Exit;
      if GoForward then
        Inc(Result)
      else
        Dec(Result);
      if GoForward and (Result >= RecordCount) then Exit; //Eof
      if not GoForward and (Result < 0) then Exit; //Bof;
    end;
  end;

begin
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  try
    if Filtered then
    begin
      if GoForward then
        if Restart then
          First
        else
          Next
      else if Restart then
        Last
      else
        Prior;
      IsFound := (RecordCount > 0) and not Bof and not Eof;
    end
    else
    begin
      if Filter > '' then               //filter as expression
      begin
        AcceptFilterList := FFilterFI;  //fields list for the filter
        CheckFilterFields;
      end;
      if GoForward then
        if Restart then
          index := 0
        else
          index := FCurRec + 1
      else if Restart then
        index := RecordCount - 1
      else
        index := FCurRec - 1;
      OldState := SetTempState(dsFilter);
      try
        index := SearchRec(index);
      finally
        RestoreState(OldState);
      end;
      if IsFound then FCurRec := index;
    end;
  finally
  end;
  if IsFound then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end;
  Result := Found;
  if Result then DoAfterScroll;
end;
{$ENDIF}

function TVolgaDataset.FastSeek(const Fn: string): Boolean;
//it is possible to use quick searching for the sorted list
var p: integer;
begin
  Result := false;
  p := Pos(UpperCase(fn), UpperCase(IndexFieldNames));
  if (p = 1) and ((Length(fn) = Length(IndexFieldNames)) or
    (IndexFieldNames[Length(fn) + 1] = ';')) then
    //when editing records are not sorted till SortRecords calling
    Result := not (vsoCaseInsensitive in SortOptions) and FIsSorted;
end;

function TVolgaDataset.Find(const KeyFields: string;
  const KeyValues: array of const; const FromFirstRec: Boolean): Boolean;
var start: integer;
begin
  Result := false;
  CheckBrowseMode;
  DoBeforeScroll;
  CursorPosChanged;
  InitKeyBuffer(kiSearch, KeyFields);
  SetKeyFields(kiSearch, KeyValues);    //set current search buffer and fill with key values
  //from what record should we start the searching? - from the first or after the current (current RecNo-1)
  start := InternalFind(_if(FromFirstRec, 0, RecNo), FastSeek(KeyFields));
  if start >= 0 then
  begin
    Result := true;
    FCurRec := start;
    Resync([rmCenter]);
    DoAfterScroll;
  end;
end;

function TVolgaDataset.FindKey(const KeyValues: array of const): Boolean;
begin
  CheckBrowseMode;
  InitKeyBuffer(kiIndex);
  SetKeyFields(kiIndex, KeyValues);
  Result := GotoKey;
end;

{$IFDEF PRO}

procedure TVolgaDataset.FindNearest(const KeyValues: array of const);
begin
  CheckBrowseMode;
  InitKeyBuffer(kiIndex);
  SetKeyFields(kiIndex, KeyValues);
  GotoNearest;
end;
{$ENDIF}

function TVolgaDataset.FindTab(const KeyFields, KeyValues: string;
  const FromFirstRec: Boolean): Boolean;
var start: integer;
begin
  Result := false;
  CheckBrowseMode;
  DoBeforeScroll;
  CursorPosChanged;
  InitKeyBuffer(kiSearch, KeyFields);
  SetKeyFields(kiSearch, KeyValues);    //set current search buffer and fill with key values
  start := InternalFind(_if(FromFirstRec, 0, RecNo), FastSeek(KeyFields));
  if start >= 0 then
  begin
    Result := true;
    FCurRec := start;
    Resync([rmCenter]);
    DoAfterScroll;
  end;
end;

procedure TVolgaDataset.FreeKeyBuffers;
var KeyIndex: TKeyIndex;
begin
  for KeyIndex := Low(TKeyIndex) to High(TKeyIndex) do
    if FKeyBuffers[KeyIndex] <> nil then
    begin
      if FCursor <> nil then
        FCursor.ClearData(FKeyBuffers[KeyIndex]^.Data)
      else
        FreeMem(FKeyBuffers[KeyIndex]^.Data);
      FKeyBuffers[KeyIndex]^.KeyFI.Free;
      FKeyBuffers[KeyIndex]^.Fields.Free;
      Dispose(FKeyBuffers[KeyIndex]);
      FKeyBuffers[KeyIndex] := nil;
    end;
end;

procedure TVolgaDataset.FreeRecordBuffer(var Buffer: PChar);
begin
  if Buffer <> nil then
  begin
    FCursor.ClearBuf(Buffer);
    FreeMem(Buffer);
    Buffer := nil;
  end;
end;

function TVolgaDataset.GetActiveRecBuf(var RecBuf: PChar): Boolean;
var pRec: PVolgaRecord;
  RI: PVolgaRecInfo;
begin
  //determination of the buffer, with which we will work
  RecBuf := nil;
  case State of
    dsBrowse:
      if not IsEmpty then RecBuf := ActiveBuffer;
    dsEdit, dsInsert, dsNewValue: RecBuf := ActiveBuffer; //current record's buffer
    dsSetKey: RecBuf := FKeyBuffer^.Data; //key fields buffer
    dsCalcFields: RecBuf := CalcBuffer; //Calculated fields buffer
    dsFilter: RecBuf := FFiltRec.Data;  //filter's buffer
    dsOldValue:
      begin
        RecBuf := ActiveBuffer;
{$IFDEF PRO}
        RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize); //current or new record
        pRec := PVolgaRecord(RI^.Bookmark);
        if (pRec <> nil) and (pRec^.OldData <> nil) then
          RecBuf := pRec^.OldData;
{$ENDIF}
      end;
  end;
  Result := RecBuf <> nil;
end;

procedure TVolgaDataset.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Move((Buffer + FRecordSize + CalcFieldsSize)^, Data^, BookmarkSize);
end;

function TVolgaDataset.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
var RI: PVolgaRecInfo;
begin
  RI := PVolgaRecInfo(Buffer + FRecordSize + CalcFieldsSize);
  Result := RI^.BookmarkFlag;
end;

function TVolgaDataset.GetCanModify: Boolean;
begin
  Result := inherited GetCanModify and not ReadOnly and not RecReadOnly;
end;

function TVolgaDataset.GetCloneCount: integer;
begin
  if FCloneList <> nil then
    Result := FCloneList.Count
  else
    Result := 0;
end;

function TVolgaDataset.GetDataSource: TDataSource;
begin                                   //return the MasterSource
  Result := FMasterLink.DataSource;
end;

function TVolgaDataset.GetDefaultValue(const FieldName: string): string;
var FI: PVolgaField;
begin
  CheckCursorOpen;
  Result := '';
  FI := FCursor.FindFieldInfo(FieldName);
  if FI <> nil then
    Result := FI^.DefValue;
end;

function TVolgaDataset.GetDescription(const FieldName: string): string;
var FI: PVolgaField;
begin
  CheckCursorOpen;
  Result := '';
  FI := FCursor.FindFieldInfo(FieldName);
  if FI <> nil then
    Result := FI^.Description;
end;

function TVolgaDataset.GetFieldData(FieldNo: Integer;
  Buffer: Pointer): Boolean;
begin
  Result := GetFieldData(FieldByNumber(FieldNo), Buffer);
end;

function TVolgaDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  Offset, lenbuf: integer;
  RecBuf, Str: PChar;
  AdrStr: Pointer;
begin
  //determination of the buffer, with which we will work
  Result := GetActiveRecBuf(RecBuf);
  if not IsCursorOpen or not Result then Exit;

  if (Buffer <> nil) then FillChar(Buffer^, Field.DataSize, $00); //clear the buffer-destination
  if Field.FieldNo > 0 then             //real (not Calculated) field
  begin                                 // Check for a nil value
    Result := (RecordCount > 0) or (State = dsInsert)
      or (State = dsFilter);
    if Result then
    begin
      Offset := GetFieldOffset(Field.FieldNo); //offset of the field in record
      if Field.DataType in [ftMemo, ftFmtMemo, ftGraphic, ftBlob] then
      begin
        //BLOB-fields have their own way of calling when reading, here only the NULL check
        AdrStr := Pointer(Pointer(RecBuf + Offset)^);
        lenbuf := PInteger(RecBuf + Offset + SIZECHAR)^;
        //Move((RecBuf + Offset + SIZECHAR)^, lenbuf, SIZEINT); //copy the length
        //Move((RecBuf + Offset)^, AdrStr, SIZECHAR); //copy the address
        Result := (AdrStr <> nil) and (lenbuf > 0);
      end
      else if (Field.DataType = ftString) and (Field.Size > MAXSLEN) then //long strings greater when 35 bytes
      begin                             //dsMaxStringSize := 8192;
        AdrStr := Pointer(Pointer(RecBuf + Offset)^);
        //Move((RecBuf + Offset)^, AdrStr, SIZECHAR); //copy in address buffer
        if AdrStr <> nil then
        begin
          Str := PChar(AdrStr);
          if vstrlen(Str) > Word(Field.DataSize) then
            lenbuf := Field.DataSize
          else
            lenbuf := vstrlen(Str);
          if Buffer <> nil then Move(Str^, Buffer^, lenbuf);
        end
        else
          Result := false;              //else  the field's buffer contains zeroes
      end                               //continue for the short strings...
      else
      begin                             //0 in numeric fields is not NULL!!
        Result := not IsBufEmpty(RecBuf + Offset, Field.DataSize) or
          (Field.DataType in [ftBoolean, ftTime, ftSmallInt, ftInteger, ftAutoInc,
          ftFloat, ftCurrency]);
        if Result and (Buffer <> nil) then //fill buffer if the destination and source are not nil
          Move((RecBuf + Offset)^, Buffer^, Field.DataSize) //copy into field's buffer
      end;
    end;
  end
  else
  begin                                 //a calculated field
    Offset := FRecordSize + Field.Offset;
    Result := Boolean(RecBuf[Offset]);  //zero byte of the calc-field shows whether the field is empty
    if Result and (Buffer <> nil) then
      Move((RecBuf + Offset + 1)^, Buffer^, Field.DataSize);
  end;
end;

function TVolgaDataset.GetFieldOffset(Num: integer): integer;
begin
  //FieldNo - starts from the one, but index FFldInfo - from zero!!
  Result := FCursor.FieldInfo[Num - 1]^.Offset;
end;

function TVolgaDataset.GetIndexField(Index: Integer): TField;
begin
  Result := nil;
  if Active and ((Index > 0) or (Index >= FKeyBuffers[kiIndex]^.Fields.Count)) then
    Result := TField(FKeyBuffers[kiIndex]^.Fields.Items[Index])
  else
    DatabaseError(V_INDEXOUTOFRANGE, Self);
end;

function TVolgaDataset.GetIndexFieldCount: Integer;
begin
  if Active then
    Result := FKeyBuffers[kiIndex]^.Fields.Count
  else
    Result := 0;
end;

function TVolgaDataset.GetKeyExclusive: Boolean;
begin
  CheckSetKeyMode;
  Result := FKeyBuffer^.Exclusive;
end;

function TVolgaDataset.GetKeyFieldCount: Integer;
begin
  CheckSetKeyMode;
  Result := FKeyBuffer^.FieldCount;
end;

{$IFDEF PRO}

function TVolgaDataset.GetListRecValues(const FieldName: string;
  const ch: char): string;
//list of values from all records, that satisfy the condition)
var
  pRec: PVolgaRecord;
  OldState: TDataSetState;
  i: integer;
  Accept: Boolean;
begin
  Result := '';
  if FieldName = '' then Exit;
  if FindField(FieldName) = nil then Exit;
  if Assigned(OnFindRecord) then        //if procedure of the record's selection is assigned
  begin
    OldState := SetTempState(dsFilter);
    try
      for i := 0 to FCursor.RecordCount - 1 do
      begin
        pRec := FCursor.RecData[i];
        if (pRec^.Status = vusAbsent) or (pRec^.Status = vusDeleted) then continue;
        Accept := true;
        FFiltRec := pRec;               //tested record
        OnFindRecord(Self, Accept);
        if Accept then Result := Result + FieldByName(FieldName).AsString + ch;
      end;
    finally
      RestoreState(OldState);
    end;
  end;
end;
{$ENDIF}

function TVolgaDataset.GetMasterFields: string;
begin
  Result := FMasterLink.FieldNames;
end;

function TVolgaDataset.GetNeedToSave: Boolean;
begin                                   //it is necessary save data?
  Result := FNeedToSave > 0;
end;

{$IFDEF PRO}

function TVolgaDataset.GetPassword: string;
begin
  if Active then
    Result := FCursor.Password
  else
    Result := FPassword;
  if Result <> FPassword then FPassword := Result;
end;
{$ENDIF}

function TVolgaDataset.GetRangeFieldCount: Integer;
begin
  if Active then
    Result := FKeyBuffers[kiRangeStart]^.Fields.Count
  else
    Result := 0;
end;

function TVolgaDataset.GetRecNo: Integer;
var
  BufPtr: PChar;
  RI: PVolgaRecInfo;
begin
  CheckActive;
  Result := -1;
  if State = dsInternalCalc then
    Exit
  else
  begin
    if State = dsCalcFields then
      BufPtr := CalcBuffer
    else
      BufPtr := ActiveBuffer;
    RI := PVolgaRecInfo(BufPtr + FRecordSize + CalcFieldsSize);
    Result := RI^.RecordNum;
  end;
end;

function TVolgaDataset.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var RI: PVolgaRecInfo;
begin
  if RecordCount < 1 then
    Result := grEOF
  else
  begin
    Result := grOK;
    case GetMode of
      gmNext:                           //go to next record
        if FCurRec >= RecordCount - 1 then
          Result := grEOF
        else
          Inc(FCurRec);
      gmPrior:                          //go to previous record
        if FCurRec <= 0 then
          Result := grBOF
        else
          Dec(FCurRec);
      gmCurrent:                        //take the current record
        if (FCurRec < 0) or (FCurRec >= RecordCount) then
          Result := grError;
    end;
    if Result = grOK then
    begin                               //copy from the internal buffer in the record's buffer
      FCursor.CopyRecordBuf(PVolgaRecord(FSortList.Items[FCurRec])^.Data, Buffer);
      RI := PVolgaRecInfo(Buffer + FRecordSize + CalcFieldsSize);
      RI^.BookmarkFlag := bfCurrent;
      RI^.Bookmark := PVolgaRecord(FSortList.Items[FCurRec]); //pointer on our record
      RI^.RecordNum := FCurRec + 1;     //record's number
      GetCalcFields(Buffer);            //calculating the values of the Calculated fields!!!
    end
    else if (Result = grError) and DoCheck then
      DatabaseError(V_RECZERO, Self);
  end;
end;

function TVolgaDataset.GetRecordCount: Integer;
begin
  Result := FSortList.Count;
end;

{$IFDEF PRO}

function TVolgaDataset.GetRecordData: Pointer;
begin
  Result := nil;
  if Active and (RecordCount > 0) then
    Result := PVolgaRecord(FSortList.Items[RecNo - 1])^.RecordData;
end;
{$ENDIF}

function TVolgaDataset.GetRecordSize: Word;
begin
  Result := FRecordSize;                //the size of the record without official additives
end;

function TVolgaDataset.GetStreamSize: integer;
begin
  CheckActive;                          //if not opened, then exception
  Result := FCursor.CalcStreamSize;
end;

function TVolgaDataset.GetSum(const FieldName: string): Extended;
var ss1: Extended;
  OldState: TDataSetState;
  i: integer;
begin
  CheckActive;                          //if not opened, then exception
  Result := 0;
  if (FieldName = '') or not ExistField(FieldName) or
    not (FindField(FieldName) is TNumericField) then Exit;

  ss1 := 0;
  OldState := SetTempState(dsFilter);
  try
    for i := 0 to RecordCount - 1 do
    begin
      FFiltRec := PVolgaRecord(FSortList.Items[i]); //record, from which we take the value
      ss1 := ss1 + FieldByName(FieldName).AsFloat; //trying to convert into number
    end;
  finally
    RestoreState(OldState);
  end;
  Result := ss1;
end;

function TVolgaDataset.GetTabValues(const KeyFields: string): string;
var KL: TList;
  j: integer;
begin
  CheckActive;                          //if not opened, then exception
  KL := TList.Create;
  GetFieldList(KL, KeyFields);                 //fields list
  Result := '';
  for j := 0 to KL.Count - 1 do
    Result := Result + TField(KL.Items[j]).AsString + Tabulator;
  KL.Free;
end;

function TVolgaDataset.GetVersion: string;
begin
  Result := 'VolgaDB Engine v' + _version;
end;

procedure TVolgaDataset.GotoCurrent(DS: TVolgaDataset);
var BM: TBookmark;
begin
  CheckCursorOpen;                      //if not opened, then exception
  if not DS.Active or not Active then Exit;
  CheckBrowseMode;
  DS.CheckBrowseMode;
  //another way of searching for clones
  if FCursor = DS.FCursor then
  begin
    BM := DS.GetBookmark;
    GotoBookmark(BM);
    DS.FreeBookmark(BM);
  end
  else
  begin
    if AnsiCompareText(UniqueFieldNames, DS.UniqueFieldNames) <> 0 then
      DatabaseError(V_KEYNOTEQUAL, Self);
    FindTab(UniqueFieldNames, DS.GetTabValues(UniqueFieldNames)); //go to the necessary record
  end;
end;

function TVolgaDataset.GotoKey: Boolean;
var start: integer;
begin
  if State = dsSetKey then SetState(dsBrowse);
  CheckBrowseMode;
  DoBeforeScroll;
  CursorPosChanged;
  FKeyBuffer := FKeyBuffers[kiIndex];
  start := InternalFind(0, FastSeek(IndexFieldNames)); //we must search in RecBuffer!!!!
  if start >= 0 then
  begin
    Result := true;
    FCurRec := start;
    Resync([rmCenter]);
    DoAfterScroll;
  end
  else
    Result := false;
end;

{$IFDEF PRO}

procedure TVolgaDataset.GotoNearest;
var start: integer;
begin
  if State = dsSetKey then SetState(dsBrowse);
  CheckBrowseMode;
  CursorPosChanged;
  FKeyBuffer := FKeyBuffers[kiIndex];
  start := InternalFindNearest(0);
  if start >= 0 then
  begin
    FCurRec := start;
    Resync([rmCenter]);
    DoAfterScroll;
  end;
end;
{$ENDIF}

{$IFDEF PRO}

function TVolgaDataset.InternalFindNearest(const istart: integer): integer;
var start, H, I, C: integer;
  desc: Boolean;
begin
  Result := -1;
  start := istart;
  if not FastSeek(IndexFieldNames) then SortRecords;
  desc := vsoDescendSort in SortOptions;

  H := FSortList.Count - 1;
  while start <= H do
  begin
    I := (start + H) shr 1;
    C := CompareRec(PVolgaRecord(FSortList.Items[I])^.Data, FKeyBuffer^.Data,
      FKeyBuffer^.KeyFI, FKeyBuffer^.FieldCount);
    if ((C < 0) and not desc) or ((C > 0) and desc) then
      start := I + 1
    else
    begin
      H := I - 1;
      if ((C >= 0) and not desc) or ((C <= 0) and desc) then
      //if C = 0 then //we stay in cycle for the searching the 1-st of the many similar
        Result := I;
    end;
  end;
end;
{$ENDIF}

function TVolgaDataset.InitKeyBuffer(KeyIndex: TKeyIndex; FieldNames: string = ''):
PKeyBuffer;
var Buffer: PKeyBuffer;
  i: integer;
begin
  Buffer := FKeyBuffers[KeyIndex];
  if IsCursorOpen then
    FCursor.ClearBuf(Buffer^.Data);     //todo:check
  FillChar((Buffer^.Data)^, FRecordSize, 0);
  case Buffer^.Kind of
    kiIndex: FieldNames := FIndexFieldNames;
    //kiSearch: - FieldNames
    kiRangeStart, kiRangeEnd: FieldNames := FRangeFieldNames;
  end;
  if (FieldNames > '') and IsCursorOpen then
  begin
    Buffer^.KeyFI.Clear;
    Buffer^.Fields.Clear;
    GetFieldList(Buffer^.Fields, FieldNames); //Dataset-fields list
    GetFldInfoList(FCursor, Buffer^.KeyFI, FieldNames); //Cursor-fields list
    if Buffer^.Fields.Count <> Buffer^.KeyFI.Count then
      DatabaseError(V_CALCORBLOB, Self);
    for i := 0 to Buffer^.KeyFI.Count - 1 do //justification (it is important for strings!)
    try
      PVolgaField(Buffer^.KeyFI.Items[i])^.Align :=
        TField(Buffer^.Fields.Items[i]).Alignment;
    except;
    end;
  end;
  Buffer^.FieldCount := Buffer^.Fields.Count;
  Buffer^.Exclusive := false;
  Result := Buffer;
end;

procedure TVolgaDataset.InsertRec(Buf: PChar);
var Msg: string;
  R: PVolgaRecord;
  RI: PVolgaRecInfo;
begin
{$IFDEF PRO}
  if CheckKeyViolation and (UniqueFieldNames <> '') and
    FCursor.CheckDuplicate(Buf, nil, UniqueFieldNames) then
    DatabaseError(V_ERRSAVENEW + V_KEYVIOLATION, Self);
{$ENDIF}
  if not FCursor.SaveRecord(Buf, nil, Msg) then //here we save the new record in the cursor!!!
    DatabaseError(V_ERRSAVENEW + Msg, Self);
  //add the new record to list
  R := FCursor.RecData[FCursor.RecordCount - 1]; //the last added record's address
  FSortList.Add(R);
  //write tokens into buffer
  RI := PVolgaRecInfo(Buf + FRecordSize + CalcFieldsSize);
  RI^.Bookmark := Pointer(R);           //pointer on our record
  RI^.BookmarkFlag := bfCurrent;
  RI^.RecordNum := FSortList.Count;     //record's number
  FCurRec := FSortList.Count - 1;
end;

procedure TVolgaDataset.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  if Append then InternalLast;          //go to the end, otherwise insert in middle
  InsertRec(PChar(Buffer));
end;

procedure TVolgaDataset.InternalCancel;
var RI: PVolgaRecInfo;
begin
  if State = dsEdit then
  begin
    RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
    FCursor.UndoRecord(PVolgaRecord(RI^.Bookmark)); //eturn the old field values before all changes
  end;
  inherited;
  Dec(FNeedToSave);
  if (FNeedToSave = 0) then
  begin
    if Assigned(FOnNeed) then FOnNeed(Self);
    if FDatabase <> nil then
      FDatabase.DeleteUpdated(Self);    //there are no changes
  end;
end;

procedure TVolgaDataset.InternalClose;
begin
  if CloneCount > 0 then
    DatabaseError(V_CLOSECLON, Self);
  //clear all chosenned buffers for ranges
  if FCursor <> nil then
    FreeKeyBuffers;
  BindFields(false);
  if DefaultFields then DestroyFields;
  FCurRec := -1;
  FLoadSize := 0;
  FLoadTime := 0;
  FSortList.Clear;
  FLoadedBuf := false;
  if FDatabase <> nil then
  begin
    FDatabase.DeleteUpdated(Self);      //for the safety
    FDatabase.Remove(Self);             //selfdestraction
  end;
end;

procedure TVolgaDataset.InternalDelete;
var Msg: string;
  RI: PVolgaRecInfo;
begin
{$IFDEF PRO}
  FShowDeleted := false;
{$ENDIF}
  RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
  if not FCursor.DeleteRecord(PVolgaRecord(RI^.Bookmark), Msg) then
    DatabaseError(V_ERRDELETE + Msg, Self);
  FSortList.Delete(FCurRec);
  Inc(FNeedToSave);                     //the changes saving is necessary
  if (FNeedToSave = 1) and Assigned(FOnNeed) then FOnNeed(Self);
  if FDatabase <> nil then
    FDatabase.AddUpdated(Self);
end;

function TVolgaDataset.InternalFind(const istart: integer; Sorted: Boolean): integer;
var start, H, I, C: integer;
  IsFound, desc: Boolean;
begin
  Result := -1;
  start := istart;
  desc := vsoDescendSort in SortOptions;
  if Sorted then
  begin                                 //searching in the sorted list
    H := FSortList.Count - 1;
    while start <= H do
    begin
      I := (start + H) shr 1;
      C := CompareRec(PVolgaRecord(FSortList.Items[I])^.Data, FKeyBuffer^.Data,
        FKeyBuffer^.KeyFI, FKeyBuffer^.FieldCount);
      if ((C < 0) and not desc) or ((C > 0) and desc) then
        start := I + 1
      else
      begin
        H := I - 1;
        if C = 0 then                   //we stay in cycle for the searching the 1-st of the many similar
          Result := I;
      end;
    end;
  end
  else
  begin                                 //when the list is unsorted - linear search
    IsFound := false;
    while not IsFound and (start < FSortList.Count) do
    begin
      IsFound := CompareRec(FKeyBuffer^.Data, PVolgaRecord(FSortList.Items[start])^.Data,
        FKeyBuffer^.KeyFI, FKeyBuffer^.FieldCount) = 0;
      if not IsFound then
        Inc(start)
      else
      begin
        Result := start;
        break;
      end;
    end;
  end;
end;

procedure TVolgaDataset.InternalFirst;
begin
  FCurRec := -1;
end;

procedure TVolgaDataset.InternalGotoBookmark(Bookmark: Pointer);
var Index: integer;
begin
  //we send the bookmark's address
  Index := FSortList.IndexOf(PVolgaBookmark(Bookmark)^.pToRec);
  if Index <> -1 then
    FCurRec := Index;
end;

procedure TVolgaDataset.InternalHandleException;
begin
  Application.HandleException(Self);
end;

procedure TVolgaDataset.InternalInitFieldDefs;
var FI: PVolgaField;
  i: integer;
begin
  FRecordSize := 0;
  try
    FieldDefs.BeginUpdate;
    if FieldDefs.Count > 0 then FieldDefs.Clear; //clear the old fields descriptions
    if IsCursorOpen then
    begin
      //reading all information about the structure of the sent FFieldLayout buffer
      for i := 0 to FCursor.FieldCount - 1 do
      begin                             //for every field
        FI := FCursor.FieldInfo[i];
        FieldDefs.Add(FI^.Name, UsingFieldType[FI^.DataType], FI^.Size, false);
        Inc(FRecordSize, FI^.DataSize); //8 bytes for blob!!!(the pointer and the length)
      end;
    end;
  finally
    FieldDefs.EndUpdate;
  end;
end;

procedure TVolgaDataset.InternalInitRecord(Buffer: PChar);
var RI: PVolgaRecInfo;
begin
  if IsCursorOpen then FCursor.ClearBuf(Buffer); //todo:check
  FillChar(Buffer^, FRecBufSize, $00);
  //here we can insert DefaultValues
  RI := PVolgaRecInfo(Buffer + FRecordSize + CalcFieldsSize);
  RI^.BookmarkFlag := bfInserted;
  RI^.RecordNum := -1;                  //record's number
end;

procedure TVolgaDataset.InternalLast;
begin
  FCurRec := FSortList.Count;           //pointer after the last dataset record
end;

procedure TVolgaDataset.InternalLoad(Stream: TStream; Keep: Boolean = false);
var Msg: string;
  Suss: Boolean;
begin
  if FCursor = nil then
  begin
    FCursor := TVolgaCursor.Create;
{$IFDEF PRO}
    FCursor.Password := FPassword;
    FCursor.OnDestroyRec := DoDestroyRec;
{$ENDIF}
    if Assigned(FOnCallback) then
      FCursor.OnCallback := DoCallback;
  end
  else
    FCursor.Clear(true);
  FCursor.PartialOpen := PartialOpen;   //opening token, even when errors
  if FSortList.Count > 0 then FSortList.Clear;
  FLoadSize := Stream.Size;
  if not Keep then
    Stream.Position := 0;               //from the very beginning of the stream
  //else we are located in the external stream's position, specified by user
  Suss := FCursor.LoadFromStream(Stream, Msg); //load from the buffer
  //always clear the buffer after the loading (or failure)
  if not Keep then
    Stream.Free;                        //clear the memory for the TMemoryStream, closes the file for TFileStream
  //else FExternalStream := nil; //separate our variable from the external stream
  if not Suss then
  begin
    FCurRec := -1;
    FCursor.Free;
    FCursor := nil;
    DatabaseError(V_OPENTABLE + Msg, Self);
  end
  else if (Msg > '') and PartialOpen then
    ShowMessage(Msg);
end;

function TVolgaDataset.InternalLocate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions; const FromStart: Boolean = true):
Boolean;
var fld_cnt, start, j: integer;
  OldState: TDataSetState;
  val, fld: Variant;
  FI: PVolgaField;
  notisnull: Boolean;
begin
  CheckBrowseMode;
  CursorPosChanged;
  Result := false;
  if (RecordCount = 0) or (KeyFields = '') or VarIsEmpty(KeyValues) or
    VarIsNull(KeyValues) then Exit;

  InitKeyBuffer(kiSearch, KeyFields);
  FKeyBuffer := FKeyBuffers[kiSearch];
  if Options = [] then
  begin
    fld_cnt := FKeyBuffer^.KeyFI.Count - 1;
    try
      if not VarIsArray(KeyValues) then
      begin
        FI := PVolgaField(FKeyBuffer^.KeyFI.Items[0]);
        AssignBufFromStr(FKeyBuffer^.Data + FI^.Offset, VarAsType(KeyValues, varString),
          FI);
      end
      else
      begin
        if fld_cnt > VarArrayHighBound(KeyValues, 1) then
          fld_cnt := VarArrayHighBound(KeyValues, 1);
        for j := 0 to fld_cnt do
        begin
          notisnull := not (VarIsNull(KeyValues[j]) or VarIsEmpty(KeyValues[j]));
          if notisnull then
          begin
            FI := PVolgaField(FKeyBuffer^.KeyFI.Items[j]);
            AssignBufFromStr(FKeyBuffer^.Data + FI^.Offset, VarAsType(KeyValues[j],
              varString), FI);
          end;
        end;
      end;
      //from what record should we start the searching? - from the first or after the current (current RecNo-1)
      start := InternalFind(_if(FromStart, 0, RecNo), FastSeek(KeyFields));
      if start >= 0 then
      begin
        Result := true;
        FCurRec := start;
      end;
    finally
    end
  end
  else
  begin
    OldState := SetTempState(dsFilter);
    try
      fld_cnt := FKeyBuffer^.Fields.Count - 1;
      start := _if(FromStart, 0, RecNo); //from the first or after the current
      while not Result and (start < RecordCount) do
      begin
        j := 0;
        Result := true;
        FFiltRec := PVolgaRecord(FSortList.Items[start]); //the record, from which we take the value
        while Result and (j <= fld_cnt) do
        begin
          if VarIsArray(KeyValues) then
            val := KeyValues[j]
          else
            val := KeyValues;           //Do we have the array of values for searching?
          fld := TField(FKeyBuffer^.Fields.Items[j]).Value; //value of the field
          Result := not (VarIsNull(val) or VarIsNull(fld));
          if Result then
          begin
            fld := VarAsType(fld, VarType(val)); //convert to the necessary type
            if TField(FKeyBuffer^.Fields.Items[j]).DataType = ftString then
            begin
              if loCaseInsensitive in Options then //only for string!
              begin
                fld := AnsiUpperCase(fld);
                val := AnsiUpperCase(val);
              end;
              if loPartialKey in Options then
                Result := Result and (Pos(val, fld) = 1)
              else
                Result := Result and (val = fld);
            end
            else
              Result := Result and (val = fld);
          end;
          Inc(j);
        end;
        if not Result then
          Inc(start)
        else
          Break;                        //found!!!
      end;
    finally
      RestoreState(OldState);
    end;
    if Result then
      FCurRec := start;
  end;
end;

{$IFDEF Linux}

function GetTickCount: Cardinal;        {(c) Vit, 2004, www.delphist.com}
begin                                   {we have no idea regarding time of starting linus, lets use regular time}
  Result := trunc(Now * 86400000);      {24*60*60*1000}
end;
{$ENDIF Linux}

procedure TVolgaDataset.InternalOpen;
var FF: TFloatField;
  i: integer;
  tic: Cardinal;
begin
  FCurRec := -1;
  FLoadedBuf := false;
  tic := GetTickCount;

  InternalLoadData;                     //abstract

  FieldDefs.Updated := false;
  FieldDefs.Update;                     //Initialize the FieldDefs ???????
  //Compute the offsets for different parts of the record's buffer
  BookmarkSize := sizeof(Pointer);      //bookmark buffer size
  if DefaultFields then                 //"DefaultFields" mean FieldCount is 0
  begin
    CreateFields;                       //Create TField components when no persistent fields have been created
{$IFDEF PRO}
    SetOriginToFields;                  //assign the origin to all fields if we have it
{$ENDIF}
  end;
  BindFields(true);                     //Bind the TField components to the physical fields
  AllocKeyBuffers;

  //enter displayformat for float-fields
  for i := 0 to FieldCount - 1 do
    if (Fields[i] is TFloatField) and not TFloatField(Fields[i]).Currency then
    begin
      FF := TFloatField(Fields[i]);
      if FF.DisplayFormat = '' then FF.DisplayFormat := '0.######';
    end;

  FRecBufSize := FRecordSize + CalcFieldsSize + sizeof(TVolgaRecInfo); //whole buffer's size
  CheckMasterRange;                     //if master-detail is specified, that set the range
  if FCursor.RecordCount > 0 then
  begin
    SortRecords;                        //sorting by the key fields of the prepared cursor
    FCurRec := -1;
  end;
  FLoadedBuf := true;
  FLoadTime := GetTickCount - tic;      //how long did the loading last
  if FDatabase <> nil then
    FDatabase.Add(Self);                //adding self to session
end;

procedure TVolgaDataset.InternalPost;
var keyfldschanged: Boolean;
  pRec: PVolgaRecord;
  RI: PVolgaRecInfo;
begin
  //write records from buffer to our internal buffer
  if State = dsEdit then
  begin
    keyfldschanged := true;
    if FIndexFieldNames > '' then
    begin
      RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
      pRec := PVolgaRecord(RI^.Bookmark);
      if (pRec <> nil) and (pRec^.OldData <> nil) then
        keyfldschanged := (CompareRec(ActiveBuffer, pRec^.OldData,
          FKeyBuffers[kiIndex]^.KeyFI, FKeyBuffers[kiIndex]^.FieldCount) <> 0);
    end;
    ModifyRec(ActiveBuffer);
    //compare new and old values of index fields
  end
  else
  begin
    keyfldschanged := false;
    InsertRec(ActiveBuffer);
    MoveInsertedToOrder;
  end;
  FIsSorted := FIsSorted and not keyfldschanged;
end;

procedure TVolgaDataset.InternalRefresh;
var keystr: string;

  procedure FindRefresh;
  var
    FI: PVolgaField;
    iPos, i: integer;
  begin
    InitKeyBuffer(kiSearch, UniqueFieldNames);
    FKeyBuffer := FKeyBuffers[kiSearch];
    iPos := 1;
    i := 0;
    try
      while (iPos <= Length(keystr)) and (i < FKeyBuffer^.KeyFI.Count) do
      begin                             //key fields values
        FI := PVolgaField(FKeyBuffer^.KeyFI.Items[i]);
        AssignBufFromStr(FKeyBuffer^.Data + FI^.Offset, ExtractValue(keystr, iPos), FI);
        Inc(i);
      end;
      FCurRec := InternalFind(0, FastSeek(UniqueFieldNames));
    finally
    end;
  end;

begin                                   //update the data, without calling Close and Open
  //remember the key fields of the current record
  keystr := GetTabValues(UniqueFieldNames);
  FLoadedBuf := false;                  //data aren't loaded
  DoBeforeScroll;
  if NeedToSave then ApplyUpdates;
{$IFDEF PRO}
  FShowDeleted := false;
{$ENDIF}
  InternalLoadData;
  CheckMasterRange;                     //if master-detail is specified, that set the range
  InitKeyBuffer(kiIndex);
  if FCursor.RecordCount > 0 then
    SortRecords;
  FLoadedBuf := true;                   //data loaded
  if keystr > '' then
    FindRefresh;                        //find "the previous current" record
  DoAfterScroll;
end;

procedure TVolgaDataset.InternalSetToRecord(Buffer: PChar);
begin
  InternalGotoBookmark(Pointer(Buffer + FRecordSize + CalcFieldsSize));
end;

{$IFDEF PRO}

function TVolgaDataset.IsChanged(const FieldName: string): Boolean;
begin
  Result := true;
  if not ExistField(FieldName) or (FSortList.Count = 0) then Exit;
  Result := FieldByName(FieldName).Value <> FieldByName(FieldName).OldValue; //for State=dsOldValue!!!!!
end;
{$ENDIF}

function TVolgaDataset.IsCursorOpen: Boolean;
begin
  Result := (FCursor <> nil) and (FCursor.FieldCount > 0);
end;

{$IFDEF PRO}
function TVolgaDataset.IsPasswordProtected: Boolean;
begin
  Result := (FCursor <> nil) and FCursor.FCiphered; 
end;
{$ENDIF}

function TVolgaDataset.Like(const KeyFields, KeyValues: string; AtStartPos,
  FromFirstRec: Boolean): Boolean;
var kv, val: string;
  LikeFI: TList;
  Vals: TStringList;
  FI: PVolgaField;
  IsFound: Boolean;
  start, i, j: integer;
begin
  CheckActive;                          //if not opened, then exception
  Result := false;
  if (KeyFields = '') or (KeyValues = '') then Exit;
  LikeFI := TList.Create;
  Vals := TStringList.Create;
  GetFldInfoList(FCursor, LikeFI, KeyFields);
  Split(KeyValues, Tabulator, Vals);
  if (LikeFI.Count > 0) and (LikeFI.Count = Vals.Count) then
  begin
    //from what record should we start the searching? - from the first or after the current (current RecNo-1)
    start := _if(FromFirstRec, 0, RecNo);
    for i := start to RecordCount - 1 do
    begin
      IsFound := true;
      for j := 0 to LikeFI.Count - 1 do
      begin
        FI := LikeFI.Items[j];
        val := AnsiUpperCase(Vals[j]);
        kv := FCursor.GetFieldText(FI, PVolgaRecord(FSortList.Items[i]), false, false);
        kv := AnsiUpperCase(kv);
        if AtStartPos then
          IsFound := IsFound and (Pos(val, kv) = 1)
        else
          IsFound := IsFound and (Pos(val, kv) > 0);      //we have found on the 1-st or any position?
      end;
      if IsFound then                     //we have found the field with such context
      begin
        DoBeforeScroll;
        FCurRec := i;
        Resync([rmCenter]);
        DoAfterScroll;
        Result := true;
        Break;
      end;
    end;                                  //if we haven't found, the cursor's position does not change
  end;
  LikeFI.Free;
  Vals.Free;
end;

function TVolgaDataset.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  CheckActive;                          //if not opened, then exception
  DoBeforeScroll;
  Result := InternalLocate(KeyFields, KeyValues, Options); //searching from the beginning
  if Result then
  begin
    Resync([rmExact, rmCenter]);
    DoAfterScroll;
  end;
end;

function TVolgaDataset.Lookup(const KeyFields: string;
  const KeyValues: Variant; const ResultFields: string): Variant;
var IsFound: Boolean;
  OldCurRec: integer;
  Buf: PChar;
begin
  CheckActive;                          //if not opened, then exception
  Result := NULL;
  OldCurRec := FCurRec;
  IsFound := InternalLocate(KeyFields, KeyValues, []);
  if IsFound then
  begin
    SetTempState(dsCalcFields);
    try
      Buf := TempBuffer;
      FCursor.CopyRecordBuf(PVolgaRecord(FSortList.Items[FCurRec])^.Data, Buf);
      CalculateFields(Buf);
      Result := FieldValues[ResultFields]; //what fields from the found record do we need?
    finally
      RestoreState(dsBrowse);
      FCurRec := OldCurRec;
    end;
  end;
end;

procedure TVolgaDataset.MasterChanged(Sender: TObject);
var s1: string;
begin
  if not Active then Exit;
  CheckBrowseMode;                      //the Browse test
  //update the Range values
  s1 :=
    TVolgaDataset(FMasterLink.DataSource.DataSet).GetTabValues(FMasterLink.FieldNames);
  SetRange(s1, s1);                     //apply new Range
end;

procedure TVolgaDataset.MasterDisabled(Sender: TObject);
begin
  if not Active then Exit;
  CancelRange;                          //delete the Range
end;

procedure TVolgaDataset.ModifyRec(Buf: PChar);
var Msg: string;
  RI: PVolgaRecInfo;
begin
  RI := PVolgaRecInfo(Buf + FRecordSize + CalcFieldsSize);
{$IFDEF PRO}
  if CheckKeyViolation and (UniqueFieldNames <> '') and
    FCursor.CheckDuplicate(Buf, PVolgaRecord(RI^.Bookmark), UniqueFieldNames) then
    DatabaseError(V_ERRSAVE + V_KEYVIOLATION, Self);
{$ENDIF}
  if not FCursor.SaveRecord(Buf, PVolgaRecord(RI^.Bookmark), Msg) then
    DatabaseError(V_ERRSAVE + Msg, Self);
end;

procedure TVolgaDataset.MoveInsertedToOrder;
//move a new record to it's sort order position
var I, start, finded, H, C: integer;
  KeyFI: TList;
  desc: Boolean;
  Buf: PChar;
begin
  if (IndexFieldNames = '') or not FIsSorted then Exit;
  if not (vsoInsertInOrder in SortOptions) then Exit;
  KeyFI := TList.Create;
  GetFldInfoList(FCursor, KeyFI, IndexFieldNames);
  try
    desc := vsoDescendSort in SortOptions;
    H := FSortList.Count - 2;  //   
    start := 0;
    finded := -1;
    Buf := PVolgaRecord(FSortList.Items[FSortList.Count - 1])^.Data;
    while start <= H do
    begin
      I := (start + H) shr 1;
      C := CompareRec(PVolgaRecord(FSortList.Items[I])^.Data, Buf, KeyFI);
      if ((C < 0) and not desc) or ((C > 0) and desc) then
        start := I + 1
      else
      begin
        H := I - 1;
        if ((C >= 0) and not desc) or ((C <= 0) and desc) then
          finded := I;
      end;
    end;
    if (finded >= 0) then
    begin  //     
      FSortList.Move(FSortList.Count - 1, finded);
      FCurRec := finded;
    end;
  finally
    KeyFI.Free;
  end;
end;

procedure TVolgaDataset.RemoveClone(DS: TVolgaDataset);
begin
  FCloneList.Remove(DS);
  DS.FCursor := nil;                    //do not delete the cursor, but delete the reference to it
end;

function TVolgaDataset.ResetCursorRange: Boolean;
begin
  Result := true;
  FIsRanged := false;
  if IsCursorOpen and Active then SortRecords;
end;

{$IFDEF PRO}

procedure TVolgaDataset.RevertRecord;
var RI: PVolgaRecInfo;
  pRec: PVolgaRecord;
begin
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  CheckActive;
  if State in dsEditModes then
    Cancel
  else
  begin
    RI := PVolgaRecInfo(ActiveBuffer + FRecordSize + CalcFieldsSize);
    pRec := PVolgaRecord(RI^.Bookmark);
    if (pRec <> nil) and ((pRec^.Status = vusDeleted)
      or (pRec^.Status = vusModified)) then
    begin
      UpdateCursorPos;
      FCursor.UndoRecord(pRec);         //return the old, unchanged fields values
      Dec(FNeedToSave);
      if (FNeedToSave = 0) then
      begin
        if Assigned(FOnNeed) then FOnNeed(Self);
        if FDatabase <> nil then
          FDatabase.DeleteUpdated(Self); //there are no more changes
      end;
      Resync([]);
    end;
  end;
{$ENDIF}
end;
{$ENDIF}

procedure TVolgaDataset.SaveToFile(const FileName: string);
begin
  if FCursor = nil then Exit;           //if not opened, then exception
  FCursor.SaveToFile(FileName, SaveOptions, IndexFieldNames, vsoCaseInsensitive in
    SortOptions);
end;

{$IFDEF PRO}

function TVolgaDataset.SaveToStream(AStream: TStream): integer;
begin
  Result := 0;
  if FCursor = nil then Exit;           //if not opened, then exception
  Result := FCursor.SaveToStream(AStream); //writing from the current position of the stream!!!!
  //after the writing we are located in the end of data in stream
end;
{$ENDIF}

procedure TVolgaDataset.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Move(Data^, (Buffer + FRecordSize + CalcFieldsSize)^, BookmarkSize);
end;

procedure TVolgaDataset.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
var RI: PVolgaRecInfo;
begin
  RI := PVolgaRecInfo(Buffer + FRecordSize + CalcFieldsSize);
  RI^.BookmarkFlag := Value;
end;

function TVolgaDataset.SetCursorRange: Boolean;
begin
  Result := true;
  FIsRanged := true;
  if not Active then Exit;
  SortRecords(true);                    //recreation with the new range
end;

procedure TVolgaDataset.SetDatabase(const Value: TVolgaDatabase);
begin
  if Value <> FDatabase then
  begin
    Close;
    FDatabase := Value;
  end;
end;

procedure TVolgaDataset.SetDataSource(Value: TDataSource);
begin                                   //set MasterSource
  if IsLinkedTo(Value) then             //Checks that a dataset is linked to a specified data source
    DatabaseError(V_CIRCULARLINK, Self);
  FMasterLink.DataSource := Value;
  if (Value <> nil) and (FMasterLink.DataSource.DataSet <> nil) and
    not (FMasterLink.DataSource.DataSet is TVolgaDataSet) then
    DatabaseError(V_NONVOLGADATASET, Self);
end;

procedure TVolgaDataset.SetDefaultValue(const FieldName, value: string);
var FI: PVolgaField;
begin
  CheckCursorOpen;
  FI := FCursor.FindFieldInfo(FieldName);
  if FI <> nil then
  begin
    FI^.DefValue := value;
    Inc(FNeedToSave);                   //the table must be saved
  end;
end;

procedure TVolgaDataset.SetDescription(const FieldName, descr: string);
var FI: PVolgaField;
begin
  CheckCursorOpen;
  FI := FCursor.FindFieldInfo(FieldName);
  if FI <> nil then
  begin
    FI^.Description := descr;
    Inc(FNeedToSave);                   //the table must be saved
  end;
end;

procedure TVolgaDataset.SetFieldData(Field: TField; Buffer: Pointer);
var lenbuf, Offset, DataSize: integer;
  AdrStr: Pointer;
  RecBuf, Str: PChar;
begin
  if not (State in dsWriteModes) then
    DatabaseError(V_NOTEDITMODE, Self);

  //determination of the buffer, with which we willl work
  GetActiveRecBuf(RecBuf);
  if Field.FieldNo > 0 then             //'real' fields (not calculated)
  begin
    // Cannot set fields while in OnCalcFields handler
    if State = dsCalcFields then DatabaseError(V_CALCEDIT, Self);
    //the read only field editing attempt
    if Field.ReadOnly and not (State in [dsSetKey, dsFilter]) then
      DatabaseErrorFmt(V_READONLYFLD, [Field.DisplayName], Self);
    if Field is TAutoIncField and not (State in [dsSetKey, dsFilter]) then
      Exit;
    Field.Validate(Buffer);

    Offset := GetFieldOffset(Field.FieldNo);
    DataSize := Field.DataSize;

    if (Field.DataType = ftString) and (Field.Size > MAXSLEN) then //BLOB-fields have their own way of calling !!!!!!!!!!!!!!
    begin
      KillBuf(RecBuf + Offset);         //clear the old value
      if Buffer <> nil then
        lenbuf := vstrlen(PChar(Buffer))
      else
        lenbuf := 0;
      if (Buffer = nil) or (lenbuf = 0) then
        FillChar((RecBuf + Offset)^, SIZECHAR, $00) //new string is empty or nil
      else
      begin
        Str := AllocMem(lenbuf + 1);    //allocate the memory fro the new string, which length is lenbuf
        Move(Buffer^, Str^, lenbuf);
        AdrStr := Pointer(Str);
        Pointer(Pointer(RecBuf + Offset)^) := AdrStr;
        //Move(AdrStr, (RecBuf + Offset)^, SIZECHAR); //the string address  in the field's buffer
      end;
    end                                 //for short strings and other fields
    else if Buffer = nil then           //field := nil, clear it
      FillChar((RecBuf + Offset)^, DataSize, $00)
    else                                //carrying data from the field buffer in the record buffer with the corresponding offset
      Move(Buffer^, (RecBuf + Offset)^, DataSize);
  end
  else
  begin                                 // a calculated or lookup field
    Inc(RecBuf, FRecordSize + Field.Offset);
    Boolean(RecBuf[0]) := LongBool(Buffer);
    if Boolean(RecBuf[0]) then Move(Buffer^, RecBuf[1], Field.DataSize);
  end;
  if not (State in [dsCalcFields, dsSetKey, dsFilter]) then
    DataEvent(deFieldChange, longint(Field));
end;

procedure TVolgaDataset.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    inherited;
    SortRecords;                        //apply filter
    First;
  end
  else
    inherited;
end;

{$IFDEF PRO}

procedure TVolgaDataset.SetFilterText(const Value: string);
//new filter-expression
begin
  ClearFilterExpr;                      //lear FFilterFI
  //call the parser for the filter
  if Value <> '' then
    with TVolgaParser.Create do
    begin
      try
        ParseText(Value);               //if there were no errors in parser then continue
        FFilterFI := TList.Create;
        PrepareFilterExpr(FFilterFI, 0); //if there were no errors in parser then continue
      finally
        Free;                           //free parser
      end;
    end;
  inherited SetFilterText(Value);
  if Active then
  begin
    CheckBrowseMode;
    SortRecords;                        //put or remove the filter
  end;
end;
{$ENDIF}

procedure TVolgaDataset.SetIndexField(Index: Integer; Value: TField);
begin
  GetIndexField(Index).Assign(Value);
end;

procedure TVolgaDataset.SetIndexFieldNames(const value: string);
begin
  if FIndexFieldNames = value then Exit;
  FIndexFieldNames := value;
  if Active then
  begin
    InitKeyBuffer(kiIndex);
    SortRecords;                        //dataset resorting
  end;
end;

procedure TVolgaDataset.SetKey;
begin
  SetKeyBuffer(kiIndex, True);
end;

procedure TVolgaDataset.SetKeyBuffer(KeyIndex: TKeyIndex; Clear: Boolean);
begin
  CheckBrowseMode;
  if Clear then InitKeyBuffer(KeyIndex);
  FKeyBuffer := FKeyBuffers[KeyIndex];
  SetState(dsSetKey);
  DataEvent(deDataSetChange, 0);
end;

procedure TVolgaDataset.SetKeyExclusive(Value: Boolean);
begin
  CheckSetKeyMode;
  FKeyBuffer.Exclusive := Value;
end;

procedure TVolgaDataset.SetKeyFieldCount(Value: Integer);
begin
  CheckSetKeyMode;
  FKeyBuffer^.FieldCount := Value;
end;

procedure TVolgaDataset.SetKeyFields(KeyIndex: TKeyIndex;
  const Values: array of const);
var
  j, last: Integer;
  SaveState: TDataSetState;
  FI: PVolgaField;
begin
  SaveState := SetTempState(dsSetKey);
  try
    FKeyBuffer := FKeyBuffers[KeyIndex]; //InitKeyBuffer(KeyIndex);
    last := FKeyBuffer^.FieldCount - 1;
    if last > High(Values) then last := High(Values);
    for j := 0 to last do
    begin                               //copy from  the value of the key in the range buffer immediately
      FI := PVolgaField(FKeyBuffer^.KeyFI.Items[j]);
      AssignBufFromVar(FKeyBuffer^.Data + FI^.Offset, Values[j], FI);
    end;
    FKeyBuffer^.FieldCount := High(Values) + 1;
  finally
    RestoreState(SaveState);
  end;
end;

procedure TVolgaDataset.SetKeyFields(KeyIndex: TKeyIndex; const Values: string);
var
  i, p: Integer;
  SaveState: TDataSetState;
  sv, fv: string;
  FI: PVolgaField;
begin
  SaveState := SetTempState(dsSetKey);
  try
    FKeyBuffer := InitKeyBuffer(KeyIndex);
    sv := Values;
    i := 0;
    while (i < FKeyBuffer^.KeyFI.Count) and (Length(sv) > 0) do
    begin
      FI := PVolgaField(FKeyBuffer^.KeyFI.Items[i]); //the field, in that we are searching
      p := Pos(Tabulator, sv);          //looking for the tabulation - the value's separator
      if p = 0 then p := Length(sv) + 1;
      if p > 1 then
      begin
        fv := Copy(sv, 1, p - 1);       //the key requisite value
        sv := Copy(sv, p + 1, MaxInt);
        AssignBufFromStr(FKeyBuffer^.Data + FI^.Offset, fv, FI); //copy the string in range buffer
      end;                              //otherwise the requisition instructions are over
      Inc(i);
    end;
    if i < FKeyBuffer^.KeyFI.Count then
      FKeyBuffer^.FieldCount := i;
  finally
    RestoreState(SaveState);
  end;
end;

procedure TVolgaDataset.SetListOfFields(const KeyFields: string; const KeyValues: array of const);
//assign the values for the several fields from the list immediately (only fkData)
var j, last: integer;
  KL: TList;
  F: TField;
begin
  if not (State in [dsEdit, dsInsert]) then
    DatabaseError(V_NOTEDITMODE, Self);
  KL := TList.Create;
  GetFieldList(KL, KeyFields);
  last := _if(KL.Count - 1 > High(KeyValues), High(KeyValues), KL.Count - 1);
  for j := 0 to last do
  begin
    F := TField(KL.Items[j]);
    if F.ReadOnly then
      DatabaseErrorFmt(V_READONLYFLD, [F.DisplayName], Self);
    if F.FieldNo <= 0 then Continue;    //uneditable fields
    F.AssignValue(KeyValues[j]);
    DataEvent(deFieldChange, longint(F));
  end;
end;

procedure TVolgaDataset.SetMasterFields(const Value: string);
begin
  FMasterLink.FieldNames := Value;      //the master-fields list must be correspondent to list
  //to the list of the detail-fields in the RangeFieldNames!!!
end;

procedure TVolgaDataset.SetOrigin(const Value: string);
begin
  FOrigin := Value;
end;

{$IFDEF PRO}

procedure TVolgaDataset.SetOriginToFields;
var i: integer;
begin
  for i := 0 to FieldCount - 1 do
    Fields[i].Origin := Origin;
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaDataset.SetPassword(const Value: string);
begin
{$IFDEF VOLGAT}
  ShowMessage(textb);
{$ELSE}
  if Value > '' then
    VolgaDB.AddPassword(Value)
  else if FPassword > '' then
    VolgaDB.DeletePassword(FPassword);
  FPassword := Value;
  if Active then
  begin
    FCursor.Password := Value;
    Inc(FNeedToSave);                   //the table must be saved
  end;
{$ENDIF}
end;
{$ENDIF}

procedure TVolgaDataset.SetRange(const StartValues, EndValues: array of const);
begin
  CheckBrowseMode;
  InitKeyBuffer(kiRangeStart);
  SetKeyFields(kiRangeStart, StartValues);
  InitKeyBuffer(kiRangeEnd);
  SetKeyFields(kiRangeEnd, EndValues);
  ApplyRange;
end;

procedure TVolgaDataset.SetRange(const StartValues, EndValues: string);
begin
  if FCursor = nil then Exit;
//  CheckBrowseMode;
  InitKeyBuffer(kiRangeStart);
  SetKeyFields(kiRangeStart, StartValues);
  InitKeyBuffer(kiRangeEnd);
  SetKeyFields(kiRangeEnd, EndValues);
  ApplyRange;
end;

procedure TVolgaDataset.SetRangeEnd;
begin
  SetKeyBuffer(kiRangeEnd, True);
end;

procedure TVolgaDataset.SetRangeStart;
begin
  SetKeyBuffer(kiRangeStart, True);
end;

procedure TVolgaDataset.SetRangeFieldNames(const Value: string);
begin
  FRangeFieldNames := Value;
  if Active then
  begin
    InitKeyBuffer(kiRangeStart);
    InitKeyBuffer(kiRangeEnd);
  end;
  CheckMasterRange;
end;

procedure TVolgaDataset.SetReadOnly(value: Boolean);
begin
  if Active then CheckBrowseMode;
  FReadOnly := value;
end;

procedure TVolgaDataset.SetRecNo(Value: Integer);
begin
  CheckBrowseMode;
  if RecNo = Value then Exit;
  if Value < 1 then
    Value := 1
  else if Value > RecordCount then
  begin
    InternalLast;
    Value := RecordCount;
  end;
  DoBeforeScroll;
  FCurRec := Value - 1;
  Resync([rmCenter]);
  DoAfterScroll;
end;

{$IFDEF PRO}

procedure TVolgaDataset.SetRecordData(const Value: Pointer);
begin
  if Active and (RecordCount > 0) then
    if (RecordData <> nil) and (Value = nil) then
      DoDestroyRec(PVolgaRecord(FSortList.Items[RecNo - 1])^.RecordData);
  PVolgaRecord(FSortList.Items[RecNo - 1])^.RecordData := Value;
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaDataset.SetShowDeleted(const Value: Boolean);
begin
  FShowDeleted := Value;
  if Active then SortRecords;
end;
{$ENDIF}

procedure TVolgaDataset.SetSortOptions(const Value: TVolgaSortOptions);
begin
  FSortOptions := Value;
  IgnoreC := vsoCaseInsensitive in SortOptions; //ignore the letter's register for the string fields?
  DescSort := vsoDescendSort in SortOptions; //decrease sorting?
  if Active then SortRecords;
end;

procedure TVolgaDataset.SetVersion(const Value: string);
begin
  //none
end;

procedure TVolgaDataset.SortRecords(ToFirst: Boolean);
var pRec: PVolgaRecord;
  Accept: Boolean;
  i, acount: integer;
  Oldrec: Pointer;
  OldState: TDataSetState;
begin
{$IFDEF PRO}
  if Filtered and (Filter > '') then
  begin
    AcceptFilterList := FFilterFI;      //fields list for the filter
    CheckFilterFields;
  end;
{$ENDIF}
  OldRec := nil;
  if FLoadedBuf then                    //when called from the InternalOpen, then Active=false!!!!
  begin
    CheckBrowseMode;
    if FSortList.Count > 0 then         //use RecNo, rather then FCurRec !!!!!!!!!!!!!
      OldRec := FSortList.Items[RecNo - 1]; //where we were lacated earlier?
    DisableControls;                    //the counter is increased separately from State
    DoBeforeScroll;
  end;
  FSortList.Clear;
  acount := FCursor.RecordCount;
  FSortList.Capacity := acount;
  //reading all undisposed records in the natural cursor's order
  OldState := SetTempState(dsFilter);
  try
    for i := 0 to acount - 1 do
    begin
      pRec := FCursor.RecData[i];
      if (pRec^.Status = vusAbsent) or ((pRec^.Status = vusDeleted)
{$IFDEF PRO}
        and not FShowDeleted
{$ENDIF}
        ) then Continue;
      //if Range is set, that filter using it
      if FIsRanged and not AcceptRange(pRec) then Continue;
      //if dataset is filtering and record does not suit, then skip it
      if Filtered then
      begin
{$IFDEF PRO}
        if Filter > '' then
        begin                           //filter as expression
          AcceptFilter(FCursor, pRec, Accept);
          if not Accept then Continue;
        end;
{$ENDIF}
        if Assigned(OnFilterRecord) then //filter as procedure
        begin
          Accept := true;
          FFiltRec := pRec;             //record, which is testing in filter
          OnFilterRecord(Self, Accept);
          if not Accept then Continue;
        end;
      end;
      FSortList.Add(pRec);
    end;
  finally
    RestoreState(OldState);
  end;
  //**********************************************
  if FKeyBuffers[kiIndex]^.KeyFI.Count > 0 then
  begin
    IgnoreC := vsoCaseInsensitive in SortOptions; //ignore the the letter's register for string fields?
    DescSort := vsoDescendSort in SortOptions; //decrease sorting?
    FSortFI := FKeyBuffers[kiIndex]^.KeyFI;
    FSortList.Sort(VolgaSort);          //the process of the sorting
    FSortFI := nil;
    FIsSorted := true;                  //dataset is sorted
    IgnoreC := false;
    DescSort := false;
  end;
  //**********************************************
  FCurRec := -1;
  if FLoadedBuf then
  begin
    CursorPosChanged;
    if (FSortList.Count > 0) and (OldRec <> nil) and not ToFirst then //try to return on "former current"
      FCurRec := FSortList.IndexOf(OldRec);
    Resync([]);
    DoAfterScroll;
    EnableControls;                     //the counter is decreased separately from State
  end;
end;

function TVolgaDataset.UpdateStatus: TUpdateStatus;
var BufPtr: PChar;
  pRec: PVolgaRecord;
  RI: PVolgaRecInfo;
begin
  CheckActive;
  if State = dsInternalCalc then
    Result := usUnModified
  else
  begin
    if State = dsCalcFields then
      BufPtr := CalcBuffer
    else
      BufPtr := ActiveBuffer;
    Result := usUnModified;
    RI := PVolgaRecInfo(BufPtr + FRecordSize + CalcFieldsSize);
    pRec := PVolgaRecord(RI^.Bookmark);
    if (pRec <> nil) then
      case pRec^.Status of
        vusModified: Result := usModified;
        vusInserted: Result := usInserted;
        vusDeleted: Result := usDeleted;
      end;
  end;
end;

//******************************************************************************
//          TVolgaTable - working with flat files
//******************************************************************************

constructor TVolgaTable.Create(AOwner: TComponent);
begin
  inherited;
  FLoadMode := lmClear;
  FTableName := '';
end;

procedure TVolgaTable.DestroyCursor;
begin
  if IsCursorOpen and (FLoadMode <> lmClone) then
  begin
    FCursor.Free;
    FCursor := nil;
  end;
end;

function TVolgaTable.BatchMove(ASource: TDataSet;
  AMode: TVolgaBatchMode): integer;
var SourceActive: Boolean;
  iMoved: integer;
  BatchMode: TVolgaBatchMode;
  AST: TVolgaDataset;
  KL, FL: TList;
  F: TField;
  kv, nm: string;
  addnew: Boolean;
  i, j: integer;
begin
{$IFDEF VOLGAT}
  ShowMessage(texta);
{$ENDIF}
  if (ASource = nil) or (ASource = Self) then
    DatabaseError(V_BATCHSOURCE, Self);
  SourceActive := ASource.Active;
  iMoved := 0;
  try
    try
      if not SourceActive then ASource.Open;
    except
      DatabaseError(V_SOURCEOPEN, Self);
    end;
    ASource.DisableControls;
    ASource.CheckBrowseMode;
    ASource.UpdateCursorPos;
    ASource.First;
    BatchMode := AMode;
    if (BatchMode = vbmCopy) or (BatchMode = vbmCopyDef) then
    begin
      Close;
      FieldDefs := ASource.FieldDefs;
      for i := 0 to FieldDefs.Count - 1 do
      begin
        FieldDefs[i].Name := vSpaceToChar(FieldDefs[i].Name);
      end;
      if ASource is TVolgaDataset then
      begin
        AST := TVolgaDataset(ASource);
        Database := AST.Database;
        UniqueFieldNames := AST.UniqueFieldNames;
        RangeFieldNames := AST.RangeFieldNames;
        IndexFieldNames := AST.IndexFieldNames;
        Notes := AST.Notes;
      end;
      if BatchMode = vbmCopy then BatchMode := vbmAppend;
    end;
    Open;                               //the table creation for the lmClear
    DisableControls;
    if BatchMode <> vbmCopyDef then     //otherwise everything is already done
    try
      KL := TList.Create;               //key fields list from the table-source
      FL := TList.Create;               //the rest fields list from the table-source
      if BatchMode = vbmAppendUpdate then
        if UniqueFieldNames = '' then
          BatchMode := vbmAppend        //unique fields are not specified
        else
        begin
          ASource.GetFieldList(KL, UniqueFieldNames); //key fields list of the source
          if KL.Count = 0 then BatchMode := vbmAppend; //Key fields are not found in the source
        end;
      for i := 0 to ASource.FieldCount - 1 do //moving the found in destination fields only
      begin                             //the spaces in the field name are replaced by underscores
        nm := vSpaceToChar(ASource.Fields[i].FieldName);
        if FindField(nm) <> nil then
          FL.Add(ASource.Fields[i]);    //add the field of the source
      end;
      for i := 0 to KL.Count - 1 do
      begin
        j := FL.IndexOf(KL.Items[i]);
        if j >= 0 then FL.Delete(j);    //delete the key fields from the list of the all source fields that are copying
      end;
      while not ASource.Eof do
      begin
        addnew := true;
        if BatchMode = vbmAppendUpdate then
        begin                           //build the string of the source-table key requisites
          kv := '';
          for j := 0 to KL.Count - 1 do
            kv := kv + TField(KL.Items[j]).AsString + Tabulator;
          if FindTab(UniqueFieldNames, kv) then //update the founded record
          begin
            Inc(iMoved);
            addnew := false;
            Edit;
          end;
        end;
        if addnew then                  //if 'add' mode or update when the record is not found
        begin
          Inc(iMoved);
          Append;
          for i := 0 to KL.Count - 1 do //moving key fields
          begin
            F := TField(KL.Items[i]);
            FieldByName(F.FieldName).AsString := F.AsString; //if types are not equal -> types transformation
          end;
        end;
        for i := 0 to FL.Count - 1 do   //moving other fields to the new or founded record
        begin
          F := TField(FL.Items[i]);
          nm := vSpaceToChar(F.FieldName); //spaces replacement
          if not F.IsNull then
          try                           //it works with Graphic fields too (refer to Borland help)
            FieldByName(nm).AsString := F.AsString;
          except;
          end;
        end;
        Post;
        ASource.Next;
      end;
      KL.Free;
      FL.Free;
    finally
      First;
    end;
  finally
    if not SourceActive then ASource.Close;
    EnableControls;
    ASource.EnableControls;
  end;
  Result := iMoved;
end;

procedure TVolgaTable.CreateTable;
var fn: string;
begin
  CheckInactive;
  try
    //create new table in memory
    InternalLoadByClearMode;            //it calls: if FieldDefs.Count = 0 then DatabaseError(V_OPENEMPTY, Self);
    //save it to "tablename"
    fn := GetFullTableName;
    if fn = '' then
      DatabaseErrorFmt(V_OPENFILE, [fn], Self)
    else
      SaveToFile(fn);
  finally
    InternalClose;                      // closes temporary cursor
  end;
end;

procedure TVolgaTable.CancelUpdates;
begin
  inherited;
  if (FLoadMode = lmClone) and (FDatabase <> nil) then
    FDatabase.DeleteUpdated(Self.CloneDS);
end;

procedure TVolgaTable.CommitUpdates;
begin
  inherited;
  if (FLoadMode = lmClone) and (FDatabase <> nil) then
    FDatabase.DeleteUpdated(Self.CloneDS);
end;

procedure TVolgaTable.DefChanged(Sender: TObject);
begin
  StoreDefs := (LoadMode = lmClear);
end;

procedure TVolgaTable.DeleteTable;
var fn: string;
begin
  CheckInactive;
  fn := GetFullTableName;
  if not FileExists(fn) then
    DatabaseErrorFmt(V_OPENFILE, [fn], Self);
  SysUtils.DeleteFile(fn);
end;

function TVolgaTable.FieldDefsStored: Boolean;
begin
  Result := StoreDefs and (FieldDefs.Count > 0);
end;

function TVolgaTable.GetFullTableName: string;
begin
  if (ExtractFileName(FTableName) = FTableName)
    and (FDatabase <> nil) then
    Result := FDatabase.DatabasePath + FTableName
  else
    Result := FTableName;               //load any file!!
end;

{$IFDEF PRO}

function TVolgaTable.GroupBy(ASource: TVolgaDataset; const GroupFlds,
  SumFlds: string; const Kind: TVolgaGroupKind): Boolean;
var GroupL, SumL: TList;
  i, OldNo, avg: integer;
  F: TField;
  OldIndex: string;
  IsComp: Boolean;
  OldOptions: TVolgaSortOptions;
  OldScroll: TDataSetNotifyEvent;
label LN;
begin                                   //create the new table, grouping existing by specified fields
{$IFDEF VOLGAT}
  ShowMessage(texta);
{$ENDIF}
  Result := false;
  if Active or not ASource.Active then Exit; //only for the new and Active tables

  GroupL := TList.Create;
  SumL := TList.Create;
  ASource.GetFieldList(GroupL, GroupFlds);
  ASource.GetFieldList(SumL, SumFlds);
  ASource.DisableControls;
  OldNo := ASource.RecNo;
  OldIndex := ASource.IndexFieldNames;
  OldScroll := ASource.AfterScroll;
  OldOptions := ASource.SortOptions;

  LoadMode := lmClear;
  FieldDefs.Clear;
  for i := 0 to GroupL.Count - 1 do
  begin
    F := TField(GroupL.Items[i]);
    FieldDefs.Add(F.FieldName, F.DataType, F.Size, false);
  end;
  for i := 0 to SumL.Count - 1 do
  begin
    F := TField(SumL.Items[i]);
    case Kind of
      vgkSum, vgkAvg: FieldDefs.Add(F.FieldName, ftFloat, 0, false);
      vgkCount: FieldDefs.Add(F.FieldName, ftInteger, 0, false);
    else
      FieldDefs.Add(F.FieldName, F.DataType, F.Size, false);
    end;
  end;
  try
    Open;                               //the table creation
    ASource.AfterScroll := nil;
    if Kind = vgkMax then
      ASource.SortOptions := ASource.SortOptions + [vsoDescendSort]
    else
      ASource.SortOptions := ASource.SortOptions - [vsoDescendSort];
    if (Kind = vgkMin) or (Kind = vgkMin) then
      ASource.IndexFieldNames := GroupFlds +
        _if(GroupFlds[Length(GroupFlds)] <> ';', ';', '') + SumFlds
    else
      ASource.IndexFieldNames := GroupFlds; //sorting the source-table by the grouping fields
    ASource.First;
    while not ASource.Eof do
    begin
      Append;
      for i := 0 to GroupL.Count - 1 do
      begin
        F := TField(GroupL.Items[i]);
        FieldByName(F.FieldName).AsString := ASource.FieldByName(F.FieldName).AsString;
      end;
      for i := 0 to SumL.Count - 1 do
      begin
        F := TField(SumL.Items[i]);
        FieldByName(F.FieldName).AsString := ASource.FieldByName(F.FieldName).AsString;
      end;
      avg := 1;
      LN: ASource.Next;
      if ASource.Eof then
      begin
        Post;
        Exit;
      end;
      IsComp := true;
      //this record has old key requisites?
      for i := 0 to GroupL.Count - 1 do
      begin
        F := TField(GroupL.Items[i]);
        IsComp := IsComp and (FieldByName(F.FieldName).AsString =
          ASource.FieldByName(F.FieldName).AsString);
      end;
      if IsComp then
      begin
        if ((Kind = vgkSum) or (Kind = vgkAvg)) then //add up the values for summa
          for i := 0 to SumL.Count - 1 do
          begin
            F := TField(SumL.Items[i]);
            FieldByName(F.FieldName).AsFloat := FieldByName(F.FieldName).AsFloat
              + ASource.FieldByName(F.FieldName).AsFloat;
          end;
        Inc(avg);
        goto LN;
      end
      else
      begin
        if (Kind = vgkAvg) or (Kind = vgkCount) then //get the average value
          for i := 0 to SumL.Count - 1 do
          begin
            F := TField(SumL.Items[i]);
            if Kind = vgkAvg then
              FieldByName(F.FieldName).AsFloat := FieldByName(F.FieldName).AsFloat / avg
            else
              FieldByName(F.FieldName).AsFloat := avg;
          end;
        Post;
      end;
    end;
    Result := true;
  finally
    ASource.AfterScroll := OldScroll;
    ASource.SortOptions := OldOptions;
    ASource.IndexFieldNames := OldIndex;
    ASource.RecNo := OldNo;
    ASource.EnableControls;
  end;
end;
{$ENDIF}

{$IFDEF PRO}

procedure TVolgaTable.ImportFromTxt(const FileName: string;
  const Delimiter: char);
var i, j: integer;
  s: string;
  TxtFile, ListFlds, KeyFlds, Rec: TStrings;
  keyf: array of integer;
  CanUpdate: Boolean;

  function ExtractFromQuotes(s: string): string;
  var a: PChar;
    Quote: char;
  begin
    Result := s;
    if Length(s) < 2 then Exit;         //no any quotes pair
    //quotes are used when Delimiter is comma or semicolon
    if (s[1] = Quote2) and (Delimiter <> Tabulator) then
      Quote := Quote2
    else if (s[1] = Quote1) and (Delimiter <> Tabulator) then
      Quote := Quote1
    else
      Exit;
    a := AllocMem(Length(s) + 1);
    a := StrPCopy(a, s);
    Result := AnsiExtractQuotedStr(a, Quote);
    FreeMem(a);
  end;

begin
  if not FileExists(FileName) then
    DatabaseErrorFmt(V_OPENFILE, [FileName], Self);
  CheckActive;
  CheckBrowseMode;
  TxtFile := TStringList.Create;
  TxtFile.LoadFromFile(FileName);
  if TxtFile.Count > 0 then
  begin
    DisableControls;
    KeyFlds := TStringList.Create;
    CanUpdate := false;                 //Append mode only or we can update records using key fields
    ListFlds := TStringList.Create;
    Rec := TStringList.Create;
    //in first line we want to see field names with delimiters
    s := TxtFile[0];
    Split(s, Delimiter, ListFlds);
    if UniqueFieldNames > '' then
    begin
      Split(UniqueFieldNames, ';', KeyFlds);
      SetLength(keyf, KeyFlds.Count);
      for j := 0 to KeyFlds.Count - 1 do
        keyf[j] := -1;
      for i := 0 to ListFlds.Count - 1 do
        for j := 0 to KeyFlds.Count - 1 do
          if AnsiUpperCase(ListFlds[i]) = AnsiUpperCase(KeyFlds[j]) then
          begin
            keyf[j] := i;
            Break;
          end;
      CanUpdate := keyf[KeyFlds.Count - 1] <> -1; //all key fields found in imported fields
    end;
    for i := 1 to TxtFile.Count - 1 do
    begin
      Split(TxtFile[i], Delimiter, Rec);
      if CanUpdate then
      begin
        s := '';
        for j := 0 to KeyFlds.Count - 1 do
          s := s + ExtractFromQuotes(Rec[keyf[j]]) + #9;
        SetLength(s, Length(s) - 1);
        if FindTab(UniqueFieldNames, s) then
          Edit
        else
          Append;
      end
      else
        Append;
      for j := 0 to ListFlds.Count - 1 do
      begin
        if ExistField(ListFlds[j]) then
        try                             //try to convert string to field value
          if j < Rec.Count then
            FieldByName(ListFlds[j]).AsString := ExtractFromQuotes(Rec[j]);
        except                          //if we can't convert, we skip this field
        end;
      end;
      Post;
    end;
    EnableControls;
    Rec.Free;
    ListFlds.Free;
  end;
  TxtFile.Free;
end;
{$ENDIF}

procedure TVolgaTable.InternalApplyUpdates;
var fn: string;
  VT: TVolgaTable;
begin
  if not NeedToSave then Exit;
  //save the changes to file or to the clone's source-file
  if (FLoadMode = lmClone) and (CloneDS is TVolgaTable) then
    VT := CloneDS as TVolgaTable
  else if FLoadMode = lmFile then
    VT := Self
  else
    Exit;
  fn := VT.GetFullTableName;            //tablename of original table
  if fn > '' then SaveToFile(fn);
  if FLoadMode = lmClone then CloneDS.SortRecords;
end;

procedure TVolgaTable.InternalClose;
begin
  if (LoadMode = lmClone) and (CloneDS <> nil) then
    CloneDS.RemoveClone(Self);          //decrease clone dataset counter
  inherited;
end;

procedure TVolgaTable.InternalLoadByClearMode;
var {aorigin,}  s: string;
  rec, len: integer;
  typ: char;
  Stream: TMemoryStream;
begin
  if FieldDefs.Count = 0 then
    DatabaseError(V_OPENEMPTY, Self);

  s := PrepareLayoutStr(FieldDefs);     //preparing the string of the fields description
  Stream := TMemoryStream.Create;
  len := SIZEINT * 4 + 1 + Length(s);
  Stream.SetSize(len);
  Stream.Write(len, SIZEINT);           //the buffer length
  typ := 'S';
  Stream.Write(typ, 1);                 //success
  rec := 0;
  Stream.Write(rec, SIZEINT);           //record count:= 0
  rec := FieldDefs.Count;
  Stream.Write(rec, SIZEINT);           //fields count
  rec := Length(s);
  Stream.Write(rec, SIZEINT);           //the header length
  Stream.Write(PChar(s)^, Length(s));   //fields info
  Stream.Position := 0;                 //from the beginnig of the stream
  InternalLoad(Stream);                 //Stream is freed there!!!
end;

procedure TVolgaTable.InternalLoadFromCursor;
begin
  if not CloneDS.Active then
    DatabaseError(V_ERRCLONE, Self);
  CloneDS.AddClone(Self);               //the clones counter increasing
end;

procedure TVolgaTable.InternalLoadFromFile;
var Stream: TMemoryStream;
  fn: string;
begin
  //if the database way is specified, then use it
  fn := GetFullTableName;
  if not FileExists(fn) then
    DatabaseErrorFmt(V_OPENFILE, [fn], Self);
  Stream := TMemoryStream.Create;
  Stream.LoadFromFile(fn);
  InternalLoad(Stream);                 //the Stream is freed there
end;

procedure TVolgaTable.InternalLoadData;
begin
  case FLoadMode of
    lmClear:                            //empty dataset
      InternalLoadByClearMode;
    lmFile:                             //loading from data file from disk
      InternalLoadFromFile;
    lmClone:                            //loading from source dataset
      InternalLoadFromCursor;
    lmStream:
      if FExternalStream <> nil then
        InternalLoad(FExternalStream, true) //save the Stream for user
      else
        DatabaseError(V_STREAMNOTEXIST, Self);
  end;
end;

procedure TVolgaTable.LoadFromCursor(CloneDB: TVolgaDataset);
begin
  if Active then Close;
  LoadMode := lmClone;
  CloneDS := CloneDB;
  Open;
end;

procedure TVolgaTable.LoadFromFile(const Fn: string);
begin
  if Active then Close;
  TableName := Fn;                      //setTableName and setLoadMode are calling!!
  Open;
end;

{$IFDEF PRO}

procedure TVolgaTable.LoadFromStream(AStream: TStream);
begin
  if Active then Close;
  LoadMode := lmStream;
  FExternalStream := AStream;           //user must create and kill the stream himself
  //start reading from the current position of the stream
  Open;
  FExternalStream := nil;               //in the AStream the pointer is located in the end of data
end;

function TVolgaTable.GetTableNotes: string;
begin                                   //note in the table
  CheckActive;
  Result := FCursor.Notes;
end;

procedure TVolgaTable.SetTableNotes(const Value: string);
begin
  CheckActive;
  FCursor.Notes := Value;
  Inc(FNeedToSave);                     //table must be saved
end;
{$ENDIF}

procedure TVolgaTable.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  //if cloned dataset was removed from form
  if (Operation = opRemove) and (AComponent = FCloneDS) then
  begin
    if Active then Close;
    CloneDS := nil;
  end;
  //if Database was removed from form
  if (Operation = opRemove) and (AComponent = FDatabase) then
  begin
    if Active then Close;
    FDatabase := nil;
  end;
end;

function TVolgaTable.PrepareLayoutStr(FD: TFieldDefs): string;
var
  FI: TFieldDef;
  dlina, nn: string;
  ft: TVolgaFieldType;
  typ: char;
  i: integer;
begin
  Result := '';
  for i := 0 to FD.Count - 1 do
  begin
    FI := TFieldDef(FD.Items[i]);
    dlina := '';
    ft := GetFldType(FI.DataType, FI.Size);
    typ := GetCharType(ft, FI.Size);
    if ft = vftString then dlina := IntToStr(FI.Size);
    nn := vSpaceToChar(FI.Name);        //enderscores instead of spaces
    Result := Result + 'XXX.' + nn + ' ' + typ + dlina + FieldSeparator;
  end;
  Result := Result + RecordSeparator;
end;

procedure TVolgaTable.RenameTable(const NewTableName: string);
var fn, newfn: string;
begin
  CheckInactive;
  fn := GetFullTableName;
  if not FileExists(fn) then
    DatabaseErrorFmt(V_OPENFILE, [fn], Self);
  if (ExtractFileName(NewTableName) = NewTableName)
    and (FDatabase <> nil) then
    newfn := FDatabase.DatabasePath + NewTableName
  else
    newfn := NewTableName;
  if FileExists(newfn) then
    SysUtils.DeleteFile(newfn);
  SysUtils.RenameFile(fn, newfn);
  TableName := NewTableName;
end;

procedure TVolgaTable.SetCloneDS(value: TVolgaDataset);
begin
  if value = nil then
  begin
    FCloneDS := nil;
    LoadMode := lmClear;
  end
  else
  begin
    FCloneDS := value;
    LoadMode := lmClone;
  end;
end;

procedure TVolgaTable.SetDatabase(const Value: TVolgaDatabase);
begin
  inherited;
  if (FDatabase <> nil) and (FDatabase.DatabasePath > '') and (FTableName > '') then
    FTableName := ExtractFileName(FTableName);
end;

{$IFDEF PRO}

procedure TVolgaTable.SetLastAutoInc(NewValue: integer);
begin
  CheckActive;
  CheckBrowseMode;
  FCursor.SetLastAutoInc(NewValue);
end;
{$ENDIF}

procedure TVolgaTable.SetTableName(const value: string);
begin
  if FTableName = value then Exit;
  if Active then Close;
  FTableName := value;
  if FTableName <> '' then
    LoadMode := lmFile
  else
    LoadMode := lmClear;
end;

procedure TVolgaTable.SetLoadMode(value: TVolgaLoadMode);
begin
  if FLoadMode = value then Exit;
  if Active then Close;                 //close dataset when the loading method was changed
  FLoadMode := value;
end;

{$IFDEF PRO}
//==============================================================================
//         TVolgaCustomQuery
//==============================================================================

constructor TVolgaCustomQuery.Create(AOwner: TComponent);
begin
  inherited;
  FPrepared := false;
  FSQL := TStringList.Create;
  TStringList(SQL).OnChange := QueryChanged;
  FParams := TParams.Create(Self);
end;

procedure TVolgaCustomQuery.DefineProperties(Filer: TFiler);

  function WriteData: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not FParams.IsEqual(TVolgaCustomQuery(Filer.Ancestor).FParams)
    else
      Result := FParams.Count > 0;
  end;

begin                                   //functions for writing parameters in dfm-file
  inherited DefineProperties(Filer);
  Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end;

procedure TVolgaCustomQuery.DestroyCursor;
begin
  UnPrepare;
  if IsCursorOpen then FCursor.Free;
  FCursor := nil;
  FSQL.Free;
  FParams.Free;
end;

function TVolgaCustomQuery.GetParamsCount: Word;
begin
  Result := FParams.Count;
end;

function TVolgaCustomQuery.GetSQL: TStrings;
begin
  Result := FSQL;
end;

procedure TVolgaCustomQuery.InternalClose;
begin
  inherited;
  UnPrepare;
end;

procedure TVolgaCustomQuery.InternalRefresh;
begin
  UnPrepare;
  inherited;
end;

function TVolgaCustomQuery.ParamByName(const Value: string): TParam;
begin
  Result := FParams.ParamByName(Value);
end;

function TVolgaCustomQuery.GetPreparedSQL(DecSep, DatSep: char; DatFmt: string): string;
//parse the SQL text and prepare it's execution
var AText: string;
  i: integer;
  quot: char;
  OldSep, OldDat: char;
begin
  quot := '"';
  if FSQL.Count = 0 then
    DatabaseError(V_EMPTYSQL, Self);
  OldSep := DecimalSeparator;
  OldDat := DateSeparator;
  DecimalSeparator := DecSep;
  DateSeparator := DatSep;
  //insert the parameters in request
  AText := FText;                       //copy and so we can anew reinsert parameters when REFRESH!!!!!!!
  SetLength(AText, vStrLen(PChar(AText))); //delete the trash, remaining after parser's working
  for i := 0 to ParamCount - 1 do
  begin
    if (FParams[i].IsNull) and (FParams[i].DataType <> ftString) then
      AText := StringReplace(AText, '?', 'Null', [])
    else if (FParams[i].DataType = ftString) then
      AText := StringReplace(AText, '?', AnsiQuotedStr(FParams[i].AsString, quot), [])
    else if (FParams[i].DataType = ftDate) or (FParams[i].DataType = ftDateTime)
      or (FParams[i].DataType = ftTime) then
      AText := StringReplace(AText, '?', Database.DateQuot +
        FormatDateTime(Database.DateFormat,
        FParams[i].AsDateTime) + Database.DateQuot, [])
    else
      AText := StringReplace(AText, '?', FParams[i].AsString, []);
  end;
  DecimalSeparator := OldSep;
  DateSeparator := OldDat;
  Result := AText;
end;

procedure TVolgaCustomQuery.QueryChanged(Sender: TObject);
//the SQL-text has changed?
var List: TParams;
begin
  if not (csReading in ComponentState) then
  begin
    if Active then Close;
    UnPrepare;
    List := TParams.Create(Self);
    try
      FText := List.ParseSQL(SQL.Text, True);
      List.AssignValues(FParams);
      FParams.Clear;
      FParams.Assign(List);             //create the parameters list
    finally
      List.Free;
    end;
    DataEvent(dePropertyChange, 0);
  end
  else
    FText := FParams.ParseSQL(SQL.Text, False);
end;

procedure TVolgaCustomQuery.ReadParamData(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FParams);
end;

procedure TVolgaCustomQuery.SetParamsList(const Value: TParams);
begin
  FParams.AssignValues(Value);
end;

procedure TVolgaCustomQuery.SetSQL(const Value: TStrings);
begin
  if SQL.Text <> Value.Text then
  begin
    if Active then Close;
    UnPrepare;
    FSQL.BeginUpdate;
    try
      FSQL.Assign(Value);
    finally
      FSQL.EndUpdate;
    end;
  end;
end;

procedure TVolgaCustomQuery.WriteParamData(Writer: TWriter);
begin
  Writer.WriteCollection(Params);
end;

procedure TVolgaCustomQuery.SetOriginToFields;
begin
  inherited SetOriginToFields;
end;
{$ENDIF}

//******************************************************************************
//          TVolgaDatabase
//******************************************************************************

constructor TVolgaDatabase.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDSList := TList.Create;
  FDSList.Clear;
  FUpdated := TList.Create;
  FUpdated.Clear;
{$IFDEF PRO}
  FDateQuot := '"';
  FDateFormat := 'mm/dd/yyyy';
{$ENDIF}
  FDatabasePath := '';
  VolgaDB.Add(Self);
end;

destructor TVolgaDatabase.Destroy;
begin
  FUpdated.Free;
  Close;
  FDSList.Free;
  VolgaDB.Delete(Self);
  inherited;
end;

procedure TVolgaDatabase.Add(DS: TVolgaDataset);
begin
  if FDSList.IndexOf(DS) < 0 then
    FDSList.Add(DS);
end;

procedure TVolgaDatabase.AddUpdated(DS: TVolgaDataset);
begin
  if FUpdated.IndexOf(DS) < 0 then
    FUpdated.Add(DS);
end;

procedure TVolgaDatabase.ApplyUpdates(const ADataSets: array of TVolgaDataset);
var i: integer;
  sss: string;
  DS: TVolgaDataset;
  AD: array of Pointer;
begin
  if UpdatedCount = 0 then Exit;        //there were no updates in this database
  sss := '';
  if Length(ADatasets) = 0 then
  begin
    SetLength(AD, FUpdated.Count);
    for i := 0 to FUpdated.Count - 1 do
      AD[i] := TVolgaDataset(FUpdated.Items[i]);
  end
  else
  begin
    SetLength(AD, High(ADataSets) + 1);
    for i := 0 to High(ADataSets) do
      AD[i] := ADataSets[i];
  end;
  try
    for I := 0 to High(AD) do
    begin
      DS := TVolgaDataset(AD[I]);
      if DS.Database <> Self then
        DatabaseErrorFmt(V_WRONGDATASET, [DS.Name]);
      if DS.State in [dsInsert, dsEdit] then DS.Post;
        DS.ApplyUpdates;
    end;
  except
    raise;
  end;
  for I := 0 to High(AD) do
    TVolgaDataset(AD[I]).CommitUpdates;
end;

procedure TVolgaDatabase.CancelUpdates(const ADataSets: array of TVolgaDataset);
var i: integer;
begin
  if Length(ADataSets) = 0 then         //CancelUpdates for all updated tables
    for i := UpdatedCount - 1 downto 0 do
    begin
      TVolgaDataset(FUpdated.Items[i]).CancelUpdates;
    end
  else
    for i := High(ADataSets) downto 0 do
    begin
      if ADataSets[i].Database <> Self then
        DatabaseErrorFmt(V_WRONGDATASET, [ADataSets[i].Name]);
      ADataSets[i].CancelUpdates;
    end;
end;

procedure TVolgaDatabase.Close;
var i: integer;
begin
  if FDSList.Count > 0 then
  begin
    //close all opened tables
    for i := FDSList.Count - 1 to 0 do
      TVolgaDataset(FDSList.Items[i]).Close;
    FDSList.Clear;
  end;
end;

procedure TVolgaDatabase.DeleteUpdated(DS: TVolgaDataset);
begin
  FUpdated.Remove(DS);
end;

function TVolgaDatabase.GetDataSet(Index: integer): TVolgaDataset;
begin
  if (Index >= 0) and (Index < DataSetCount) then
    Result := TVolgaDataset(FDSList.Items[Index])
  else
    Result := nil;
end;

function TVolgaDatabase.GetDataSetCount: integer;
begin
  Result := FDSList.Count;
end;

function TVolgaDatabase.GetUpdatedCount: integer;
begin
  Result := FUpdated.Count;
end;

function TVolgaDatabase.GetUpdatedDataSet(Index: integer): TVolgaDataset;
begin
  if (Index >= 0) and (Index < UpdatedCount) then
    Result := TVolgaDataset(FUpdated.Items[Index])
  else
    Result := nil;
end;

procedure TVolgaDatabase.Remove(DS: TVolgaDataset);
begin
  FDSList.Remove(DS);
end;

procedure TVolgaDatabase.SetDatabasePath(const Value: string);
const slash = {$IFDEF Linux} '/'{$ELSE} '\'{$ENDIF Linux};
begin
  if FDatabasePath <> Value then
  begin
    FDatabasePath := Value;
    if (FDatabasePath > '') and (FDatabasePath[Length(FDatabasePath)] <> slash)
    then FDatabasePath := FDatabasePath + slash;
  end;
end;

{$IFDEF PRO}

procedure TVolgaDatabase.Restructure(const Fn: string; AList: TStrings);
type
  TDefFld = record
    NewName: string;
    OldName: string;
    DataType: TFieldType;
    Size: integer;
  end;
  PDefFld = ^TDefFld;
var oldTable, newTable: TVolgaTable;
  Def: PDefFld;
  i, j: integer;
  line, notes, passw: string;

  procedure FreeListDef;
  var i: integer;
  begin
    for i := 0 to AList.Count - 1 do
      if AList.Objects[i] <> nil then
        Dispose(PDefFld(AList.Objects[i]));
  end;

begin
{$IFDEF VOLGAT}
  ShowMessage(texta);
{$ENDIF}
  //one string of the following type: NewField=OldField,DataType,size
  if (AList.Count = 0) or (Fn = '') then
    DatabaseError(V_RESTRNODATA, Self);
  //move the descriptions list in the structures array
  for i := 0 to AList.Count - 1 do
  begin
    line := AList[i];
    New(Def);
    Def^.NewName := '';
    Def^.OldName := '';
    Def^.DataType := ftString;
    Def^.Size := 0;
    j := Pos('=', line);
    if j > 1 then
    try
      Def^.NewName := Copy(line, 1, j - 1);
      line := Copy(Line, j + 1, MaxInt);
      j := Pos(',', line);
      if j = 0 then j := Length(line) + 1;
      if j > 1 then                     //name of the old field (if it presents)
        Def^.OldName := Copy(line, 1, j - 1);
      line := Copy(Line, j + 1, MaxInt);
      j := Pos(',', line);
      if j = 0 then j := Length(line) + 1;
      if j > 1 then                     //type of the new field
        Def^.DataType := GetFldType2(Copy(line, 1, j - 1));
      line := Copy(Line, j + 1, MaxInt);
      if Length(line) > 0 then
        Def^.Size := StrToInt(line);
      AList.Objects[i] := TObject(Def);
    except
      Dispose(Def);
      FreeListDef;
      DatabaseError(V_RESTRLISTERROR, Self);
    end
    else
    begin
      FreeListDef;
      DatabaseError(V_RESTRLISTERROR, Self);
    end;
  end;
  //open the old table
  oldTable := TVolgaTable.Create(Application);
  oldTable.Database := Self;
  oldTable.TableName := Fn;
  try
    oldTable.Open;                      //if some troubles - we are not guilty
    notes := oldTable.Notes;
    passw := oldTable.Password;
    //create new FieldDefs and table using it
    newTable := TVolgaTable.Create(Application);
    newTable.LoadMode := lmClear;
    newTable.FieldDefs.Clear;
    newTable.SaveOptions := [savAutoBackup];
    newtable.Notes := notes;
    for i := 0 to AList.Count - 1 do
    begin
      Def := PDefFld(AList.Objects[i]);
      newTable.FieldDefs.Add(Def^.NewName, Def^.DataType, Def^.Size, false);
    end;
    try
      newTable.Open;
      //copy descriptions and default values
      for i := 0 to AList.Count - 1 do
      begin
        Def := PDefFld(AList.Objects[i]);
        if Def^.OldName > '' then
        begin
          newTable.SetDescription(Def^.NewName, oldTable.GetDescription(Def^.OldName));
          newTable.SetDefaultValue(Def^.NewName, oldTable.GetDefaultValue(Def^.OldName));
        end;
      end;
      //copy data
      while not oldTable.Eof do
      begin
        newTable.Append;
        for i := 0 to AList.Count - 1 do
        begin
          Def := PDefFld(AList.Objects[i]);
          if Def^.OldName > '' then
            newTable.FieldByName(Def^.NewName).AsString :=
              oldTable.FieldByName(Def^.OldName).AsString;
        end;
        newTable.Post;
        oldTable.Next;
      end;
      //save data from the new table with this filename
      newTable.Password := passw;
      newTable.SaveToFile(DatabasePath + Fn);
    finally
      newTable.Free;
    end;
  finally
    FreeListDef;
    oldTable.Free;
  end;
end;
{$ENDIF}

//==============================================================================
// TVolgaSession
//==============================================================================

constructor TVolgaSession.Create;
begin
  FDatabases := TList.Create;
{$IFDEF PRO}
  Sapphire := TSapphire.Create;
  FPasswords := TStringList.Create;
  FPasswords.Sorted := true;
{$IFNDEF Linux}
  CoInitialize(nil);
{$ENDIF non-Linux}
{$ENDIF}
end;

destructor TVolgaSession.Destroy;
var i: integer;
begin
  for i := 0 to FDatabases.Count - 1 do
    TVolgaDatabase(FDatabases[i]).Free;
  inherited;
  FDatabases.Free;
{$IFDEF PRO}
  Sapphire.Free;
  FPasswords.Free;
{$IFNDEF Linux}
  CoUninitialize;
{$ENDIF non-Linux}
{$ENDIF}
end;

procedure TVolgaSession.Add(Database: TVolgaDatabase);
begin
  FDatabases.Add(Database);
end;

procedure TVolgaSession.Delete(Database: TVolgaDatabase);
begin
  FDatabases.Remove(Database);
end;

function TVolgaSession.GetDatabase(Index: Integer): TVolgaDatabase;
begin
  Result := FDatabases[Index];
end;

function TVolgaSession.GetDatabaseCount: Integer;
begin
  Result := FDatabases.Count;
end;

procedure TVolgaSession.GetTableNames(const DatabasePath, Pattern: string;
  Extensions: Boolean; List: TStrings);
var TempDatabase: TVolgaDatabase;
  i: integer;
begin
  for i := 0 to FDatabases.Count - 1 do
  begin
    TempDatabase := TVolgaDatabase(FDatabases[i]);
    if (AnsiUpperCase(TempDatabase.DatabasePath) = AnsiUpperCase(DatabasePath))
    then
    begin
      GetTableNames(TempDatabase, Pattern, Extensions, List);
      Exit;
    end;
  end;
  TempDatabase := TVolgaDatabase.Create(nil);
  TempDatabase.DatabasePath := DatabasePath;
  GetTableNames(TempDatabase, Pattern, Extensions, List);
  TempDatabase.Free;
end;

procedure TVolgaSession.GetTableNames(const Database: TVolgaDatabase;
  Pattern: string; Extensions: Boolean; List: TStrings);
var rez: Integer;
  SearchRec: TSearchRec;

  function GetName: string;
  var i: integer;
  begin
    Result := ExtractFileName(SearchRec.Name);
    if not Extensions then
    begin
      i := Pos('.', Result);
      if i > 0 then
        Result := Copy(Result, 1, i - 1);
    end;
  end;

begin
  rez := SysUtils.FindFirst(Database.DatabasePath + Pattern, faAnyFile, SearchRec);
  if rez = 0 then
  begin
    List.Add(GetName);
    while (SysUtils.FindNext(SearchRec) = 0) do
      List.Add(GetName);
  end;
  SysUtils.FindClose(SearchRec);
end;

{$IFDEF PRO}

procedure TVolgaSession.AddPassword(const passw: string);
var i: integer;
begin
  if (passw > '') and not FPasswords.Find(passw, i) then
    FPasswords.Add(passw);
end;

procedure TVolgaSession.DeletePassword(const passw: string);
var i: integer;
begin
  if FPasswords.Find(passw, i) then
    FPasswords.Delete(i);
end;

procedure TVolgaSession.ClearPasswords;
begin
  FPasswords.Clear;
end;

function TVolgaSession.CheckPassw(Original: PChar; var newpasw: string): Boolean;
var i: integer;
  new: string;
  tryagain: Boolean;

  function TryPassw(psw: string): Boolean;
  var EncBuf: PChar;

    function CheckCrc: Boolean;
    var k: integer;
    begin
      crcl := $FF;
      crch := $FF;
      for k := 0 to LenPassw - 1 do
        Crc(ord(EncBuf[k]));
      Result := (crcl = 0) and (crch = 0);
    end;

  begin
    Result := false;
    EncBuf := AllocMem(LenPassw);
    Move(Original^, EncBuf^, LenPassw);
    Sapphire.Key := psw;
    try
      Sapphire.DecryptBuf(EncBuf, LenPassw - 2); //without CRC
      if CheckCrc and (EncBuf[0] = '{') and (EncBuf[LenPassw - 3] = '}') then
      begin
        Result := true;      //give the good GUID
        //write the decoded password
        Move(EncBuf^, Original^, LenPassw);
        newpasw := psw;
      end;
    finally
      FreeMem(EncBuf);
    end;
  end;

begin
  Result := false;
  //try to sent the password first
  if newpasw > '' then
  begin
    Result := TryPassw(newpasw);
    if Result then AddPassword(new);    //add the new password in session
    if Result then Exit;
  end;
  //decipher the sent key by the means of all passwords from the list
  for i := 0 to FPasswords.Count - 1 do
  begin
    Result := TryPassw(FPasswords[i]);
    if Result then Exit;
  end;
  //if the password is wrong -  call the password dialogue thrice
  if FTryCount < 1 then FTryCount := 3;
  for i := 1 to FTryCount do
  begin
    if not GetPassword(new) then Exit;
    //decipher the sent key by the means of the entered password
    Result := TryPassw(new);
    if Result then AddPassword(new);    //add new password in session
    if Result then Exit;
    if Assigned(FOnInvalidPassword) then
    begin
      tryagain := i < FTryCount;
      FOnInvalidPassword(new, tryagain);
      if tryagain then Continue;
    end
    else if i < FTryCount then
      ShowMessage(V_TRYAGAIN);
  end;
end;

function LoginDlg(const ACaption, APrompt: string;
  var Value: string): Boolean;
var
  Form: TForm;
  Prompt: TLabel;
  Edit: TEdit;
begin
  Result := False;
  Form := TForm.Create(Application);
  with Form do
  try
    Caption := ACaption;
    ClientHeight := 116;
    ClientWidth := 205;
{$IFDEF Linux}
    BorderStyle := fbsDialog;
{$ELSE}
    BorderStyle := bsDialog;
{$ENDIF Linux}
    Position := poScreenCenter;
    Prompt := TLabel.Create(Form);
    with Prompt do
    begin
      Parent := Form;
      AutoSize := false;
      SetBounds(18, 7, 170, 28);
      WordWrap := true;
      Caption := APrompt;
    end;
    Edit := TEdit.Create(Form);
    with Edit do
    begin
      Parent := Form;
      SetBounds(Prompt.Left, 45, 170, Height);
      MaxLength := 255;
{$IFDEF Linux}
      EchoMode := emPassword;
{$ELSE}
      PasswordChar := '*';
{$ENDIF Linux}
      Text := Value;
      SelectAll;
    end;
    with TButton.Create(Form) do
    begin
      Parent := Form;
      ModalResult := mrOk;
      Default := True;
      Caption := SMsgDlgOK;
      SetBounds(18, 77, 75, 25);
    end;
    with TButton.Create(Form) do
    begin
      Parent := Form;
      ModalResult := mrCancel;
      Cancel := True;
      SetBounds(113, 77, 75, 25);
      Caption := SMsgDlgCancel;
    end;
    if ShowModal = mrOk then
    begin
      Value := Edit.Text;
      Result := True;
    end;
  finally
    Form.Release;
  end;
end;

function TVolgaSession.GetPassword(var NewPassw: string): Boolean;
begin
  if Assigned(FOnPassword) then
  begin
    Result := FOnPassword(NewPassw)
  end
  else
    Result := LoginDlg(V_PASSWCAPTION, V_PASSWPROMPT, NewPassw);
end;

function TVolgaSession.CheckStreamHeader(AStream: TStream): Boolean;
var MyCursor: TVolgaCursor;
  s: string;
begin  //check if a stream has correct Volga header and can be loded into a table
  MyCursor := TVolgaCursor.Create;
  s := '';
  Result := true;
  if not MyCursor.LoadFieldInfo(AStream, s) or (s > '') then
    Result := false;
  MyCursor.Free;
end;
{$ENDIF}

initialization
  VolgaDB := TVolgaSession.Create;
{$IFDEF PRO}
  VolgaDB.TryCount := 3;
{$ENDIF}

finalization
  VolgaDB.Free;

end.

